"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const inversify_1 = require("inversify");
const jszip_1 = tslib_1.__importDefault(require("jszip"));
const md5_1 = require("../../../tools/hash/md5");
const CartridgeDetector_1 = tslib_1.__importDefault(require("../../../machine/stella/cartridge/CartridgeDetector"));
const Storage_1 = tslib_1.__importDefault(require("./Storage"));
let AddCartridge = class AddCartridge {
    constructor(_storage) {
        this._storage = _storage;
        this._addCartridge = () => {
            if (this._input) {
                this._input.removeEventListener('change', this._onCartridgeAdded);
            }
            this._input = document.createElement('input');
            this._input.type = 'file';
            this._input.accept = '.bin, .a26, .zip';
            this._input.multiple = true;
            this._input.addEventListener('change', this._onCartridgeAdded);
            this._input.click();
        };
        this._onCartridgeAdded = (e) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const target = e.target;
            const cartridges = (yield Promise.all(Array.prototype.map.call(target.files, (f) => this._processFile(f)))).reduce((acc, x) => acc.concat(x), []);
            const cartridgesDeduped = Array.from(new Map(cartridges.map(c => [c.cartridge.hash, c])).values());
            yield this._storage.insertCartridges(cartridgesDeduped);
            this._ports.onNewCartridges_.send(Array.from(cartridgesDeduped.map(c => c.cartridge)));
        });
        this._input = null;
        this._detector = new CartridgeDetector_1.default();
    }
    init(ports) {
        this._ports = ports;
        ports.addCartridge_.subscribe(this._addCartridge);
    }
    _processFile(file) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            try {
                const content = yield new Promise((resolve, reject) => {
                    const reader = new FileReader();
                    reader.addEventListener('load', () => resolve(new Uint8Array(reader.result)));
                    reader.addEventListener('error', () => reject(reader.error));
                    reader.readAsArrayBuffer(file);
                });
                if (file.name.match(/.zip$/i)) {
                    const zip = new jszip_1.default();
                    yield zip.loadAsync(content);
                    return yield Promise.all(zip
                        .file(/\.(bin|a26)$/i)
                        .filter(f => !f.dir)
                        .map(f => f.async('uint8array').then(c => this._createCartridge(f.name.replace(/.*\//, ''), c))));
                }
                else {
                    return [this._createCartridge(file.name, content)];
                }
            }
            catch (_) {
                return [];
            }
        });
    }
    _createCartridge(filename, content) {
        const hash = md5_1.calculateFromUint8Array(content);
        const name = filename.replace(/\.[^\.]*$/, '');
        const tvMode = this._tvModeFromName(name);
        const cartridgeType = this._detector.detectCartridgeType(content);
        return {
            cartridge: {
                hash,
                name,
                tvMode,
                cartridgeType,
                emulatePaddles: false,
                volume: 100
            },
            image: content
        };
    }
    _tvModeFromName(name) {
        if (name.match(/[\W\-_]pal\d*[\W\-_]/i)) {
            return "pal";
        }
        if (name.match(/[\W\-_]secam\d*[\W\-_]/i)) {
            return "secam";
        }
        return "ntsc";
    }
};
AddCartridge = tslib_1.__decorate([
    inversify_1.injectable(),
    tslib_1.__param(0, inversify_1.inject(Storage_1.default)),
    tslib_1.__metadata("design:paramtypes", [Storage_1.default])
], AddCartridge);
exports.default = AddCartridge;
//# sourceMappingURL=AddCartridge.js.map