"use strict";
var Emulation_1;
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const inversify_1 = require("inversify");
const async_mutex_1 = require("async-mutex");
const EmulationServiceInterface_1 = tslib_1.__importDefault(require("../../../web/stella/service/EmulationServiceInterface"));
const EmulationService_1 = tslib_1.__importDefault(require("../../../web/stella/service/worker/EmulationService"));
const DriverManager_1 = tslib_1.__importDefault(require("../../../web/stella/service/DriverManager"));
const Factory_1 = tslib_1.__importDefault(require("../../../machine/cpu/Factory"));
const Video_1 = tslib_1.__importDefault(require("../../../web/driver/Video"));
const KeyboardIO_1 = tslib_1.__importDefault(require("../../../web/stella/driver/KeyboardIO"));
const WebAudio_1 = tslib_1.__importDefault(require("../../../web/stella/driver/WebAudio"));
const Storage_1 = tslib_1.__importDefault(require("./Storage"));
const FullscreenVideo_1 = tslib_1.__importDefault(require("../../../web/driver/FullscreenVideo"));
const TouchIO_1 = tslib_1.__importDefault(require("../../../web/stella/driver/TouchIO"));
const MouseAsPaddle_1 = tslib_1.__importDefault(require("../../../web/driver/MouseAsPaddle"));
const Gamepad_1 = tslib_1.__importDefault(require("../../../web/driver/Gamepad"));
const CANVAS_ID = 'stellerator-canvas';
const WORKER_URL = 'worker/stellerator.min.js';
function error(msg) {
    return {
        state: "error",
        error: msg
    };
}
function cpuType(cartidge, settings) {
    const cpuEmulation = cartidge.cpuEmulation || settings.cpuEmulation;
    switch (cpuEmulation) {
        case "cycle":
            return Factory_1.default.Type.stateMachine;
        case "instruction":
            return Factory_1.default.Type.batchedAccess;
        default:
            throw new Error(`cannot happen: invalid cpu emulation mod e${cpuEmulation}`);
    }
}
function tvMode(cartridge) {
    switch (cartridge.tvMode) {
        case "ntsc":
            return 0;
        case "pal":
            return 1;
        case "secam":
            return 2;
        default:
            throw new Error(`cannot happen: invalid TV mode ${cartridge.tvMode}`);
    }
}
function config(cartridge, settings) {
    return {
        tvMode: tvMode(cartridge),
        enableAudio: cartridge.volume * settings.volume > 0,
        randomSeed: typeof cartridge.rngSeed === 'undefined' ? -1 : cartridge.rngSeed,
        emulatePaddles: cartridge.emulatePaddles,
        frameStart: typeof cartridge.firstVisibleLine === 'undefined' ? -1 : cartridge.firstVisibleLine,
        pcmAudio: (cartridge.audioEmulation || settings.audioEmulation) === "pcm",
        cpuType: cpuType(cartridge, settings)
    };
}
function mapTvEmulation(tvEmulation) {
    switch (tvEmulation) {
        case "composite":
            return "composite";
        case "svideo":
            return "svideo";
        case "none":
            return "none";
        default:
            throw new Error(`invalid TV emulation setting: ${tvEmulation}`);
    }
}
function mapScaling(scaling) {
    switch (scaling) {
        case "qis":
            return "qis";
        case "bilinear":
            return "bilinear";
        case "none":
            return "none";
        default:
            throw new Error(`invalid scaling setting`);
    }
}
function videoSettings(cartridge, settings) {
    var _a;
    return {
        gamma: settings.gammaCorrection,
        tvEmulation: mapTvEmulation(settings.tvEmulation),
        scalingMode: mapScaling(settings.scaling),
        phosphorLevel: ((_a = cartridge === null || cartridge === void 0 ? void 0 : cartridge.phosphorLevel) !== null && _a !== void 0 ? _a : settings.phosphorLevel) / 100,
        scanlineLevel: settings.scanlineIntensity / 100
    };
}
let Emulation = Emulation_1 = class Emulation {
    constructor(_storage) {
        this._storage = _storage;
        this._onEmulationStarted = ({ hash, switches }) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            try {
                yield this._emulationMutex.runExclusive(() => this._startEmulation(hash, switches));
            }
            catch (e) {
                this._ports.onEmulationStateChange_.send(error(e.message));
            }
        });
        this._onEmulationStopped = () => this._emulationMutex.runExclusive(() => this._stopEmulation());
        this._onEmulationPaused = () => this._emulationMutex.runExclusive(() => this._pauseEmulation());
        this._onEmulationResumed = () => this._emulationMutex.runExclusive(() => this._resumeEmulation());
        this._onEmulationReset = () => this._emulationMutex.runExclusive(() => this._resetEmulation());
        this._onToggleFullscreen = () => this._fullscreenDriver && this._fullscreenDriver.toggle();
        this._onSetLimitFramerate = (limitFramerate) => this._emulationService.setRateLimit(limitFramerate);
        this._onInputTogglePause = () => this._ports.onInputDriverEvent_.send("pause");
        this._onInputReset = () => this._ports.onInputDriverEvent_.send("reset");
        this._onInputToggleFullscreen = () => this._ports.onInputDriverEvent_.send("fullscreen");
        this._onWindowResize = () => this._videoDriver && this._videoDriver.resize();
        this._onUpdateConsoleSwitches = (switches) => this._emulationMutex.runExclusive(() => this._updateConsoleSwitches(switches));
        this._onMutation = (mutations) => {
            if (!mutations.some(m => m.addedNodes)) {
                return;
            }
            const canvas = document.getElementById(CANVAS_ID);
            if (canvas === this._canvas || this._handledCanvasElements.has(canvas)) {
                return;
            }
            if (canvas) {
                this._handledCanvasElements.add(canvas);
            }
            if (canvas && this._canvas) {
                this._canvasMutex.runExclusive(() => this._rebindCanvas(canvas));
            }
            else if (canvas) {
                this._canvasMutex.runExclusive(() => this._bindCanvas(canvas));
            }
            else {
                this._canvasMutex.runExclusive(() => this._unbindCanvas());
            }
        };
        this._emulationMutex = new async_mutex_1.Mutex();
        this._canvasMutex = new async_mutex_1.Mutex();
        this._driverManager = new DriverManager_1.default();
        this._currentCartridgeHash = null;
        this._canvas = null;
        this._videoDriver = null;
        this._fullscreenDriver = null;
        this._mutationObserver = new MutationObserver(this._onMutation);
        this._keyboardDriver = new KeyboardIO_1.default(document);
        this._paddleDriver = new MouseAsPaddle_1.default();
        this._audioDriver = new WebAudio_1.default();
        this._gamepadDriver = new Gamepad_1.default();
        this._touchDriver = null;
        this._currentConfig = null;
        this._handledCanvasElements = new WeakSet();
        this._emulationService = new EmulationService_1.default(WORKER_URL);
        this._emulationServiceReady = this._emulationService.init();
        this._driverManager.bind(this._emulationService);
        this._audioDriver.init();
        this._keyboardDriver.hardReset.addHandler(this._onInputReset);
        this._keyboardDriver.togglePause.addHandler(this._onInputTogglePause);
        this._keyboardDriver.toggleFullscreen.addHandler(this._onInputToggleFullscreen);
        this._driverManager.addDriver(this._gamepadDriver, (context, driver) => driver.bind([context.getJoystick(0), context.getJoystick(1)], {
            ["start"]: context.getControlPanel().getResetButton(),
            ["select"]: context.getControlPanel().getSelectSwitch()
        }));
        this._gamepadDriver.init();
        window.addEventListener('resize', this._onWindowResize);
    }
    init(ports) {
        this._ports = ports;
        ports.startEmulation_.subscribe(this._onEmulationStarted);
        ports.stopEmulation_.subscribe(this._onEmulationStopped);
        ports.pauseEmulation_.subscribe(this._onEmulationPaused);
        ports.resumeEmulation_.subscribe(this._onEmulationResumed);
        ports.resetEmulation_.subscribe(this._onEmulationReset);
        ports.toggleFullscreen_.subscribe(this._onToggleFullscreen);
        ports.setLimitFramerate_.subscribe(this._onSetLimitFramerate);
        ports.updateConsoleSwitches_.subscribe(this._onUpdateConsoleSwitches);
        this._emulationService.stateChanged.addHandler(Emulation_1._onEmulationStateChange, this);
        this._emulationService.frequencyUpdate.addHandler(Emulation_1._onFrequencyChange, this);
        this._gamepadDriver.gamepadCountChanged.addHandler(n => {
            console.log(n);
            this._ports.onUpdateGamepadCount_.send(n);
        });
        this._ports.onUpdateGamepadCount_.send(Gamepad_1.default.probeGamepadCount());
        this._mutationObserver.observe(document.body, {
            attributes: false,
            characterData: false,
            childList: true,
            subtree: true
        });
    }
    updateCartridge(cartridge) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (cartridge.hash !== this._currentCartridgeHash) {
                return;
            }
            switch (this._emulationService.getState()) {
                case EmulationServiceInterface_1.default.State.running:
                case EmulationServiceInterface_1.default.State.paused:
                    break;
                default:
                    return;
            }
            const settings = yield this._storage.getSettings();
            this._audioDriver.setMasterVolume((cartridge.volume * settings.volume) / 10000);
            if (this._videoDriver) {
                this._videoDriver.updateConfig(videoSettings(cartridge, settings));
            }
        });
    }
    updateSettings(settings) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            switch (this._emulationService.getState()) {
                case EmulationServiceInterface_1.default.State.running:
                case EmulationServiceInterface_1.default.State.paused:
                    break;
                default:
                    return;
            }
            const cartridge = yield this._storage.getCartridge(this._currentCartridgeHash);
            this._audioDriver.setMasterVolume((cartridge.volume * settings.volume) / 10000);
            if (this._videoDriver) {
                this._videoDriver.updateConfig(videoSettings(cartridge, settings));
            }
        });
    }
    static _onEmulationStateChange(state, self) {
        self._ports.onEmulationStateChange_.send(self._emulationState(state));
        if (state === EmulationServiceInterface_1.default.State.running) {
            self._audioDriver.resume();
        }
        else {
            self._audioDriver.pause();
        }
    }
    static _onFrequencyChange(frequency, self) {
        if (self._emulationService.getState() === EmulationServiceInterface_1.default.State.running) {
            self._ports.onEmulationStateChange_.send({
                state: "running",
                frequency
            });
        }
    }
    _startEmulation(hash, switches) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._emulationServiceReady;
            yield this._emulationService.stop();
            const [cartidge, image, settings] = yield Promise.all([
                this._storage.getCartridge(hash),
                this._storage.getCartridgeImage(hash),
                this._storage.getSettings()
            ]);
            if (!(cartidge && image)) {
                throw new Error(`invalid cartridge hash ${hash}`);
            }
            this._currentCartridgeHash = hash;
            const newConfig = config(cartidge, settings);
            if (!this._currentConfig || this._currentConfig.pcmAudio !== newConfig.pcmAudio) {
                this._driverManager.removeDriver(this._audioDriver);
                this._driverManager.addDriver(this._audioDriver, (context, driver) => driver.bind(newConfig.pcmAudio, newConfig.pcmAudio ? [context.getPCMChannel()] : context.getWaveformChannels()));
            }
            this._audioDriver.setMasterVolume((cartidge.volume * settings.volume) / 10000);
            if (this._videoDriver)
                this._videoDriver.updateConfig(videoSettings(cartidge, settings));
            this._currentConfig = config(cartidge, settings);
            yield this._emulationService.start(image, this._currentConfig, cartidge.cartridgeType);
            this._updateConsoleSwitches(switches);
            yield this._emulationService.resume();
        });
    }
    _stopEmulation() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._emulationServiceReady;
            yield this._emulationService.stop();
            this._currentCartridgeHash = null;
        });
    }
    _pauseEmulation() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._emulationServiceReady;
            yield this._emulationService.pause();
        });
    }
    _resumeEmulation() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._emulationServiceReady;
            if (this._emulationService.getState() === EmulationServiceInterface_1.default.State.paused) {
                yield this._emulationService.resume();
            }
        });
    }
    _resetEmulation() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._emulationServiceReady;
            yield this._emulationService.reset();
        });
    }
    _updateConsoleSwitches(consoleSwitches) {
        const context = this._emulationService.getEmulationContext();
        if (!context) {
            return;
        }
        const controlPanel = context.getControlPanel();
        controlPanel.getDifficultySwitchP0().toggle(consoleSwitches.difficultyP0 === "pro");
        controlPanel.getDifficultySwitchP1().toggle(consoleSwitches.difficultyP1 === "pro");
        controlPanel.getColorSwitch().toggle(consoleSwitches.color === "bw");
    }
    _createAndBindVideoDriver(canvas) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._emulationServiceReady;
            this._removeVideoDriver();
            const [settings, cartridge] = yield this._emulationMutex.runExclusive(() => Promise.all([
                this._storage.getSettings(),
                this._currentCartridgeHash && this._storage.getCartridge(this._currentCartridgeHash)
            ]));
            this._canvas = canvas;
            this._videoDriver = new Video_1.default(this._canvas, videoSettings(cartridge, settings));
            this._videoDriver.init();
            this._driverManager.addDriver(this._videoDriver, (context, driver) => driver.bind(context.getVideo()));
            this._fullscreenDriver = new FullscreenVideo_1.default(this._videoDriver);
        });
    }
    _bindCanvas(canvas) {
        var _a;
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const settings = yield this._storage.getSettings();
            yield this._createAndBindVideoDriver(canvas);
            if ((_a = settings.touchControls) !== null && _a !== void 0 ? _a : TouchIO_1.default.isSupported()) {
                this._touchDriver = new TouchIO_1.default(canvas, settings.virtualJoystickSensitivity, settings.leftHanded);
                this._touchDriver.toggleFullscreen.addHandler(this._onInputToggleFullscreen);
                this._touchDriver.togglePause.addHandler(this._onInputTogglePause);
                this._driverManager.addDriver(this._touchDriver, (context, driver) => driver.bind(context.getJoystick(0), context.getControlPanel()));
            }
            this._driverManager
                .addDriver(this._keyboardDriver, (context, driver) => driver.bind(context.getJoystick(0), context.getJoystick(1), context.getControlPanel()))
                .addDriver(this._paddleDriver, (context, driver) => driver.bind(context.getPaddle(0)));
        });
    }
    _rebindCanvas(canvas) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._createAndBindVideoDriver(canvas);
        });
    }
    _unbindCanvas() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this._touchDriver) {
                this._driverManager.removeDriver(this._touchDriver);
                this._touchDriver = null;
            }
            this._driverManager.removeDriver(this._keyboardDriver).removeDriver(this._paddleDriver);
            yield this._removeVideoDriver();
            this._handledCanvasElements.delete(this._canvas);
            this._canvas = null;
        });
    }
    _removeVideoDriver() {
        if (!this._videoDriver) {
            return;
        }
        this._driverManager.removeDriver(this._videoDriver);
        this._videoDriver.close();
        this._videoDriver = null;
        this._fullscreenDriver.disengage();
        this._fullscreenDriver = null;
    }
    _emulationState(stateIncoming) {
        switch (stateIncoming) {
            case EmulationServiceInterface_1.default.State.error:
                return error(this._emulationService.getLastError().message);
            case EmulationServiceInterface_1.default.State.running:
                return { state: "running" };
            case EmulationServiceInterface_1.default.State.paused:
                return { state: "paused" };
            case EmulationServiceInterface_1.default.State.stopped:
                return { state: "stopped" };
            default:
                throw new Error(`cannot happen: invalid emulation state ${stateIncoming}`);
        }
    }
};
Emulation = Emulation_1 = tslib_1.__decorate([
    inversify_1.injectable(),
    tslib_1.__param(0, inversify_1.inject(Storage_1.default)),
    tslib_1.__metadata("design:paramtypes", [Storage_1.default])
], Emulation);
exports.default = Emulation;
//# sourceMappingURL=Emulation.js.map