"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const dexie_1 = tslib_1.__importDefault(require("dexie"));
const inversify_1 = require("inversify");
const SETTINGS_ID = 0;
const DB_NAME = process.env.DEVELOPMENT ? 'stellerator-ng-dev' : 'stellerator-ng';
class Database extends dexie_1.default {
    constructor() {
        super(DB_NAME);
        this.version(1).stores({
            cartridges: '&hash',
            roms: '&hash',
            settings: '&id'
        });
        this.version(2)
            .stores({
            cartridges: '&hash',
            roms: '&hash',
            settings: '&id'
        })
            .upgrade(transaction => {
            transaction
                .table('settings')
                .toCollection()
                .modify((settings) => {
                const _settings = settings;
                delete _settings.smoothScaling;
                delete _settings.phosphorEmulation;
                delete _settings.videoSync;
                settings.tvEmulation = "composite";
                settings.scaling = "qis";
                settings.phosphorLevel = 50;
                settings.scanlineIntensity = 20;
            });
            transaction
                .table('cartridges')
                .toCollection()
                .modify(cartridge => {
                delete cartridge.phosphorEmulation;
            });
        });
    }
}
let Storage = class Storage {
    constructor() {
        this._database = new Database();
        this._defaultSettings = null;
    }
    setDefaults(settings) {
        this._defaultSettings = settings;
    }
    getAllCartridges() {
        return this._database.cartridges.toArray();
    }
    insertCartridges(cartridges) {
        return this._database.transaction('rw', [this._database.cartridges, this._database.roms], () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield Promise.all([
                this._database.cartridges.bulkPut(cartridges.map(c => c.cartridge)),
                this._database.roms.bulkPut(cartridges.map(c => ({ hash: c.cartridge.hash, image: c.image })))
            ]);
        }));
    }
    updateCartridge(cartridge) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this._database.transaction('rw', [this._database.cartridges], () => tslib_1.__awaiter(this, void 0, void 0, function* () {
                if (!(yield this._database.cartridges.get(cartridge.hash))) {
                    return;
                }
                yield this._database.cartridges.put(cartridge);
            }));
        });
    }
    deleteCartridge(hash) {
        return this._database.transaction('rw', [this._database.cartridges, this._database.roms], () => tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield Promise.all([this._database.cartridges.delete(hash), this._database.roms.delete(hash)]);
        }));
    }
    deleteAllCartridges() {
        return this._database.cartridges.clear();
    }
    getCartridge(hash) {
        return this._database.cartridges.get(hash);
    }
    getCartridgeImage(hash) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const rom = yield this._database.roms.get(hash);
            return rom && rom.image;
        });
    }
    getSettings() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const record = yield this._database.settings.get(SETTINGS_ID);
            if (!record) {
                return this._defaultSettings;
            }
            const { id } = record, settings = tslib_1.__rest(record, ["id"]);
            return settings;
        });
    }
    saveSettings(settings) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._database.settings.put(Object.assign(Object.assign({}, settings), { id: SETTINGS_ID }));
        });
    }
    dropDatabase() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._database.close();
            yield indexedDB.deleteDatabase(DB_NAME);
            this._database = new Database();
        });
    }
};
Storage = tslib_1.__decorate([
    inversify_1.injectable()
], Storage);
exports.default = Storage;
//# sourceMappingURL=Storage.js.map