"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const Instruction_1 = tslib_1.__importDefault(require("./cpu/Instruction"));
const Disassembler_1 = tslib_1.__importDefault(require("./cpu/Disassembler"));
const hex = tslib_1.__importStar(require("../tools/hex"));
const binary = tslib_1.__importStar(require("../tools/binary"));
const util = tslib_1.__importStar(require("util"));
class Debugger {
    constructor(_traceSize = 2048, _stepMaxCycles = 10000000) {
        this._traceSize = _traceSize;
        this._stepMaxCycles = _stepMaxCycles;
        this._breakpoints = new Uint8Array(0x10000);
        this._breakpointDescriptions = new Array(0x10000);
        this._breakpointsEnabled = false;
        this._traceEnabled = false;
        this._trace = new Uint16Array(this._traceSize);
        this._traceLength = 0;
        this._traceIndex = 0;
    }
    attach(board) {
        this._board = board;
        this._bus = this._board.getBus();
        this._cpu = this._board.getCpu();
        this._disassembler = new Disassembler_1.default(this._bus);
        this._board.trap.addHandler(this._trapHandler, this);
        this._traceLength = 0;
        this._traceIndex = 0;
        return this;
    }
    detach() {
        if (!this._board) {
            return this;
        }
        this._board.cpuClock.removeHandler(this._cpuClockHandler);
        this._board.trap.removeHandler(this._trapHandler);
        return this;
    }
    clearAllBreakpoints() {
        for (let i = 0; i < 0x10000; i++) {
            this._breakpoints[i] = 0;
        }
        return this;
    }
    setBreakpoint(address, description = '-') {
        address %= 0x10000;
        this._breakpoints[address] = 1;
        this._breakpointDescriptions[address] = description;
        return this;
    }
    clearBreakpoint(address) {
        this._breakpoints[address % 0x10000] = 0;
        return this;
    }
    dumpBreakpoints() {
        let result = '';
        for (let address = 0; address < 0x10000; address++) {
            if (this._breakpoints[address]) {
                result += hex.encode(address, 4) + ': ' + this._breakpointDescriptions[address] + '\n';
            }
        }
        return result.replace(/\n$/, '');
    }
    loadBlock(block, at, from = 0, to = block.length - 1) {
        for (let i = 0; i <= to - from; i++) {
            this._poke(at + i, block[i]);
        }
    }
    disassembleAt(start, length) {
        let i = 0, result = '', instruction, address;
        while (i < length) {
            address = (start + i) % 0x10000;
            instruction = Instruction_1.default.opcodes[this._peek(address)];
            result +=
                (this._breakpoints[address] ? '(B) ' : '    ') +
                    hex.encode(address, 4) +
                    ':   ' +
                    this._disassembler.disassembleAt(address) +
                    '\n';
            i += instruction.getSize();
        }
        return result.replace(/\n$/, '');
    }
    disassemble(length) {
        return this.disassembleAt(this._cpu.state.p, length);
    }
    trace(length = this._traceSize) {
        let result = '';
        length = Math.min(length, this._traceLength);
        for (let i = 0; i < length; i++) {
            result +=
                this.disassembleAt(this._trace[(this._traceSize + this._traceIndex - length + i) % this._traceSize], 1) + '\n';
        }
        return result + this.disassemble(1);
    }
    dumpAt(start, length) {
        let result = '', address;
        for (let i = 0; i < length; i++) {
            address = (start + i) % 0x10000;
            result += hex.encode(address, 4) + ':   ' + hex.encode(this._peek(address), 2) + '\n';
        }
        return result.replace(/\n$/, '');
    }
    dumpState() {
        const state = this._cpu.state;
        switch (this._cpu.executionState) {
            case 0:
        }
        let result = '' +
            'A = ' +
            hex.encode(state.a, 2) +
            '   ' +
            'X = ' +
            hex.encode(state.x, 2) +
            '   ' +
            'Y = ' +
            hex.encode(state.y, 2) +
            '   ' +
            'S = ' +
            hex.encode(state.s, 2) +
            '   ' +
            'P = ' +
            hex.encode(state.p, 4) +
            '\n' +
            'flags = ' +
            binary.encode(state.flags, 8) +
            '\n' +
            'state: ' +
            this._humanReadableExecutionState();
        const boardState = this._board.getBoardStateDebug();
        if (boardState) {
            result += '\n' + '\n' + boardState;
        }
        return result;
    }
    dumpStack() {
        return this.dumpAt(0x0100 + this._cpu.state.s, 0x100 - this._cpu.state.s);
    }
    step(instructions) {
        let instruction = 0, cycles = 0, lastExecutionState = this._cpu.executionState, cpuCycles = 0;
        const timer = this._board.getTimer();
        const cpuClockHandler = (c) => (cpuCycles += c);
        this._board.cpuClock.addHandler(cpuClockHandler);
        this._lastTrap = undefined;
        this._board.resume();
        while (instruction < instructions && !this._lastTrap && cycles < this._stepMaxCycles) {
            timer.tick(1);
            cycles++;
            if (lastExecutionState !== this._cpu.executionState) {
                lastExecutionState = this._cpu.executionState;
                if (lastExecutionState === 1) {
                    instruction++;
                }
            }
        }
        this._board.cpuClock.removeHandler(cpuClockHandler);
        this._board.suspend();
        return { cycles, cpuCycles };
    }
    stepClock(cycles) {
        this._lastTrap = undefined;
        this._board.resume();
        const usedCycles = this._board.getTimer().tick(cycles);
        this._board.suspend();
        return usedCycles;
    }
    setBreakpointsEnabled(breakpointsEnabled) {
        this._breakpointsEnabled = breakpointsEnabled;
        this._attachToCpuIfNecessary();
        return this;
    }
    setTraceEnabled(traceEnabled) {
        this._traceEnabled = traceEnabled;
        this._attachToCpuIfNecessary();
        return this;
    }
    getBoard() {
        return this._board;
    }
    getLastTrap() {
        return this._lastTrap;
    }
    _humanReadableExecutionState() {
        if (this._cpu.isHalt()) {
            return 'halted';
        }
        switch (this._cpu.executionState) {
            case 0:
                return 'boot';
            case 1:
                return 'fetch';
            case 2:
                return 'execute';
        }
    }
    _attachToCpuIfNecessary() {
        if (this._traceEnabled || this._breakpointsEnabled) {
            this._lastInstructionPointer = this._cpu.getLastInstructionPointer() || 0;
            this._board.cpuClock.addHandler(this._cpuClockHandler, this);
            this._board.setClockMode(0);
        }
        else {
            this._board.cpuClock.removeHandler(this._cpuClockHandler, this);
            this._board.setClockMode(1);
        }
    }
    _cpuClockHandler(clocks, ctx) {
        const lastInstructionPointer = ctx._cpu.getLastInstructionPointer();
        if (ctx._cpu.executionState !== 1 ||
            lastInstructionPointer === ctx._lastInstructionPointer) {
            return;
        }
        ctx._lastInstructionPointer = lastInstructionPointer;
        if (ctx._traceEnabled) {
            ctx._trace[ctx._traceIndex] = lastInstructionPointer;
            ctx._traceIndex = (ctx._traceIndex + 1) % ctx._traceSize;
            if (ctx._traceLength < ctx._traceSize) {
                ctx._traceLength++;
            }
        }
        if (ctx._breakpointsEnabled && ctx._breakpoints[ctx._cpu.state.p]) {
            ctx._board.triggerTrap(2, util.format('breakpoint "%s" at %s', ctx._breakpointDescriptions[ctx._cpu.state.p] || '', hex.encode(ctx._cpu.state.p)));
        }
    }
    _trapHandler(trap, dbg) {
        dbg._lastTrap = trap;
    }
    _peek(address) {
        return this._bus.peek(address % 0x10000);
    }
    _poke(address, value) {
        this._bus.poke(address % 0x10000, value & 0xff);
    }
}
exports.default = Debugger;
//# sourceMappingURL=Debugger.js.map