"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const CpuInterface_1 = tslib_1.__importDefault(require("./CpuInterface"));
const vector_1 = require("./statemachine/vector");
const Compiler_1 = tslib_1.__importDefault(require("./statemachine/Compiler"));
class StateMachineCpu {
    constructor(_bus, _rng) {
        this._bus = _bus;
        this._rng = _rng;
        this.executionState = 0;
        this.state = new CpuInterface_1.default.State();
        this._invalidInstructionCallback = null;
        this._interruptPending = false;
        this._nmiPending = false;
        this._halt = false;
        this._pollInterruptsAfterLastInstruction = false;
        this._lastInstructionPointer = 0;
        this._operations = new Array(255);
        this._opBoot = vector_1.boot(this.state);
        this._opIrq = vector_1.irq(this.state);
        this._opNmi = vector_1.nmi(this.state);
        const compiler = new Compiler_1.default(this.state);
        for (let op = 0; op < 256; op++) {
            this._operations[op] = compiler.compile(op);
        }
        this.reset();
    }
    reset() {
        this.state.a = this._rng ? this._rng.int(0xff) : 0;
        this.state.x = this._rng ? this._rng.int(0xff) : 0;
        this.state.y = this._rng ? this._rng.int(0xff) : 0;
        this.state.s = 0xfd;
        this.state.p = this._rng ? this._rng.int(0xffff) : 0;
        this.state.flags =
            (this._rng ? this._rng.int(0xff) : 0) | 4 | 32 | 16;
        this.state.irq = false;
        this.state.nmi = false;
        this.executionState = 0;
        this._interruptPending = false;
        this._nmiPending = false;
        this._halt = false;
        this._lastResult = this._opBoot.reset(undefined);
        this._lastInstructionPointer = 0;
        return this;
    }
    setInterrupt(i) {
        this._interruptPending = i;
        return this;
    }
    isInterrupt() {
        return this._interruptPending;
    }
    nmi() {
        this._nmiPending = true;
        return this;
    }
    halt() {
        this._halt = true;
        return this;
    }
    resume() {
        this._halt = false;
        return this;
    }
    isHalt() {
        return this._halt;
    }
    setInvalidInstructionCallback(callback) {
        this._invalidInstructionCallback = callback;
        return this;
    }
    getInvalidInstructionCallback() {
        return this._invalidInstructionCallback;
    }
    getLastInstructionPointer() {
        return this._lastInstructionPointer;
    }
    cycle() {
        if (this._halt && (!this._lastResult || this._lastResult.cycleType === 0)) {
            return this;
        }
        if (this.executionState === 1) {
            this._fetch();
            return this;
        }
        let value;
        switch (this._lastResult.cycleType) {
            case 0:
                value = this._bus.read(this._lastResult.address);
                break;
            case 1:
                value = this._lastResult.value;
                this._bus.write(this._lastResult.address, value);
                break;
            default:
                throw new Error('invalid cycle type');
        }
        if (this._lastResult.pollInterrupts) {
            this._pollInterrupts();
            this._lastResult.pollInterrupts = false;
            this._pollInterruptsAfterLastInstruction = false;
        }
        this._lastResult = this._lastResult.nextStep(value);
        if (this._lastResult === null) {
            this.executionState = 1;
        }
        return this;
    }
    _fetch() {
        if (this._pollInterruptsAfterLastInstruction) {
            this._pollInterrupts();
        }
        this._lastInstructionPointer = this.state.p;
        let operation;
        const opcode = this._bus.read(this.state.p);
        if (this.state.nmi) {
            operation = this._opNmi;
            this._pollInterruptsAfterLastInstruction = false;
        }
        else if (this.state.irq) {
            operation = this._opIrq;
            this._pollInterruptsAfterLastInstruction = false;
        }
        else {
            operation = this._operations[opcode];
            this.state.p = (this.state.p + 1) & 0xffff;
            this._pollInterruptsAfterLastInstruction = true;
        }
        if (!operation) {
            if (this._invalidInstructionCallback) {
                this._invalidInstructionCallback(this);
            }
            return;
        }
        this.executionState = 2;
        this._lastResult = operation.reset(undefined);
    }
    _pollInterrupts() {
        this.state.irq = false;
        if (this._nmiPending) {
            this.state.nmi = true;
            this._nmiPending = false;
            return;
        }
        if (this._interruptPending && !this.state.nmi && !(this.state.flags & 4)) {
            this.state.irq = true;
        }
    }
}
exports.default = StateMachineCpu;
//# sourceMappingURL=StateMachineCpu.js.map