"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const microevent_ts_1 = require("microevent.ts");
const BoardInterface_1 = tslib_1.__importDefault(require("../board/BoardInterface"));
const Bus_1 = tslib_1.__importDefault(require("./Bus"));
const Pia_1 = tslib_1.__importDefault(require("./Pia"));
const Tia_1 = tslib_1.__importDefault(require("./tia/Tia"));
const Config_1 = tslib_1.__importDefault(require("./Config"));
const ControlPanel_1 = tslib_1.__importDefault(require("./ControlPanel"));
const DigitalJoystick_1 = tslib_1.__importDefault(require("../io/DigitalJoystick"));
const Paddle_1 = tslib_1.__importDefault(require("../io/Paddle"));
const factory_1 = require("../../tools/rng/factory");
const Factory_1 = tslib_1.__importDefault(require("../cpu/Factory"));
class Board {
    constructor(_config, cartridge, cpuFactory) {
        this._config = _config;
        this.trap = new microevent_ts_1.Event();
        this.clock = new microevent_ts_1.Event();
        this.cpuClock = new microevent_ts_1.Event();
        this.systemReset = new microevent_ts_1.Event();
        this._clockMode = 1;
        this._cpuCycles = 0;
        this._trap = false;
        this._audioEnabled = true;
        this._suspended = true;
        this._subClock = 0;
        this._clockHz = 0;
        this._sliceSize = 0;
        this._timer = {
            tick: (clocks) => this.tick(clocks),
            start: (scheduler) => this._start(scheduler),
            stop: () => this._stop(),
            isRunning: () => !!this._runTask,
        };
        this._rng = factory_1.createRng(_config.randomSeed < 0 ? Math.random() : _config.randomSeed);
        cartridge.randomize(this._rng);
        const bus = new Bus_1.default();
        if (typeof cpuFactory === 'undefined') {
            cpuFactory = (_bus, rng) => new Factory_1.default(_config.cpuType).create(_bus, rng);
        }
        const controlPanel = new ControlPanel_1.default(), joystick0 = new DigitalJoystick_1.default(), joystick1 = new DigitalJoystick_1.default(), paddles = new Array(4);
        for (let i = 0; i < 4; i++) {
            paddles[i] = new Paddle_1.default();
        }
        const cpu = cpuFactory(bus, this._rng);
        const pia = new Pia_1.default(controlPanel, joystick0, joystick1, this._rng);
        const tia = new Tia_1.default(_config, joystick0, joystick1, paddles);
        cpu.setInvalidInstructionCallback(() => this._onInvalidInstruction());
        tia.setCpu(cpu)
            .setBus(bus)
            .setCpuTimeProvider(() => this.getCpuTime());
        cartridge
            .setCpu(cpu)
            .setBus(bus)
            .setCpuTimeProvider(() => this.getCpuTime())
            .setRng(this._rng);
        pia.setBus(bus);
        bus.setTia(tia).setPia(pia).setCartridge(cartridge);
        this._bus = bus;
        this._cpu = cpu;
        this._tia = tia;
        this._pia = pia;
        this._cartridge = cartridge;
        this._controlPanel = controlPanel;
        this._joystick0 = joystick0;
        this._joystick1 = joystick1;
        this._paddles = paddles;
        this._bus.event.trap.addHandler((payload) => this.triggerTrap(1, payload.message));
        this._clockHz = Config_1.default.getClockHz(_config);
        this._sliceSize = 228 * (_config.tvMode === 0 ? 262 : 312);
        this.reset();
    }
    getCpu() {
        return this._cpu;
    }
    getBus() {
        return this._bus;
    }
    getVideoOutput() {
        return this._tia;
    }
    getWaveformChannels() {
        return [0, 1].map((i) => this._tia.getWaveformChannel(i));
    }
    getPCMChannel() {
        return this._tia.getPCMChannel();
    }
    getTimer() {
        return this._timer;
    }
    getConfig() {
        return this._config;
    }
    reset() {
        this._cpu.reset();
        this._tia.reset();
        this._pia.reset();
        this._cartridge.reset();
        this._controlPanel.getResetButton().toggle(false);
        this._controlPanel.getSelectSwitch().toggle(false);
        this._controlPanel.getColorSwitch().toggle(false);
        this._controlPanel.getDifficultySwitchP0().toggle(true);
        this._controlPanel.getDifficultySwitchP1().toggle(true);
        this._subClock = 0;
        this._cpuCycles = 0;
        this.systemReset.dispatch();
        return this;
    }
    boot() {
        let cycles = 0, cpuCycles = 0;
        this.reset();
        if (this._cpu.executionState !== 0) {
            throw new Error('Already booted!');
        }
        while (this._cpu.executionState !== 1) {
            this._cycle();
            cycles++;
            if (this._subClock === 0) {
                cpuCycles++;
            }
        }
        this.cpuClock.dispatch(cpuCycles);
        this.clock.dispatch(cycles);
        return this;
    }
    suspend() {
        this._suspended = true;
        this._updateAudioState();
    }
    resume() {
        this._suspended = false;
        this._updateAudioState();
    }
    setAudioEnabled(state) {
        this._audioEnabled = state;
        this._updateAudioState();
    }
    triggerTrap(reason, message) {
        this._stop();
        this._trap = true;
        if (this.trap.hasHandlers) {
            this.trap.dispatch(new BoardInterface_1.default.TrapPayload(reason, this, message));
        }
        else {
            throw new Error(message);
        }
        return this;
    }
    getControlPanel() {
        return this._controlPanel;
    }
    getJoystick0() {
        return this._joystick0;
    }
    getJoystick1() {
        return this._joystick1;
    }
    getBoardStateDebug() {
        const sep = '============';
        return ('TIA:\n' +
            sep +
            '\n' +
            this._tia.getDebugState() +
            '\n' +
            `\n` +
            `PIA:\n` +
            `${sep}\n` +
            `${this._pia.getDebugState()}\n`);
    }
    setClockMode(clockMode) {
        this._clockMode = clockMode;
        return this;
    }
    getClockMode() {
        return this._clockMode;
    }
    getPaddle(idx) {
        return this._paddles[idx];
    }
    getCpuTime() {
        return (this._cpuCycles / Config_1.default.getClockHz(this._config)) * 3;
    }
    tick(requestedCycles) {
        let i = 0, cycles = 0, cpuCycles = 0, lastExecutionState = this._cpu.executionState;
        this._trap = false;
        while (i++ < requestedCycles && !this._trap) {
            this._cycle();
            cycles++;
            if (this._subClock === 0) {
                cpuCycles++;
                this._cpuCycles++;
            }
            if (lastExecutionState !== this._cpu.executionState) {
                lastExecutionState = this._cpu.executionState;
                if (this._cpu.executionState === 1) {
                    this._cartridge.notifyCpuCycleComplete();
                    if (this._clockMode === 0 &&
                        cpuCycles > 0 &&
                        this.cpuClock.hasHandlers) {
                        this.cpuClock.dispatch(cpuCycles);
                        cpuCycles = 0;
                    }
                }
            }
        }
        if (cpuCycles > 0 && this.cpuClock.hasHandlers) {
            this.cpuClock.dispatch(cpuCycles);
        }
        if (cycles > 0 && this.clock.hasHandlers) {
            this.clock.dispatch(cycles);
        }
        return cycles;
    }
    getSubclock() {
        return this._subClock;
    }
    static _executeSlice(board, _timeSlice) {
        const slice = _timeSlice ? Math.round((_timeSlice * board._clockHz) / 1000) : board._sliceSize;
        return (board.tick(slice) / board._clockHz) * 1000;
    }
    _updateAudioState() {
        this._tia.setAudioEnabled(this._audioEnabled && !this._suspended);
    }
    _cycle() {
        this._tia.cycle();
        if (this._subClock++ >= 2) {
            this._pia.cycle();
            this._cpu.cycle();
            this._subClock = 0;
        }
    }
    _start(scheduler) {
        if (this._runTask) {
            return;
        }
        this._runTask = scheduler.start(Board._executeSlice, this, 1000 / (this._config.tvMode === 0 ? 60 : 50));
    }
    _stop() {
        if (!this._runTask) {
            return;
        }
        this._runTask.stop();
        this._runTask = undefined;
    }
    _onInvalidInstruction() {
        this.triggerTrap(0, 'invalid instruction');
    }
}
exports.default = Board;
//# sourceMappingURL=Board.js.map