"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const microevent_ts_1 = require("microevent.ts");
class Pia {
    constructor(_controlPanel, _joystick0, _joystick1, _rng) {
        this._controlPanel = _controlPanel;
        this._joystick0 = _joystick0;
        this._joystick1 = _joystick1;
        this._rng = _rng;
        this.trap = new microevent_ts_1.Event();
        this.ram = new Uint8Array(128);
        this._bus = null;
        this._timerValue = 255;
        this._subTimer = 0;
        this._timerDivide = 1024;
        this._interruptFlag = 0;
        this._timerWrapped = false;
        this._flagSetDuringThisCycle = false;
        this.reset();
    }
    reset() {
        for (let i = 0; i < 128; i++) {
            this.ram[i] = this._rng ? this._rng.int(0xff) : 0;
        }
        this._interruptFlag = 0;
        this._flagSetDuringThisCycle = false;
        this._timerDivide = 1024;
        this._subTimer = 0;
        this._rng.int(0xff);
        this._timerValue = 0;
        this._timerWrapped = false;
    }
    read(address) {
        if (address & 0x0200) {
            if (address & 0x0004) {
                return this._readTimer(address);
            }
            else {
                return this._readIo(address);
            }
        }
        else {
            return this.ram[address & 0x7f];
        }
    }
    peek(address) {
        if (address & 0x0200) {
            if (address & 0x0004) {
                return this._peekTimer(address);
            }
            else {
                return this._readIo(address);
            }
        }
        else {
            return this.ram[address & 0x7f];
        }
    }
    write(address, value) {
        if (address & 0x0200) {
            if (address & 0x0004) {
                return this._writeTimer(address, value);
            }
            else {
                return this._writeIo(address, value);
            }
        }
        else {
            this.ram[address & 0x7f] = value;
        }
    }
    cycle() {
        this._cycleTimer();
    }
    getDebugState() {
        return `divider: ${this._timerDivide} raw timer: INTIM: ${this._timerValue}`;
    }
    setBus(bus) {
        this._bus = bus;
        return this;
    }
    _writeIo(address, value) { }
    _writeTimer(address, value) {
        this._interruptFlag = 0;
        switch (address & 0x0297) {
            case 663:
                return this._setTimer(1024, value);
            case 662:
                return this._setTimer(64, value);
            case 661:
                return this._setTimer(8, value);
            case 660:
                return this._setTimer(1, value);
        }
    }
    _setTimer(divide, value) {
        this._timerDivide = divide;
        this._timerValue = value;
        this._subTimer = 0;
        this._timerWrapped = false;
    }
    _readIo(address) {
        switch (address & 0x0283) {
            case 640:
                return ((this._joystick1.getUp().read() ? 0 : 0x01) |
                    (this._joystick1.getDown().read() ? 0 : 0x02) |
                    (this._joystick1.getLeft().read() ? 0 : 0x04) |
                    (this._joystick1.getRight().read() ? 0 : 0x08) |
                    (this._joystick0.getUp().read() ? 0 : 0x10) |
                    (this._joystick0.getDown().read() ? 0 : 0x20) |
                    (this._joystick0.getLeft().read() ? 0 : 0x40) |
                    (this._joystick0.getRight().read() ? 0 : 0x80));
            case 642:
                return ((this._controlPanel.getResetButton().read() ? 0 : 0x01) |
                    (this._controlPanel.getSelectSwitch().read() ? 0 : 0x02) |
                    (this._controlPanel.getColorSwitch().read() ? 0 : 0x08) |
                    (this._controlPanel.getDifficultySwitchP0().read() ? 0 : 0x40) |
                    (this._controlPanel.getDifficultySwitchP1().read() ? 0 : 0x80));
        }
        return this._bus.getLastDataBusValue();
    }
    _readTimer(address) {
        if (address & 0x01) {
            const flag = this._interruptFlag;
            return flag & 0x80;
        }
        else {
            if (!this._flagSetDuringThisCycle) {
                this._interruptFlag = 0;
                this._timerWrapped = false;
            }
            return this._timerValue;
        }
    }
    _peekTimer(address) {
        return address & 0x01 ? this._interruptFlag & 0x80 : this._timerValue;
    }
    _cycleTimer() {
        this._flagSetDuringThisCycle = false;
        if (this._timerWrapped) {
            this._timerValue = (this._timerValue + 0xff) & 0xff;
        }
        else if (this._subTimer === 0 && --this._timerValue < 0) {
            this._timerValue = 0xff;
            this._flagSetDuringThisCycle = true;
            this._interruptFlag = 0xff;
            this._timerWrapped = true;
        }
        if (++this._subTimer === this._timerDivide) {
            this._subTimer = 0;
        }
    }
}
exports.default = Pia;
(function (Pia) {
    class TrapPayload {
        constructor(reason, pia, message) {
            this.reason = reason;
            this.pia = pia;
            this.message = message;
        }
    }
    Pia.TrapPayload = TrapPayload;
})(Pia || (Pia = {}));
exports.default = Pia;
//# sourceMappingURL=Pia.js.map