"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const cartridgeUtil = tslib_1.__importStar(require("./util"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
class Cartridge3E extends AbstractCartridge_1.default {
    constructor(buffer) {
        super();
        this._banks = null;
        this._ramSelect = false;
        this._ramBanks = new Array(0x0100);
        this._bus = null;
        if ((buffer.length & 0x07ff) !== 0) {
            throw new Error(`buffer length ${buffer.length} is not a multiple of 2k`);
        }
        const bankCount = buffer.length >>> 11;
        if (bankCount < 2) {
            throw new Error('image must have at least 2k');
        }
        this._banks = new Array(bankCount);
        for (let i = 0; i < bankCount; i++) {
            this._banks[i] = new Uint8Array(0x0800);
        }
        for (let i = 0; i <= 0xff; i++) {
            this._ramBanks[i] = new Uint8Array(0x0400);
        }
        this._ramBank = this._ramBanks[0];
        this._bank1 = this._banks[bankCount - 1];
        this._bank0 = this._banks[0];
        for (let i = 0; i < 0x0800; i++) {
            for (let j = 0; j < bankCount; j++) {
                this._banks[j][i] = buffer[0x0800 * j + i];
            }
        }
    }
    static matchesBuffer(buffer) {
        const signatureCounts = cartridgeUtil.searchForSignatures(buffer, [[0x85, 0x3e, 0xa9, 0x00]]);
        return signatureCounts[0] >= 1;
    }
    reset() {
        this._bank0 = this._banks[0];
    }
    randomize(rng) {
        for (let i = 0; i < this._ramBanks.length; i++) {
            for (let j = 0; j < 0x0400; j++) {
                this._ramBanks[i][j] = rng.int(0xff);
            }
        }
    }
    setBus(bus) {
        this._bus = bus;
        this._bus.event.read.addHandler(Cartridge3E._onBusAccess, this);
        this._bus.event.write.addHandler(Cartridge3E._onBusAccess, this);
        return this;
    }
    read(address) {
        address &= 0x0fff;
        if (this._ramSelect) {
            if (address < 0x0400) {
                return this._ramBank[address];
            }
            if (address < 0x0800) {
                return (this._ramBank[address & 0x03ff] = this._bus.getLastDataBusValue());
            }
            return this._bank1[address & 0x07ff];
        }
        return address < 0x0800 ? this._bank0[address] : this._bank1[address & 0x07ff];
    }
    peek(address) {
        address &= 0x0fff;
        if (this._ramSelect) {
            if (address < 0x0400) {
                return this._ramBank[address];
            }
            if (address < 0x0800) {
                return this._bus.getLastDataBusValue();
            }
            return this._bank1[address & 0x07ff];
        }
        return address < 0x0800 ? this._bank0[address] : this._bank1[address & 0x07ff];
    }
    write(address, value) {
        if (!this._ramSelect) {
            return;
        }
        address &= 0x0fff;
        if (address >= 0x0400 && address < 0x0800) {
            this._ramBank[address & 0x03ff] = value;
        }
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_3E;
    }
    static _onBusAccess(accessType, self) {
        switch (self._bus.getLastAddresBusValue()) {
            case 0x003f:
                self._ramSelect = false;
                self._bank0 = self._banks[self._bus.getLastDataBusValue() % self._banks.length];
                break;
            case 0x003e:
                self._ramSelect = true;
                self._ramBank = self._ramBanks[self._bus.getLastDataBusValue() % 32];
                break;
        }
    }
}
exports.default = Cartridge3E;
//# sourceMappingURL=Cartridge3E.js.map