"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const Soc_1 = tslib_1.__importDefault(require("./harmony/Soc"));
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
const cartridgeUtil = tslib_1.__importStar(require("./util"));
class CartridgeCDF extends AbstractCartridge_1.default {
    constructor(buffer) {
        super();
        this._handleBxCDF0 = (address) => {
            const thumbulator = this._soc.getThumbulator(), r2 = thumbulator.readRegister(2), r3 = thumbulator.readRegister(3);
            switch (address) {
                case 0x000006e2:
                    this._musicStreams[r2 % 3].frequency = r3;
                    return 0;
                case 0x000006e6:
                    this._musicStreams[r2 % 3].counter = 0;
                    return 0;
                case 0x000006ea:
                    thumbulator.writeRegister(2, this._musicStreams[r2 % 3].counter);
                    return 0;
                case 0x000006ee:
                    this._musicStreams[r2 % 3].waveformSize = r3;
                    return 0;
            }
            return 3;
        };
        this._handleBxCDF1 = (address) => {
            const thumbulator = this._soc.getThumbulator(), r2 = thumbulator.readRegister(2), r3 = thumbulator.readRegister(3);
            switch (address) {
                case 0x00000752:
                    this._musicStreams[r2 % 3].frequency = r3;
                    return 0;
                case 0x00000756:
                    this._musicStreams[r2 % 3].counter = 0;
                    return 0;
                case 0x0000075a:
                    thumbulator.writeRegister(2, this._musicStreams[r2 % 3].counter);
                    return 0;
                case 0x0000075e:
                    this._musicStreams[r2 % 3].waveformSize = r3;
                    return 0;
            }
            return 3;
        };
        this._banks = new Array(7);
        this._currentBank = null;
        this._rom = null;
        this._ram = null;
        this._displayRam = null;
        this._musicStreams = new Array(3);
        this._clockAccumulator = 0;
        this._lastCpuTime = 0;
        this._soc = null;
        this._fastFetch = false;
        this._digitalAudio = false;
        this._fastJumpCountdown = 0;
        this._fastFetchPending = false;
        this._jmpOperandAddress = 0;
        this._ldaOperandAddress = 0;
        this._datastreamBase = 0;
        this._datastreamIncrementBase = 0;
        this._waveformBase = 0;
        this._jumpstream = 0;
        this._jumpstreamMask = 0;
        this._amplitudeStream = 0;
        this._bus = null;
        this._cpuTimeProvider = null;
        const version = CartridgeCDF.getVersion(buffer);
        switch (version) {
            case 0:
                this._jumpstreamMask = 0xff;
                this._amplitudeStream = 34;
                this._datastreamBase = 0x06e0;
                this._datastreamIncrementBase = 0x0768;
                this._waveformBase = 0x7f0;
                break;
            case 1:
                this._jumpstreamMask = 0xff;
                this._amplitudeStream = 34;
                this._datastreamBase = 0x00a0;
                this._datastreamIncrementBase = 0x0128;
                this._waveformBase = 0x01b0;
                break;
            case 2:
                this._jumpstreamMask = 0xfe;
                this._amplitudeStream = 35;
                this._datastreamBase = 0x0098;
                this._datastreamIncrementBase = 0x0124;
                this._waveformBase = 0x01b0;
                break;
            default:
                throw new Error('not a CDF image: missing signature');
        }
        if (buffer.length !== 0x8000) {
            throw new Error(`not a CDF image: invalid lenght ${buffer.length}`);
        }
        this._soc = new Soc_1.default(version === 0 ? this._handleBxCDF0 : this._handleBxCDF1);
        this._soc.trap.addHandler(message => this.triggerTrap(2, message));
        this._rom = this._soc.getRom();
        for (let i = 0; i < 0x8000; i++) {
            this._rom[i] = buffer[i];
        }
        for (let i = 0; i < 7; i++) {
            this._banks[i] = new Uint8Array(this._rom.buffer, 0x1000 * (i + 1), 0x1000);
        }
        this._ram = this._soc.getRam();
        this._displayRam = new Uint8Array(this._soc.getRam().buffer, 0x0800, 0x1000);
        for (let i = 0; i < 3; i++) {
            this._musicStreams[i] = new MusicStream();
        }
        this.reset();
    }
    static getVersion(buffer) {
        const sig = 'CDF'.split('').map(x => x.charCodeAt(0)), startAddress = cartridgeUtil.searchForSignature(buffer, [...sig, -1, ...sig, -1, ...sig]);
        if (startAddress < 0) {
            return 3;
        }
        switch (buffer[startAddress + 3]) {
            case 0:
                return 0;
            case 1:
                return 1;
            case 'J'.charCodeAt(0):
                return 2;
            default:
                return 3;
        }
    }
    static matchesBuffer(buffer) {
        return CartridgeCDF.getVersion(buffer) !== 3;
    }
    init() {
        return this._soc.init();
    }
    reset() {
        for (let i = 0; i < 0x0200; i++) {
            this._soc.setRam32(i << 2, this._soc.getRom32(i << 2));
        }
        this._fastFetch = false;
        this._digitalAudio = false;
        this._fastJumpCountdown = 0;
        this._fastFetchPending = false;
        this._jmpOperandAddress = 0;
        this._ldaOperandAddress = 0;
        this._currentBank = this._banks[6];
        for (let i = 0; i < 3; i++) {
            this._musicStreams[i].reset();
        }
        this._lastCpuTime = 0;
        this._clockAccumulator = 0;
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_cdf;
    }
    setBus(bus) {
        this._bus = bus;
        return this;
    }
    setCpuTimeProvider(provider) {
        this._cpuTimeProvider = provider;
        return this;
    }
    read(address) {
        return this._access(address, this._bus.getLastDataBusValue());
    }
    peek(address) {
        return this._currentBank[address & 0x0fff];
    }
    write(address, value) {
        this._access(address, value);
    }
    _access(address, value) {
        address &= 0x0fff;
        const romValue = this._currentBank[address];
        if (this._fastJumpCountdown-- > 0 && address === this._jmpOperandAddress) {
            this._jmpOperandAddress++;
            return this._datastreamReadWithIncrement(this._jumpstream, 0x0100);
        }
        if (this._fastFetch &&
            romValue === 0x4c &&
            (this._currentBank[(address + 1) & 0x0fff] & this._jumpstreamMask) === 0 &&
            this._currentBank[(address + 2) & 0x0fff] === 0) {
            this._fastJumpCountdown = 2;
            this._jmpOperandAddress = (address + 1) & 0x0fff;
            this._jumpstream = 33 + this._currentBank[(address + 1) & 0x0fff];
            return romValue;
        }
        this._fastJumpCountdown = 0;
        if (this._fastFetch &&
            this._fastFetchPending &&
            this._ldaOperandAddress === address &&
            romValue <= this._amplitudeStream) {
            this._fastFetchPending = false;
            if (romValue === this._amplitudeStream) {
                this._clockMusicStreams();
                if (this._digitalAudio) {
                    const counter = this._musicStreams[0].counter, sampleAddress = this._soc.getRam32(this._waveformBase) + (counter >>> 21);
                    let sample = 0;
                    if (sampleAddress < 0x8000) {
                        sample = this._rom[sampleAddress];
                    }
                    if (sampleAddress > 0x40000000 && sampleAddress < 0x40002000) {
                        sample = this._ram[sampleAddress - 0x40000000];
                    }
                    if ((counter & 0x100000) === 0) {
                        sample >>>= 4;
                    }
                    return sample & 0x0f;
                }
                else {
                    let acc = 0;
                    for (let i = 0; i < 3; i++) {
                        acc += this._displayRam[(this._getWaveform(i) +
                            (this._musicStreams[i].counter >>> this._musicStreams[i].waveformSize)) &
                            0x0fff];
                    }
                    return acc;
                }
            }
            return this._datastreamRead(romValue);
        }
        this._fastFetchPending = false;
        if (address >= 0x0ff0) {
            switch (address) {
                case 0x0ff0:
                    this._datastreamWriteWithIncrement(32, value, 0x0100);
                    break;
                case 0x0ff1: {
                    let pointer = this._getDatastreamPointer(32);
                    pointer <<= 8;
                    pointer &= 0xf0000000;
                    pointer |= value << 20;
                    this._setDatastreamPointer(32, pointer);
                    break;
                }
                case 0x0ff2:
                    this._fastFetch = (value & 0x0f) === 0;
                    this._digitalAudio = (value & 0xf0) === 0;
                    break;
                case 0x0ff3:
                    switch (value) {
                        case 254:
                        case 255:
                            this._soc.run(0x080b);
                            break;
                    }
                    break;
            }
            if (address > 0x0ff4 && address < 0x0ffc) {
                this._currentBank = this._banks[address - 0x0ff5];
            }
        }
        if (this._fastFetch && romValue === 0xa9) {
            this._fastFetchPending = true;
            this._ldaOperandAddress = (address + 1) & 0x0fff;
        }
        return romValue;
    }
    _clockMusicStreams() {
        const cpuTime = this._cpuTimeProvider();
        this._clockAccumulator += (cpuTime - this._lastCpuTime) * 20000;
        this._lastCpuTime = cpuTime;
        const clocks = Math.floor(this._clockAccumulator);
        this._clockAccumulator -= clocks;
        if (clocks === 0) {
            return;
        }
        for (let i = 0; i < 3; i++) {
            this._musicStreams[i].increment(clocks);
        }
    }
    _getDatastreamPointer(stream) {
        return this._soc.getRam32(this._datastreamBase + 4 * stream);
    }
    _setDatastreamPointer(stream, value) {
        this._soc.setRam32(this._datastreamBase + 4 * stream, value);
    }
    _getDatastreamIncrement(stream) {
        return this._soc.getRam32(this._datastreamIncrementBase + 4 * stream);
    }
    _datastreamRead(stream) {
        const pointer = this._getDatastreamPointer(stream), value = this._displayRam[pointer >>> 20];
        this._setDatastreamPointer(stream, (pointer + (this._getDatastreamIncrement(stream) << 12)) | 0);
        return value;
    }
    _datastreamReadWithIncrement(stream, increment) {
        const pointer = this._getDatastreamPointer(stream), value = this._displayRam[pointer >>> 20];
        this._setDatastreamPointer(stream, (pointer + (increment << 12)) | 0);
        return value;
    }
    _datastreamWriteWithIncrement(stream, value, increment) {
        const pointer = this._getDatastreamPointer(stream);
        this._displayRam[pointer >>> 20] = value;
        this._setDatastreamPointer(stream, (pointer + (increment << 12)) | 0);
    }
    _getWaveform(index) {
        const value = this._soc.getRam32(this._waveformBase + 4 * index);
        return (value - 0x40000000 - 0x0800) & 0x0fff;
    }
}
exports.default = CartridgeCDF;
class MusicStream {
    constructor() {
        this.counter = 0;
        this.frequency = 0;
        this.waveformSize = 27;
    }
    reset() {
        this.counter = this.frequency = 0;
        this.waveformSize = 27;
    }
    increment(clocks) {
        this.counter = (this.counter + clocks * this.frequency) | 0;
    }
}
//# sourceMappingURL=CartridgeCDF.js.map