"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
const mixerTable = new Uint8Array([0x0, 0x4, 0x5, 0x9, 0x6, 0xa, 0xb, 0xf]);
class CartridgeDPC extends AbstractCartridge_1.default {
    constructor(buffer) {
        super();
        this._bank0 = new Uint8Array(0x1000);
        this._bank1 = new Uint8Array(0x1000);
        this._fetcherData = new Uint8Array(0x0800);
        this._fetchers = new Array(8);
        this._rng = 1;
        this._cpuTimeProvider = null;
        this._lastCpuTime = 0;
        this._clockAccumulator = 0;
        if (buffer.length < 0x2800) {
            throw new Error(`buffer is not a DPC image: too small ${buffer.length}`);
        }
        for (let i = 0; i < 8; i++) {
            this._fetchers[i] = new Fetcher();
        }
        for (let i = 0; i < 0x1000; i++) {
            this._bank0[i] = buffer[i];
            this._bank1[i] = buffer[0x1000 + i];
        }
        for (let i = 0; i < 0x0800; i++) {
            this._fetcherData[i] = buffer[0x2000 + i];
        }
        this.reset();
    }
    reset() {
        this._bank = this._bank1;
        this._rng = 1;
        this._fetchers.forEach(fetcher => fetcher.reset());
        this._lastCpuTime = 0;
        this._clockAccumulator = 0;
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_8k_DPC;
    }
    setBus(bus) {
        this._bus = bus;
        return this;
    }
    setCpuTimeProvider(provider) {
        this._cpuTimeProvider = provider;
        return this;
    }
    read(address) {
        return this._access(address, this._bus.getLastDataBusValue());
    }
    peek(address) {
        return this._bank[address & 0x0fff];
    }
    write(address, value) {
        this._access(address, value);
    }
    _access(address, value) {
        address &= 0x0fff;
        if (address > 0x7f) {
            switch (address) {
                case 0x0ff8:
                    this._bank = this._bank0;
                    break;
                case 0x0ff9:
                    this._bank = this._bank1;
                    break;
                default:
                    break;
            }
            return this._bank[address];
        }
        if (address < 0x08) {
            if (address & 0x04) {
                this._clockMusicFetchers();
                return mixerTable[(this._fetchers[5].mask & 0x4) | (this._fetchers[6].mask & 0x2) | (this._fetchers[7].mask & 0x1)];
            }
            else {
                return this._randomNext();
            }
        }
        if (address < 0x40) {
            const fetcher = this._fetchers[(address - 8) & 0x07], mask = fetcher.mask;
            let fetchedData = this._fetcherData[0x07ff - fetcher.pointer];
            fetcher.next();
            switch ((address - 8) >>> 3) {
                case 0:
                    return fetchedData;
                case 1:
                    return fetchedData & mask;
                case 2:
                    fetchedData &= mask;
                    return ((fetchedData << 4) | (fetchedData >>> 4)) & 0xff;
                case 3:
                    fetchedData &= mask;
                    return (((fetchedData & 0x01) << 7) |
                        ((fetchedData & 0x02) << 5) |
                        ((fetchedData & 0x04) << 3) |
                        ((fetchedData & 0x08) << 1) |
                        ((fetchedData & 0x10) >>> 1) |
                        ((fetchedData & 0x20) >>> 3) |
                        ((fetchedData & 0x40) >>> 5) |
                        ((fetchedData & 0x80) >>> 7));
                case 4:
                    return (fetchedData & mask) >>> 1;
                case 5:
                    return (fetchedData << 1) & mask;
                case 6:
                    return mask;
            }
        }
        if (address < 0x60) {
            const fetcher = this._fetchers[(address - 0x40) & 0x07];
            switch ((address - 0x40) >>> 3) {
                case 0:
                    this._clockMusicFetchers();
                    fetcher.setStart(value);
                    break;
                case 1:
                    this._clockMusicFetchers();
                    fetcher.setEnd(value);
                    break;
                case 2:
                    fetcher.setLow(value);
                    break;
                case 3:
                    fetcher.setHigh(value);
                    if (address > 0x5c) {
                        fetcher.setMusicMode(value);
                    }
                    break;
            }
            return this._bank[address];
        }
        if (address >= 0x70 && address < 0x78) {
            this._rng = 1;
            return this._bank[address];
        }
        return this._bank[address];
    }
    _randomNext() {
        const oldRng = this._rng;
        this._rng =
            ((this._rng << 1) |
                (~((this._rng >>> 7) ^ (this._rng >>> 5) ^ ((this._rng >>> 4) ^ (this._rng >>> 3))) & 0x01)) &
                0xff;
        return oldRng;
    }
    _clockMusicFetchers() {
        const cpuTime = this._cpuTimeProvider();
        this._clockAccumulator += (cpuTime - this._lastCpuTime) * 20000;
        this._lastCpuTime = cpuTime;
        const clocks = Math.floor(this._clockAccumulator);
        this._clockAccumulator -= clocks;
        if (clocks === 0) {
            return;
        }
        for (let i = 5; i < 8; i++) {
            this._fetchers[i].forwardClock(clocks);
        }
    }
}
exports.default = CartridgeDPC;
class Fetcher {
    constructor() {
        this.pointer = 0;
        this.start = 0;
        this.end = 0;
        this.musicMode = false;
        this.mask = 0x00;
    }
    contructor() {
        this.reset();
    }
    reset() {
        this.pointer = this.start = this.end = this.mask = 0;
        this.musicMode = false;
    }
    next() {
        if (this.musicMode) {
            return;
        }
        this.pointer = (this.pointer + 0x07ff) & 0x07ff;
        this._updateMask();
    }
    setStart(start) {
        this.start = start;
        this.mask = 0x00;
        this._updateMask();
    }
    setEnd(end) {
        this.end = end;
        this._updateMask();
    }
    setLow(value) {
        this.pointer = (this.pointer & 0x0700) | value;
        this._updateMask();
    }
    setHigh(value) {
        this.pointer = (this.pointer & 0xff) | ((value & 0x07) << 8);
        this._updateMask();
    }
    setMusicMode(value) {
        this.musicMode = (value & 0x10) !== 0;
    }
    forwardClock(clocks) {
        if (!this.musicMode) {
            return;
        }
        const period = this.start + 1, newPointerLow = ((this.pointer & 0xff) + period - clocks % period) % period, distanceStart = (256 + this.start - newPointerLow) & 0xff, distanceEnd = (256 + this.end - newPointerLow) & 0xff;
        this.pointer = (this.pointer & 0x0700) | newPointerLow;
        this.mask = distanceStart > distanceEnd ? 0 : 0xff;
    }
    _updateMask() {
        if ((this.pointer & 0xff) === this.start) {
            this.mask = 0xff;
        }
        if ((this.pointer & 0xff) === this.end) {
            this.mask = 0x00;
        }
    }
}
//# sourceMappingURL=CartridgeDPC.js.map