"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const Soc_1 = tslib_1.__importDefault(require("./harmony/Soc"));
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
const cartridgeUtil = tslib_1.__importStar(require("./util"));
class CartridgeDPCPlus extends AbstractCartridge_1.default {
    constructor(buffer) {
        super();
        this._banks = new Array(6);
        this._fetchers = new Array(8);
        this._fractionalFetchers = new Array(8);
        this._musicFetchers = new Array(3);
        this._parameters = new Uint8Array(8);
        this._parameterIndex = 0;
        this._rng = 0;
        this._clockAccumulator = 0;
        this._lastCpuTime = 0;
        this._fastFetch = false;
        this._ldaPending = false;
        this._cpuTimeProvider = null;
        this._soc = new Soc_1.default();
        if (buffer.length < 28 * 0x0400 || buffer.length > 0x8000) {
            throw new Error(`not a DPC+ image: invalid lenght ${buffer.length}`);
        }
        this._rom = this._soc.getRom();
        for (let i = 0; i < 6; i++) {
            this._banks[i] = new Uint8Array(this._rom.buffer, 0x0c00 + i * 0x1000, 0x1000);
        }
        this._ram = this._soc.getRam();
        this._imageRam = new Uint8Array(this._ram.buffer, 0x0c00, 0x1000);
        const offset = 0x8000 - buffer.length;
        for (let i = 0; i < buffer.length; i++) {
            this._rom[offset + i] = buffer[i];
        }
        for (let i = 0; i < 8; i++) {
            this._fetchers[i] = new Fetcher();
            this._fractionalFetchers[i] = new FractionalFetcher();
        }
        for (let i = 0; i < 3; i++) {
            this._musicFetchers[i] = new MusicFetcher();
        }
        this._soc.trap.addHandler(message => this.triggerTrap(2, message));
        this.reset();
    }
    static matchesBuffer(buffer) {
        const signatureCounts = cartridgeUtil.searchForSignatures(buffer, ['DPC+'.split('').map(x => x.charCodeAt(0))]);
        return signatureCounts[0] === 2;
    }
    init() {
        return this._soc.init();
    }
    reset() {
        this._soc.reset();
        this._currentBank = this._banks[5];
        for (let i = 0; i < 0x0300; i++) {
            this._soc.setRam32(i << 2, this._soc.getRom32(i << 2));
        }
        for (let i = 0x1b00; i < 0x2000; i++) {
            this._soc.setRam32((0x0300 + i - 0x1b00) << 2, this._soc.getRom32(i << 2));
        }
        this._currentBank = this._banks[5];
        for (let i = 0; i < 8; i++) {
            this._parameters[i] = 0;
        }
        for (let i = 0; i < 8; i++) {
            this._fetchers[i].reset();
            this._fractionalFetchers[i].reset();
        }
        for (let i = 0; i < 3; i++) {
            this._musicFetchers[i].reset();
        }
        this._parameterIndex = 0;
        this._fastFetch = this._ldaPending = false;
        this._rng = 0x2b435044;
        this._lastCpuTime = 0;
        this._clockAccumulator = 0;
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_dpc_plus;
    }
    setBus(bus) {
        this._bus = bus;
        return this;
    }
    setCpuTimeProvider(provider) {
        this._cpuTimeProvider = provider;
        return this;
    }
    read(address) {
        return this._access(address, this._bus.getLastDataBusValue());
    }
    peek(address) {
        return this._currentBank[address & 0x0fff];
    }
    write(address, value) {
        this._access(address, value);
    }
    triggerTrap(reason, message) {
        super.triggerTrap(reason, message);
    }
    _access(address, value) {
        address &= 0x0fff;
        const readResult = this._currentBank[address];
        if (this._fastFetch && this._ldaPending && address > 0x7f && address < 0x0ff6 && readResult < 0x28) {
            address = readResult;
        }
        this._ldaPending = false;
        if (address < 0x28) {
            const idx = address & 0x07, fetcher = this._fetchers[idx], fractionalFetcher = this._fractionalFetchers[idx];
            let result = 0;
            switch ((address >>> 3) & 0x07) {
                case 0x00:
                    switch (idx) {
                        case 0x00:
                            this._advanceRng();
                            return this._rng & 0xff;
                        case 0x01:
                            this._rewindRng();
                            return this._rng & 0xff;
                        case 0x02:
                            return (this._rng >>> 8) & 0xff;
                        case 0x03:
                            return (this._rng >>> 16) & 0xff;
                        case 0x04:
                            return (this._rng >>> 24) & 0xff;
                        case 0x05: {
                            this._clockMusicFetchers();
                            let acc = 0;
                            for (let i = 0; i < 3; i++) {
                                acc += this._imageRam[(this._musicFetchers[i].waveform << 5) + this._musicFetchers[i].waveformSample()];
                            }
                            return acc & 0xff;
                        }
                    }
                    return 0;
                case 0x01:
                    result = this._imageRam[fetcher.pointer];
                    fetcher.increment();
                    return result;
                case 0x02:
                    result = this._imageRam[fetcher.pointer] & fetcher.mask();
                    fetcher.increment();
                    return result;
                case 0x03:
                    result = this._imageRam[fractionalFetcher.pointer >>> 8];
                    fractionalFetcher.increment();
                    return result;
                case 0x04:
                    return idx < 4 ? fetcher.mask() : 0;
                default:
                    return 0;
            }
        }
        else if (address < 0x80) {
            const idx = address & 0x07, fetcher = this._fetchers[idx], fractionalFetcher = this._fractionalFetchers[idx];
            switch (((address - 0x28) >>> 3) & 0x0f) {
                case 0x00:
                    fractionalFetcher.setPointerLo(value);
                    break;
                case 0x01:
                    fractionalFetcher.setPointerHi(value);
                    break;
                case 0x02:
                    fractionalFetcher.setFraction(value);
                    break;
                case 0x03:
                    fetcher.top = value;
                    break;
                case 0x04:
                    fetcher.bottom = value;
                    break;
                case 0x05:
                    fetcher.setPointerLo(value);
                    break;
                case 0x06:
                    switch (idx) {
                        case 0x00:
                            this._fastFetch = value === 0;
                            break;
                        case 0x01:
                            if (this._parameterIndex < 8) {
                                this._parameters[this._parameterIndex++] = value;
                            }
                            break;
                        case 0x02:
                            this._dispatchFunction(value);
                            break;
                        case 0x05:
                        case 0x06:
                        case 0x07:
                            this._musicFetchers[idx - 0x05].waveform = value & 0x7f;
                            break;
                    }
                    break;
                case 0x07:
                    fetcher.decrement();
                    this._imageRam[fetcher.pointer] = value;
                    break;
                case 0x08:
                    fetcher.setPointerHi(value);
                    break;
                case 0x09:
                    switch (idx) {
                        case 0x00:
                            this._rng = 0x2b435044;
                            break;
                        case 0x01:
                            this._rng = (this._rng & 0xffffff00) | value;
                            break;
                        case 0x02:
                            this._rng = (this._rng & 0xffff00ff) | (value << 8);
                            break;
                        case 0x03:
                            this._rng = (this._rng & 0xff00ffff) | (value << 16);
                            break;
                        case 0x04:
                            this._rng = (this._rng & 0x00ffffff) | (value << 24);
                            break;
                        case 0x05:
                        case 0x06:
                        case 0x07:
                            this._musicFetchers[idx - 0x05].frequency = this._soc.getRam32(0x2000 - 0x400 + (value << 2));
                            break;
                    }
                    break;
                case 0x0a:
                    this._imageRam[fetcher.pointer] = value;
                    fetcher.increment();
                    break;
                default:
                    break;
            }
        }
        else if (address > 0x0ff5 && address < 0x0ffc) {
            this._currentBank = this._banks[address - 0x0ff6];
        }
        if (this._fastFetch && address > 0x7f && address < 0x0ff6) {
            this._ldaPending = readResult === 0xa9;
        }
        return readResult;
    }
    _clockMusicFetchers() {
        const cpuTime = this._cpuTimeProvider();
        this._clockAccumulator += (cpuTime - this._lastCpuTime) * 20000;
        this._lastCpuTime = cpuTime;
        const clocks = Math.floor(this._clockAccumulator);
        this._clockAccumulator -= clocks;
        if (clocks === 0) {
            return;
        }
        for (let i = 0; i < 3; i++) {
            this._musicFetchers[i].increment(clocks);
        }
    }
    _dispatchFunction(index) {
        const romBase = this._parameters[0] + (this._parameters[1] << 8);
        switch (index) {
            case 0:
                this._parameterIndex = 0;
                break;
            case 1:
                for (let i = 0; i < this._parameters[3]; i++) {
                    this._ram[0x0c00 + ((this._fetchers[this._parameters[2] & 0x07].pointer + i) & 0x0fff)] = this._rom[0x0c00 + (romBase + i) % 0x7400];
                }
                this._parameterIndex = 0;
                break;
            case 2:
                for (let i = 0; i < this._parameters[3]; i++) {
                    this._ram[0x0c00 + ((this._fetchers[this._parameters[2] & 0x07].pointer + i) & 0x0fff)] = this._parameters[0];
                }
                this._parameterIndex = 0;
                break;
            case 254:
            case 255:
                this._soc.run(0x0c0b);
                break;
        }
    }
    _advanceRng() {
        this._rng = (this._rng & (1 << 10) ? 0x10adab1e : 0x00) ^ ((this._rng >>> 11) | (this._rng << 21));
    }
    _rewindRng() {
        this._rng =
            this._rng & (1 << 31)
                ? ((0x10adab1e ^ this._rng) << 11) | ((0x10adab1e ^ this._rng) >>> 21)
                : (this._rng << 11) | (this._rng >>> 21);
    }
}
exports.default = CartridgeDPCPlus;
class Fetcher {
    constructor() {
        this.pointer = 0;
        this.top = 0;
        this.bottom = 0;
    }
    contructor() {
        this.reset();
    }
    reset() {
        this.pointer = this.top = this.bottom = 0;
    }
    setPointerHi(value) {
        this.pointer = (this.pointer & 0xff) | ((value & 0x0f) << 8);
    }
    setPointerLo(value) {
        this.pointer = (this.pointer & 0x0f00) | (value & 0xff);
    }
    increment() {
        this.pointer = (this.pointer + 1) & 0xfff;
    }
    decrement() {
        this.pointer = (this.pointer + 0xfff) & 0xfff;
    }
    mask() {
        return ((this.top - (this.pointer & 0xff)) & 0xff) > ((this.top - this.bottom) & 0xff) ? 0xff : 0;
    }
}
class FractionalFetcher {
    constructor() {
        this.pointer = 0;
        this.fraction = 0;
    }
    contructor() {
        this.reset();
    }
    reset() {
        this.pointer = this.fraction = 0;
    }
    setPointerHi(value) {
        this.pointer = (this.pointer & 0x00ffff) | ((value & 0x0f) << 16);
    }
    setPointerLo(value) {
        this.pointer = (this.pointer & 0x0f00ff) | ((value & 0xff) << 8);
    }
    setFraction(value) {
        this.fraction = value;
        this.pointer &= 0x0fff00;
    }
    increment() {
        this.pointer = (this.pointer + this.fraction) & 0x0fffff;
    }
}
class MusicFetcher {
    constructor() {
        this.frequency = ~0;
        this.counter = ~0;
        this.waveform = 0;
        this.reset();
    }
    reset() {
        this.frequency = 0;
        this.waveform = 0;
        this.counter = 0;
    }
    increment(clocks) {
        this.counter = (this.counter + clocks * this.frequency) | 0;
    }
    waveformSample() {
        return this.counter >>> 27;
    }
}
//# sourceMappingURL=CartridgeDPCPlus.js.map