"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const cartridgeUtil = tslib_1.__importStar(require("./util"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
class CartrdigeE7 extends AbstractCartridge_1.default {
    constructor(buffer) {
        super();
        this._banks = new Array(8);
        this._ram0 = new Uint8Array(0x0400);
        this._ram1Banks = new Array(4);
        this._ram0Enabled = false;
        if (buffer.length !== 0x4000) {
            throw new Error(`buffer is not a 16k cartridge image: wrong length ${buffer.length}`);
        }
        for (let i = 0; i < 8; i++) {
            this._banks[i] = new Uint8Array(0x0800);
        }
        for (let i = 0; i < 4; i++) {
            this._ram1Banks[i] = new Uint8Array(0x100);
        }
        for (let i = 0; i < 0x0800; i++) {
            for (let j = 0; j < 8; j++) {
                this._banks[j][i] = buffer[j * 0x0800 + i];
            }
        }
        this.reset();
    }
    static matchesBuffer(buffer) {
        const signatureCounts = cartridgeUtil.searchForSignatures(buffer, [
            [0xad, 0xe2, 0xff],
            [0xad, 0xe5, 0xff],
            [0xad, 0xe5, 0x1f],
            [0xad, 0xe7, 0x1f],
            [0x0c, 0xe7, 0x1f],
            [0x8d, 0xe7, 0xff],
            [0x8d, 0xe7, 0x1f]
        ]);
        for (let i = 0; i < signatureCounts.length; i++) {
            if (signatureCounts[i] > 0) {
                return true;
            }
        }
        return false;
    }
    reset() {
        this._bank0 = this._banks[0];
        this._ram1 = this._ram1Banks[0];
        this._ram0Enabled = false;
    }
    read(address) {
        this._handleBankswitch(address & 0x0fff);
        return this.peek(address);
    }
    peek(address) {
        address &= 0x0fff;
        if (address < 0x0800) {
            if (this._ram0Enabled) {
                return address >= 0x0400 ? this._ram0[address - 0x0400] : 0;
            }
            else {
                return this._bank0[address];
            }
        }
        if (address <= 0x09ff) {
            return address >= 0x0900 ? this._ram1[address - 0x0900] : 0;
        }
        return this._banks[7][0x07ff - (0x0fff - address)];
    }
    write(address, value) {
        address &= 0x0fff;
        this._handleBankswitch(address);
        if (address < 0x0400) {
            if (this._ram0Enabled) {
                this._ram0[address] = value;
            }
            else {
                super.write(address, value);
            }
        }
        else if (address < 0x0800) {
            super.write(address, value);
        }
        else if (address < 0x08ff) {
            this._ram1[address - 0x0800] = value;
        }
        else {
            super.write(address, value);
        }
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_16k_E7;
    }
    randomize(rng) {
        for (let i = 0; i < 4; i++) {
            for (let j = 0; j < this._ram1Banks[i].length; j++) {
                this._ram1Banks[i][j] = rng.int(0xff);
            }
        }
        for (let i = 0; i < this._ram0.length; i++) {
            this._ram0[i] = rng.int(0xff);
        }
    }
    _handleBankswitch(address) {
        if (address < 0x0fe0) {
            return;
        }
        if (address <= 0x0fe6) {
            this._bank0 = this._banks[address & 0x000f];
            this._ram0Enabled = false;
        }
        else if (address === 0x0fe7) {
            this._ram0Enabled = true;
        }
        else if (address <= 0x0feb) {
            this._ram1 = this._ram1Banks[address - 0x0fe8];
        }
    }
}
exports.default = CartrdigeE7;
//# sourceMappingURL=CartridgeE7.js.map