"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
const cartridgeUtil = tslib_1.__importStar(require("./util"));
class CartridgeEF extends AbstractCartridge_1.default {
    constructor(buffer, _supportSC = true) {
        super();
        this._supportSC = _supportSC;
        this._bus = null;
        this._banks = new Array(16);
        this._ram = new Uint8Array(0x80);
        this._hasSC = false;
        if (buffer.length !== 0x10000) {
            throw new Error(`buffer is not a 64k cartridge image: wrong length ${buffer.length}`);
        }
        for (let i = 0; i < 16; i++) {
            this._banks[i] = new Uint8Array(0x1000);
        }
        for (let i = 0; i < 0x1000; i++) {
            for (let j = 0; j < 16; j++) {
                this._banks[j][i] = buffer[j * 0x1000 + i];
            }
        }
        this.reset();
    }
    static matchesBuffer(buffer) {
        const matchMagic = (magicString) => {
            const magic = magicString.split('').map(x => x.charCodeAt(0));
            for (let i = 0; i < magic.length; i++) {
                if (magic[i] !== buffer[0xfff8 + i]) {
                    return false;
                }
            }
            return true;
        };
        if (buffer.length !== 0x10000) {
            return false;
        }
        if (matchMagic('efef') || matchMagic('efsc')) {
            return true;
        }
        const signatureCounts = cartridgeUtil.searchForSignatures(buffer, [
            [0x0c, 0xe0, 0xff],
            [0xad, 0xe0, 0xff],
            [0x0c, 0xe0, 0x1f],
            [0xad, 0xe0, 0x1f]
        ]);
        for (let i = 0; i < 4; i++) {
            if (signatureCounts[i] > 0) {
                return true;
            }
        }
        return false;
    }
    reset() {
        this._bank = this._banks[15];
        this._hasSC = false;
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_64k_EF;
    }
    randomize(rng) {
        for (let i = 0; i < this._ram.length; i++) {
            this._ram[i] = rng.int(0xff);
        }
    }
    setBus(bus) {
        this._bus = bus;
        return this;
    }
    read(address) {
        this._access(address & 0x0fff, this._bus.getLastDataBusValue());
        return this.peek(address);
    }
    peek(address) {
        address &= 0x0fff;
        if (this._hasSC && address >= 0x0080 && address < 0x0100) {
            return this._ram[address - 0x80];
        }
        else {
            return this._bank[address];
        }
    }
    write(address, value) {
        address &= 0x0fff;
        if (address < 0x80 && this._supportSC) {
            this._hasSC = true;
        }
        this._access(address, value);
    }
    _access(address, value) {
        if (address < 0x80 && this._hasSC) {
            this._ram[address] = value;
            return;
        }
        if (address >= 0x0fe0 && address <= 0x0fef) {
            this._bank = this._banks[address - 0x0fe0];
        }
    }
}
exports.default = CartridgeEF;
//# sourceMappingURL=CartridgeEF.js.map