"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
const cartridgeUtil = tslib_1.__importStar(require("./util"));
class CartridgeFA2 extends AbstractCartridge_1.default {
    constructor(buffer) {
        super();
        this._banks = new Array(7);
        this._ram = new Uint8Array(0x100);
        this._savedRam = new Uint8Array(0x100);
        this._accessCounter = 0;
        this._accessCounterLimit = 0;
        if (buffer.length !== 0x7000 && buffer.length !== 0x7400) {
            throw new Error(`buffer is not a 28k/29k cartridge image: wrong length ${buffer.length}`);
        }
        for (let i = 0; i < 7; i++) {
            this._banks[i] = new Uint8Array(0x1000);
        }
        const offset = buffer.length === 0x7000 ? 0 : 0x0400;
        for (let i = 0; i < 0x1000; i++) {
            for (let j = 0; j < 7; j++) {
                this._banks[j][i] = buffer[j * 0x1000 + i + offset];
            }
        }
        this.reset();
    }
    static matchesBuffer(buffer) {
        const signatureCounts = cartridgeUtil.searchForSignatures(buffer, [
            [0xa0, 0xc1, 0x1f, 0xe0],
            [0x00, 0x80, 0x02, 0xe0]
        ]);
        return signatureCounts[0] > 0 || signatureCounts[1] > 0;
    }
    reset() {
        this._accessCounter = 0;
        this._accessCounterLimit = 0;
        this._bank = this._banks[0];
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_FA2;
    }
    randomize(rng) {
        for (let i = 0; i < this._ram.length; i++) {
            this._ram[i] = rng.int(0xff);
        }
    }
    setBus(bus) {
        this._bus = bus;
        return this;
    }
    read(address) {
        this.write(address & 0x0fff, this._bus.getLastDataBusValue());
        return this.peek(address);
    }
    peek(address) {
        address &= 0x0fff;
        if (address >= 0x0100 && address < 0x0200) {
            return this._ram[address - 0x0100];
        }
        else if (address === 0x0ff4) {
            return this._accessCounter >= this._accessCounterLimit
                ?
                    this._bank[address] & ~0x40
                :
                    this._bank[address] | 0x40;
        }
        else {
            return this._bank[address];
        }
    }
    write(address, value) {
        address &= 0x0fff;
        this._accessCounter++;
        if (address < 0x0100) {
            this._ram[address] = value;
            return;
        }
        if (address === 0x0ff4) {
            return this._handleIo();
        }
        if (address >= 0x0ff5 && address <= 0x0ffb) {
            this._bank = this._banks[address - 0x0ff5];
        }
    }
    _handleIo() {
        if (this._accessCounter < this._accessCounterLimit) {
            return;
        }
        if (this._ram[0xff] === 1) {
            for (let i = 0; i < 0x100; i++) {
                this._ram[i] = this._savedRam[i];
            }
            this._accessCounterLimit = 10;
        }
        else if (this._ram[0xff] === 2) {
            for (let i = 0; i < 0x100; i++) {
                this._savedRam[i] = this._ram[i];
            }
            this._accessCounterLimit = 100;
        }
        else {
            return;
        }
        this._accessCounter = 0;
        this._ram[0xff] = 0;
    }
}
exports.default = CartridgeFA2;
//# sourceMappingURL=CartridgeFA2.js.map