"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
class CartridgePP extends AbstractCartridge_1.default {
    constructor(buffer) {
        super();
        this._banks = new Array(8);
        this._segments = new Array(4);
        this._ram = new Uint8Array(0x40);
        this._bankSwitchPending = false;
        this._pendingBank = 0;
        this._accessCounter = 0;
        this._previousAddressBusValue = 0;
        if (buffer.length !== 0x2000 && buffer.length !== 0x2003) {
            throw new Error(`buffer is not a PP cartridge image; wrong length ${buffer.length} bytes`);
        }
        for (let i = 0; i < 8; i++) {
            this._banks[i] = new Uint8Array(0x0400);
            for (let j = 0; j < 0x0400; j++) {
                this._banks[i][j] = buffer[i * 0x0400 + j];
            }
        }
        this.reset();
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_8k_pp;
    }
    reset() {
        this._switchLayout(0);
        this._bankSwitchPending = false;
    }
    randomize(rng) {
        for (let i = 0; i < 0x40; i++) {
            this._ram[i] = rng.int(0xff);
        }
    }
    setBus(bus) {
        this._bus = bus;
        bus.event.read.addHandler(CartridgePP._onBusAccess, this);
        bus.event.write.addHandler(CartridgePP._onBusAccess, this);
        return this;
    }
    read(address) {
        address &= 0x0fff;
        this._access(address, this._bus.getLastDataBusValue());
        if (address < 0x40) {
            return this._ram[address];
        }
        if (address < 0x80) {
            return this._bus.getLastDataBusValue();
        }
        return this._segments[address >>> 10][address & 0x03ff];
    }
    write(address, value) {
        this._access(address & 0x0fff, value);
    }
    peek(address) {
        address &= 0x0fff;
        if (address < 0x40) {
            return this._ram[address];
        }
        if (address < 0x80) {
            return this._bus.getLastDataBusValue();
        }
        return this._segments[address >>> 10][address & 0x03ff];
    }
    static _onBusAccess(accessType, self) {
        let address = self._bus.getLastAddresBusValue();
        if (self._bankSwitchPending && address !== self._previousAddressBusValue && --self._accessCounter === 0) {
            self._switchLayout(self._pendingBank);
            self._bankSwitchPending = false;
        }
        if (address & 0x1000) {
            return;
        }
        address &= 0xff;
        if (address >= 0x30 && address <= 0x3f) {
            self._pendingBank = address & 0x0f;
            self._accessCounter = 3;
            self._bankSwitchPending = true;
        }
    }
    _access(address, value) {
        if (address >= 0x40 && address < 0x80) {
            this._ram[address - 0x40] = value;
        }
    }
    _switchLayout(index) {
        switch (index) {
            case 0:
            case 8:
                return this._configureSegments(0, 0, 1, 2);
            case 1:
            case 9:
                return this._configureSegments(0, 1, 3, 2);
            case 2:
            case 10:
                return this._configureSegments(4, 5, 6, 7);
            case 3:
            case 11:
                return this._configureSegments(7, 4, 3, 2);
            case 4:
            case 12:
                return this._configureSegments(0, 0, 6, 7);
            case 5:
            case 13:
                return this._configureSegments(0, 1, 7, 6);
            case 6:
            case 14:
                return this._configureSegments(3, 2, 4, 5);
            case 7:
            case 15:
                return this._configureSegments(6, 0, 5, 1);
            default:
                throw new Error('illegal layout index');
        }
    }
    _configureSegments(zero, one, two, three) {
        this._segments[0] = this._banks[zero];
        this._segments[1] = this._banks[one];
        this._segments[2] = this._banks[two];
        this._segments[3] = this._banks[three];
    }
}
exports.default = CartridgePP;
//# sourceMappingURL=CartridgePP.js.map