"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const AbstractCartridge_1 = tslib_1.__importDefault(require("./AbstractCartridge"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("./CartridgeInfo"));
const Header_1 = tslib_1.__importDefault(require("./supercharger/Header"));
const blob_1 = require("./supercharger/blob");
class CartridgeSupercharger extends AbstractCartridge_1.default {
    constructor(buffer) {
        super();
        this._loadCount = 0;
        this._loads = null;
        this._headers = null;
        this._rom = new Uint8Array(0x800);
        this._ramBanks = new Array(3);
        this._bank0 = null;
        this._bank1 = null;
        this._bank1Type = 1;
        this._transitionCount = 0;
        this._pendingWriteData = 0;
        this._pendingWrite = false;
        this._lastAddressBusValue = -1;
        this._writeRamEnabled = false;
        this._loadInProgress = false;
        this._loadTimestamp = 0;
        this._cpuTimeProvider = null;
        if (buffer.length % 8448 !== 0) {
            throw new Error(`not a supercharger image --- invalid size`);
        }
        this._loadCount = buffer.length / 8448;
        this._loads = new Array(this._loadCount);
        this._headers = new Array(this._loadCount);
        for (let i = 0; i < this._loadCount; i++) {
            this._loads[i] = new Uint8Array(8448);
        }
        for (let i = 0; i < 8448; i++) {
            for (let j = 0; j < this._loadCount; j++) {
                this._loads[j][i] = buffer[j * 8448 + i];
            }
        }
        for (let i = 0; i < this._loadCount; i++) {
            this._headers[i] = new Header_1.default(this._loads[i]);
            if (!this._headers[i].verify()) {
                console.log(`load ${i} has invalid checksum`);
            }
        }
        for (let i = 0; i < 3; i++) {
            this._ramBanks[i] = new Uint8Array(0x0800);
        }
        this._setupRom();
        this.reset();
    }
    reset() {
        this._setBankswitchMode(0);
        this._transitionCount = 0;
        this._pendingWrite = false;
        this._pendingWriteData = 0;
        this._lastAddressBusValue = -1;
        this._writeRamEnabled = false;
        this._loadInProgress = false;
        this._loadTimestamp = 0;
    }
    setBus(bus) {
        this._bus = bus;
        this._bus.event.read.addHandler(CartridgeSupercharger._onBusAccess, this);
        this._bus.event.write.addHandler(CartridgeSupercharger._onBusAccess, this);
        return this;
    }
    setCpuTimeProvider(provider) {
        this._cpuTimeProvider = provider;
        return this;
    }
    setRng(rng) {
        this._rng = rng;
        return this;
    }
    read(address) {
        return this._access(address, this._bus.getLastDataBusValue());
    }
    peek(address) {
        address &= 0x0fff;
        return address < 0x0800 ? this._bank0[address] : this._bank1[address & 0x07ff];
    }
    write(address, value) {
        this._access(address, value);
    }
    getType() {
        return CartridgeInfo_1.default.CartridgeType.bankswitch_supercharger;
    }
    static _onBusAccess(type, self) {
        const address = self._bus.getLastAddresBusValue();
        if (address !== self._lastAddressBusValue && !self._loadInProgress) {
            if (self._transitionCount <= 5) {
                self._transitionCount++;
            }
            self._lastAddressBusValue = address;
        }
    }
    _access(address, value) {
        address &= 0x0fff;
        if (this._loadInProgress) {
            if ((this._cpuTimeProvider() - this._loadTimestamp) > 1E-3) {
                this._loadInProgress = false;
            }
            else {
                return value;
            }
        }
        const readValue = address < 0x0800 ? this._bank0[address] : this._bank1[address & 0x07ff];
        if ((address & 0x0f00) === 0 && (!this._pendingWrite || !this._writeRamEnabled)) {
            this._pendingWriteData = address & 0x00ff;
            this._transitionCount = 0;
            this._pendingWrite = true;
            return readValue;
        }
        if (address === 0x0ff8) {
            this._setBankswitchMode((this._pendingWriteData & 28) >>> 2);
            this._writeRamEnabled = (this._pendingWriteData & 0x02) > 0;
            this._pendingWrite = false;
            return readValue;
        }
        if (address === 0x0ff9 && this._bank1Type === 1 && ((this._lastAddressBusValue & 0x1fff) < 0xff)) {
            this._loadIntoRam(value);
            return readValue;
        }
        if (this._pendingWrite && this._writeRamEnabled && this._transitionCount === 5) {
            this._pendingWrite = false;
            if (address < 0x0800) {
                this._bank0[address] = this._pendingWriteData;
            }
            else if (this._bank1Type === 0) {
                this._bank1[address & 0x07ff] = this._pendingWriteData;
            }
            else {
                return readValue;
            }
            return this._pendingWriteData;
        }
        return readValue;
    }
    _setBankswitchMode(mode) {
        switch (mode) {
            case 0:
                return this._configureBanks(2, 1);
            case 1:
                return this._configureBanks(0, 1);
            case 2:
                return this._configureBanks(2, 0, 0);
            case 3:
                return this._configureBanks(0, 0, 2);
            case 4:
                return this._configureBanks(2, 1);
            case 5:
                return this._configureBanks(1, 1);
            case 6:
                return this._configureBanks(2, 0, 1);
            case 7:
                return this._configureBanks(1, 0, 2);
            default:
                throw new Error('invalid bankswitching mode');
        }
    }
    _configureBanks(bank0, bank1Type, bank1 = 0) {
        this._bank0 = this._ramBanks[bank0];
        this._bank1Type = bank1Type;
        this._bank1 = bank1Type === 0 ? this._ramBanks[bank1] : this._rom;
    }
    _setupRom() {
        for (let i = 0; i < 0x0800; i++) {
            this._rom[i] = 0;
        }
        for (let i = 0; i < blob_1.bios.length; i++) {
            this._rom[i] = blob_1.bios[i];
        }
        this._rom[0x07ff] = this._rom[0x07fd] = 0xf8;
        this._rom[0x07fe] = this._rom[0x07fc] = 0x07;
    }
    _loadIntoRam(loadId) {
        let loadIndex;
        for (loadIndex = 0; loadIndex < this._loadCount; loadIndex++) {
            if (this._headers[loadIndex].multiloadId === loadId || this._loadCount === 1) {
                break;
            }
        }
        if (loadIndex >= this._loadCount) {
            console.log(`no load with id ${loadId}`);
        }
        const header = this._headers[loadIndex], load = this._loads[loadIndex];
        for (let blockIdx = 0; blockIdx < header.blockCount; blockIdx++) {
            const location = header.blockLocation[blockIdx];
            let bank = location & 0x03;
            if (bank > 2) {
                bank = 0;
                console.log(`invalid bank for block ${blockIdx}, load ${loadIndex}`);
            }
            const base = ((location & 28) >>> 2) * 256;
            let checksum = location + header.blockChecksum[blockIdx];
            for (let i = 0; i < 256; i++) {
                checksum += load[256 * blockIdx + i];
                this._ramBanks[bank][base + i] = load[256 * blockIdx + i];
            }
            if ((checksum & 0xff) !== 0x55) {
                console.log(`load ${loadIndex}, block ${blockIdx}: invalid checksum`);
            }
        }
        this._rom[0x7f0] = header.controlWord;
        this._rom[0x7f1] = this._rng.int(0xff);
        this._rom[0x7f2] = header.startAddressLow;
        this._rom[0x7f3] = header.startAddressHigh;
        this._loadInProgress = true;
        this._loadTimestamp = this._cpuTimeProvider();
    }
}
exports.default = CartridgeSupercharger;
//# sourceMappingURL=CartridgeSupercharger.js.map