"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const thumbulator_ts_1 = tslib_1.__importDefault(require("thumbulator.ts"));
const microevent_ts_1 = require("microevent.ts");
const hex_1 = require("../../../../tools/hex");
function hostIsLittleEndian() {
    const buffer8 = new Uint8Array([1, 2, 3, 4]), buffer32 = new Uint32Array(buffer8.buffer);
    return buffer32[0] === 0x04030201;
}
class Soc {
    constructor(blx32Handler = () => 3) {
        this.trap = new microevent_ts_1.Event();
        this._thumbulatorBus = {
            read16: (address) => {
                if (address & 0x01) {
                    this._triggerTrap(`unaligned 16 bit ARM read from ${hex_1.encode(address, 8, false)}`);
                    return 0;
                }
                const region = address >>> 28, addr = address & 0x0fffffff;
                switch (region) {
                    case 0x0:
                        if (addr < 0x8000) {
                            return this.getRom16(addr);
                        }
                        break;
                    case 0x4:
                        if (addr < 0x2000) {
                            return this.getRam16(addr);
                        }
                        break;
                    case 0xe:
                        switch (addr) {
                            case 0x001fc000:
                                return this._armMamcr;
                        }
                        break;
                    default:
                }
                this._triggerTrap(`invalid 16 bit ARM read from ${hex_1.encode(address, 8, false)}`);
            },
            read32: (address) => {
                if (address & 0x03) {
                    this._triggerTrap(`unaligned 32 bit ARM read from ${hex_1.encode(address, 8, false)}`);
                    return 0;
                }
                const region = address >>> 28, addr = address & 0x0fffffff;
                switch (region) {
                    case 0x0:
                        if (addr < 0x8000) {
                            return this.getRom32(addr);
                        }
                        break;
                    case 0x4:
                        if (addr < 0x2000) {
                            return this.getRam32(addr);
                        }
                        break;
                    case 0xe:
                        switch (addr) {
                            case 0x8004:
                            case 0x8008:
                                return 0;
                        }
                        break;
                    default:
                }
                this._triggerTrap(`invalid 32 bit ARM read from ${hex_1.encode(address, 8, false)}`);
            },
            write16: (address, value) => {
                if (address & 0x01) {
                    this._triggerTrap(`unaligned 16 bit ARM write: ${hex_1.encode(value, 4)} -> ${hex_1.encode(address, 8, false)}`);
                    return;
                }
                const region = address >>> 28, addr = address & 0x0fffffff;
                switch (region) {
                    case 0x04:
                        if (addr < 0x2000) {
                            this.setRam16(addr, value & 0xffff);
                            return;
                        }
                        break;
                    case 0xe:
                        switch (addr) {
                            case 0x001fc000:
                                this._armMamcr = value;
                                return;
                        }
                        break;
                }
                this._triggerTrap(`invalid 16 bit ARM write: ${hex_1.encode(value, 4)} -> ${hex_1.encode(address, 8, false)}`);
            },
            write32: (address, value) => {
                if (address & 0x03) {
                    this._triggerTrap(`unaligned 32 bit ARM write: ${hex_1.encode(value, 8, false)} -> ${hex_1.encode(address, 8, false)}`);
                    return;
                }
                const region = address >>> 28, addr = address & 0x0fffffff;
                switch (region) {
                    case 0x4:
                        if (addr < 0x2000) {
                            this.setRam32(addr, value);
                            return;
                        }
                    case 0xe:
                        switch (addr) {
                            case 0x8004:
                            case 0x8008:
                                return;
                        }
                        break;
                }
                this._triggerTrap(`invalid 32 bit ARM write: ${hex_1.encode(value, 8, false)} -> ${hex_1.encode(address, 8, false)}`);
            }
        };
        this._romBuffer = new ArrayBuffer(0x8000);
        this._ramBuffer = new ArrayBuffer(0x2000);
        this._armMamcr = 0;
        this._thumbulator = null;
        if (hostIsLittleEndian()) {
            this.getRom16 = address => this._rom16[address >>> 1];
            this.getRom32 = address => this._rom32[address >>> 2];
            this.getRam16 = address => this._ram16[address >>> 1];
            this.getRam32 = address => this._ram32[address >>> 2];
            this.setRam16 = (address, value) => (this._ram16[address >>> 1] = value);
            this.setRam32 = (address, value) => (this._ram32[address >>> 2] = value);
        }
        else {
            this.getRom16 = address => this._rom8[address] | (this._rom8[address + 1] << 8);
            this.getRom32 = address => this._rom8[address] |
                (this._rom8[address + 1] << 8) |
                (this._rom8[address + 2] << 16) |
                (this._rom8[address + 3] << 24);
            this.getRam16 = address => this._ram8[address] | (this._ram8[address + 1] << 8);
            this.getRam32 = address => this._ram8[address] |
                (this._ram8[address + 1] << 8) |
                (this._ram8[address + 2] << 16) |
                (this._ram8[address + 3] << 24);
            this.setRam16 = (address, value) => {
                this._ram8[address] = value & 0xff;
                this._ram8[address + 1] = (value >>> 8) & 0xff;
            };
            this.setRam32 = (address, value) => {
                this._ram8[address] = value & 0xff;
                this._ram8[address + 1] = (value >>> 8) & 0xff;
                this._ram8[address + 2] = (value >>> 16) & 0xff;
                this._ram8[address + 3] = (value >>> 24) & 0xff;
            };
        }
        this._rom8 = new Uint8Array(this._romBuffer);
        this._rom16 = new Uint16Array(this._romBuffer);
        this._rom32 = new Uint32Array(this._romBuffer);
        this._ram8 = new Uint8Array(this._ramBuffer);
        this._ram16 = new Uint16Array(this._ramBuffer);
        this._ram32 = new Uint32Array(this._ramBuffer);
        this._thumbulator = new thumbulator_ts_1.default(this._thumbulatorBus, {
            stopAddress: 32772,
            trapOnBx32: blx32Handler
        });
        this.reset();
    }
    init() {
        return this._thumbulator.init();
    }
    reset() { }
    getRom() {
        return this._rom8;
    }
    getRam() {
        return this._ram8;
    }
    run(entry) {
        this._thumbulator.reset();
        this._thumbulator.enableDebug(false);
        for (let i = 0; i <= 12; i++) {
            this._thumbulator.writeRegister(i, 0);
        }
        this._thumbulator.writeRegister(13, 0x40001fb4);
        this._thumbulator.writeRegister(14, 32772 + 1);
        this._thumbulator.writeRegister(15, entry);
        this._armMamcr = 0;
        const trap = this._thumbulator.run(500000);
        if (trap !== 20 && trap !== 10) {
            this._triggerTrap(`ARM execution trapped: ${trap}`);
        }
    }
    getThumbulator() {
        return this._thumbulator;
    }
    _triggerTrap(message) {
        this._thumbulator.abort();
        this.trap.dispatch(message);
    }
}
exports.default = Soc;
//# sourceMappingURL=Soc.js.map