"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const microevent_ts_1 = require("microevent.ts");
class FrameManager {
    constructor(_config) {
        this._config = _config;
        this.newFrame = new microevent_ts_1.Event();
        this.vblank = false;
        this.surfaceBuffer = null;
        this._vblankLines = 0;
        this._kernelLines = 0;
        this._overscanLines = 0;
        this._linesWithoutVsync = 0;
        this._state = 0;
        this._vsync = false;
        this._lineInState = 0;
        this._surfaceFactory = null;
        this._surface = null;
        this._frameStart = -1;
        switch (this._config.tvMode) {
            case 0:
                this._vblankLines = 40;
                this._kernelLines = 192;
                this._overscanLines = 30;
                break;
            case 1:
            case 2:
                this._vblankLines = 48;
                this._kernelLines = 228;
                this._overscanLines = 36;
                break;
            default:
                throw new Error(`invalid tv mode ${this._config.tvMode}`);
        }
        this._frameStart = this._config.frameStart;
        this.reset();
    }
    reset() {
        this.vblank = false;
        this.surfaceBuffer = null;
        this._linesWithoutVsync = 0;
        this._state = 0;
        this._vsync = false;
        this._lineInState = 0;
        this._surface = null;
    }
    nextLine() {
        if (!this._surfaceFactory) {
            return;
        }
        this._lineInState++;
        switch (this._state) {
            case 0:
            case 1:
                if (++this._linesWithoutVsync > 150) {
                    this._setState(2);
                }
                break;
            case 2:
                if (this._frameStart >= 0) {
                    if (this._lineInState > this._frameStart) {
                        this._startFrame();
                    }
                }
                else {
                    if (this._lineInState >=
                        (this.vblank ? this._vblankLines : this._vblankLines - 10)) {
                        this._startFrame();
                    }
                }
                break;
            case 3:
                if (this._lineInState >= this._kernelLines + 20) {
                    this._finalizeFrame();
                }
                break;
            case 4:
                if (this._lineInState >= this._overscanLines - 20) {
                    this._setState(0);
                }
                break;
        }
    }
    isRendering() {
        return this._state === 3 && !!this._surface;
    }
    setVblank(vblank) {
        if (this._surfaceFactory) {
            this.vblank = vblank;
        }
    }
    setVsync(vsync) {
        if (!this._surfaceFactory || vsync === this._vsync) {
            return;
        }
        this._vsync = vsync;
        switch (this._state) {
            case 0:
                this._linesWithoutVsync = 0;
            case 2:
            case 4:
                if (vsync) {
                    this._setState(1);
                }
                break;
            case 1:
                if (!vsync) {
                    this._setState(2);
                }
                break;
            case 3:
                if (vsync) {
                    this._finalizeFrame();
                }
                break;
        }
    }
    getHeight() {
        return this._kernelLines + 20;
    }
    setSurfaceFactory(factory) {
        this._surfaceFactory = factory;
    }
    getCurrentLine() {
        return this._state === 3 ? this._lineInState : 0;
    }
    getDebugState() {
        return `${this._getReadableState()}, line = ${this._lineInState}, ` + `vblank = ${this.vblank ? '1' : '0'}`;
    }
    _getReadableState() {
        switch (this._state) {
            case 0:
                return `wait for vsync start`;
            case 1:
                return `wait for vsync end`;
            case 2:
                return `wait for frame start`;
            case 3:
                return `frame`;
            case 4:
                return `overscan`;
        }
    }
    _startFrame() {
        this._setState(3);
        this._surface = this._surfaceFactory();
        this.surfaceBuffer = this._surface.getBuffer();
    }
    _finalizeFrame() {
        if (this._state !== 3) {
            throw new Error(`finalize frame in invalid state ${this._state}`);
        }
        this.newFrame.dispatch(this._surface);
        this._setState(4);
    }
    _setState(newState) {
        this._state = newState;
        this._lineInState = 0;
    }
}
exports.default = FrameManager;
//# sourceMappingURL=FrameManager.js.map