"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const microevent_ts_1 = require("microevent.ts");
const WaveformAudio_1 = tslib_1.__importDefault(require("./WaveformAudio"));
const PCMAudio_1 = tslib_1.__importDefault(require("./PCMAudio"));
const Missile_1 = tslib_1.__importDefault(require("./Missile"));
const Playfield_1 = tslib_1.__importDefault(require("./Playfield"));
const Player_1 = tslib_1.__importDefault(require("./Player"));
const Ball_1 = tslib_1.__importDefault(require("./Ball"));
const LatchedInput_1 = tslib_1.__importDefault(require("./LatchedInput"));
const PaddleReader_1 = tslib_1.__importDefault(require("./PaddleReader"));
const FrameManager_1 = tslib_1.__importDefault(require("./FrameManager"));
const DelayQueue_1 = tslib_1.__importDefault(require("./DelayQueue"));
const palette = tslib_1.__importStar(require("./palette"));
class Tia {
    constructor(_config, joystick0, joystick1, paddles) {
        this._config = _config;
        this.newFrame = new microevent_ts_1.Event();
        this.trap = new microevent_ts_1.Event();
        this._cpu = null;
        this._bus = null;
        this._delayQueue = new DelayQueue_1.default(10, 20);
        this._hstate = 0;
        this._hctr = 0;
        this._collisionUpdateRequired = false;
        this._movementClock = 0;
        this._movementInProgress = false;
        this._extendedHblank = false;
        this._xDelta = 0;
        this._linesSinceChange = 0;
        this._maxLinesTotal = 0;
        this._colorBk = 0xff000000;
        this._priority = 0;
        this._collisionMask = 0;
        this._player0 = new Player_1.default(31744, () => this._flushLineCache());
        this._player1 = new Player_1.default(17344, () => this._flushLineCache());
        this._missile0 = new Missile_1.default(8760, () => this._flushLineCache());
        this._missile1 = new Missile_1.default(4390, () => this._flushLineCache());
        this._playfield = new Playfield_1.default(1099, () => this._flushLineCache());
        this._ball = new Ball_1.default(2197, () => this._flushLineCache());
        this._waveformAudio = new Array(2);
        this._pcmAudio = null;
        this._audio = new Array(2);
        this._frameManager = new FrameManager_1.default(this._config);
        this._frameManager.newFrame.addHandler(Tia._onNewFrame, this);
        this._palette = this._getPalette(this._config);
        this._input0 = new LatchedInput_1.default(joystick0.getFire());
        this._input1 = new LatchedInput_1.default(joystick1.getFire());
        this._pcmAudio = new PCMAudio_1.default(this._config);
        const pcmChannels = this._pcmAudio.getChannels();
        for (let i = 0; i < 2; i++) {
            this._waveformAudio[i] = new WaveformAudio_1.default(this._config);
            this._audio[i] = this._config.pcmAudio ? pcmChannels[i] : this._waveformAudio[i];
        }
        const clockFreq = this._getClockFreq(this._config);
        this._paddles = new Array(4);
        for (let i = 0; i < 4; i++) {
            this._paddles[i] = new PaddleReader_1.default(clockFreq, paddles[i]);
        }
        this.reset();
    }
    reset() {
        this._hctr = 0;
        this._movementInProgress = false;
        this._extendedHblank = false;
        this._movementClock = 0;
        this._priority = 0;
        this._hstate = 0;
        this._collisionMask = 0;
        this._colorBk = 0xff000000;
        this._linesSinceChange = 0;
        this._collisionUpdateRequired = false;
        this._maxLinesTotal = 0;
        this._xDelta = 0;
        this._delayQueue.reset();
        this._frameManager.reset();
        this._missile0.reset();
        this._missile1.reset();
        this._player0.reset();
        this._player1.reset();
        this._playfield.reset();
        this._ball.reset();
        this._audio[0].reset();
        this._audio[1].reset();
        this._input0.reset();
        this._input1.reset();
        for (let i = 0; i < 4; i++) {
            this._paddles[i].reset();
        }
        if (this._cpu) {
            this._cpu.resume();
        }
    }
    setCpu(cpu) {
        this._cpu = cpu;
        return this;
    }
    setCpuTimeProvider(provider) {
        for (let i = 0; i < 4; i++) {
            this._paddles[i].setCpuTimeProvider(provider);
        }
        return this;
    }
    getWidth() {
        return 160;
    }
    getHeight() {
        return this._frameManager.getHeight();
    }
    setSurfaceFactory(factory) {
        this._frameManager.setSurfaceFactory(factory);
        return this;
    }
    getWaveformChannel(i) {
        return this._waveformAudio[i];
    }
    getPCMChannel() {
        return this._pcmAudio;
    }
    setAudioEnabled(state) {
        this._audio[0].setActive(state && this._config.enableAudio);
        this._audio[1].setActive(state && this._config.enableAudio);
    }
    read(address) {
        const lastDataBusValue = this._bus.getLastDataBusValue();
        let result;
        switch (address & 0x0f) {
            case 8:
                result = (this._config.emulatePaddles ? this._paddles[0].inpt() : 0) | (lastDataBusValue & 0x40);
                break;
            case 9:
                result = (this._config.emulatePaddles ? this._paddles[1].inpt() : 0) | (lastDataBusValue & 0x40);
                break;
            case 10:
                result = (this._config.emulatePaddles ? this._paddles[2].inpt() : 0) | (lastDataBusValue & 0x40);
                break;
            case 11:
                result = (this._config.emulatePaddles ? this._paddles[3].inpt() : 0) | (lastDataBusValue & 0x40);
                break;
            case 12:
                result = this._input0.inpt() | (lastDataBusValue & 0x40);
                break;
            case 13:
                result = this._input1.inpt() | (lastDataBusValue & 0x40);
                break;
            case 0:
                result =
                    (this._collisionMask & 8760 & 31744 ? 0x40 : 0) |
                        (this._collisionMask & 8760 & 17344 ? 0x80 : 0);
                break;
            case 1:
                result =
                    (this._collisionMask & 4390 & 17344 ? 0x40 : 0) |
                        (this._collisionMask & 4390 & 31744 ? 0x80 : 0);
                break;
            case 2:
                result =
                    (this._collisionMask & 31744 & 2197 ? 0x40 : 0) |
                        (this._collisionMask & 31744 & 1099 ? 0x80 : 0);
                break;
            case 3:
                result =
                    (this._collisionMask & 17344 & 2197 ? 0x40 : 0) |
                        (this._collisionMask & 17344 & 1099 ? 0x80 : 0);
                break;
            case 4:
                result =
                    (this._collisionMask & 8760 & 2197 ? 0x40 : 0) |
                        (this._collisionMask & 8760 & 1099 ? 0x80 : 0);
                break;
            case 5:
                result =
                    (this._collisionMask & 4390 & 2197 ? 0x40 : 0) |
                        (this._collisionMask & 4390 & 1099 ? 0x80 : 0);
                break;
            case 7:
                result =
                    (this._collisionMask & 8760 & 4390 ? 0x40 : 0) |
                        (this._collisionMask & 31744 & 17344 ? 0x80 : 0);
                break;
            case 6:
                result = this._collisionMask & 2197 & 1099 ? 0x80 : 0;
                break;
            default:
                result = 0;
                break;
        }
        return (result & 0xc0) | (lastDataBusValue & 0x3f);
    }
    peek(address) {
        return this.read(address);
    }
    write(address, value) {
        let v = 0;
        switch (address & 0x3f) {
            case 2:
                this._cpu.halt();
                break;
            case 3:
                this._flushLineCache();
                this._rsync();
                break;
            case 0:
                this._frameManager.setVsync((value & 0x02) > 0);
                break;
            case 1:
                this._input0.vblank(value);
                this._input1.vblank(value);
                for (let i = 0; i < 4; i++) {
                    this._paddles[i].vblank(value);
                }
                this._delayQueue.push(1, value, 1);
                break;
            case 29:
                this._delayQueue.push(29, value, 1);
                break;
            case 30:
                this._delayQueue.push(30, value, 1);
                break;
            case 34:
                this._delayQueue.push(34, value, 2);
                break;
            case 35:
                this._delayQueue.push(35, value, 2);
                break;
            case 18:
                this._flushLineCache();
                this._missile0.resm(this._resxCounter(), this._hstate === 0);
                break;
            case 19:
                this._flushLineCache();
                this._missile1.resm(this._resxCounter(), this._hstate === 0);
                break;
            case 40:
                this._missile0.resmp(value, this._player0);
                break;
            case 41:
                this._missile1.resmp(value, this._player1);
                break;
            case 43:
                this._delayQueue.push(43, value, 2);
                break;
            case 4:
                this._flushLineCache();
                this._missile0.nusiz(value);
                this._player0.nusiz(value, this._hstate === 0);
                break;
            case 5:
                this._flushLineCache();
                this._missile1.nusiz(value);
                this._player1.nusiz(value, this._hstate === 0);
                break;
            case 42:
                this._delayQueue.push(42, value, 6);
                break;
            case 9:
                this._flushLineCache();
                this._colorBk = this._palette[(value & 0xff) >>> 1];
                break;
            case 6:
                v = this._palette[(value & 0xff) >>> 1];
                this._missile0.setColor(v);
                this._player0.setColor(v);
                this._playfield.setColorP0(v);
                break;
            case 7:
                v = this._palette[(value & 0xff) >>> 1];
                this._missile1.setColor(v);
                this._player1.setColor(v);
                this._playfield.setColorP1(v);
                break;
            case 13:
                this._delayQueue.push(13, value, 2);
                break;
            case 14:
                this._delayQueue.push(14, value, 2);
                break;
            case 15:
                this._delayQueue.push(15, value, 2);
                break;
            case 10:
                this._setPriority(value);
                this._playfield.ctrlpf(value);
                this._ball.ctrlpf(value);
                break;
            case 8:
                this._flushLineCache();
                v = this._palette[(value & 0xff) >>> 1];
                this._playfield.setColor(v);
                this._ball.color = v;
                break;
            case 27:
                this._delayQueue
                    .push(27, value, 1)
                    .push(241, 0, 1);
                break;
            case 28:
                this._delayQueue
                    .push(28, value, 1)
                    .push(240, 0, 1)
                    .push(242, 0, 1);
                break;
            case 16:
                this._flushLineCache();
                this._player0.resp(this._resxCounter());
                break;
            case 17:
                this._flushLineCache();
                this._player1.resp(this._resxCounter());
                break;
            case 11:
                this._delayQueue.push(11, value, 1);
                break;
            case 12:
                this._delayQueue.push(12, value, 1);
                break;
            case 32:
                this._delayQueue.push(32, value, 2);
                break;
            case 33:
                this._delayQueue.push(33, value, 2);
                break;
            case 37:
                this._player0.vdelp(value);
                break;
            case 38:
                this._player1.vdelp(value);
                break;
            case 31:
                this._delayQueue.push(31, value, 1);
                break;
            case 36:
                this._delayQueue.push(36, value, 2);
                break;
            case 20:
                this._flushLineCache();
                this._ball.resbl(this._resxCounter());
                break;
            case 39:
                this._ball.vdelbl(value);
                break;
            case 44:
                this._flushLineCache();
                this._collisionMask = 0;
                break;
            case 21:
                this._audio[0].audc(value);
                break;
            case 22:
                this._audio[1].audc(value);
                break;
            case 23:
                this._audio[0].audf(value);
                break;
            case 24:
                this._audio[1].audf(value);
                break;
            case 25:
                this._audio[0].audv(value);
                break;
            case 26:
                this._audio[1].audv(value);
                break;
        }
    }
    getDebugState() {
        return ('' +
            `hclock: ${this._hctr}   line: ${this._frameManager.getCurrentLine()}\n` +
            this._frameManager.getDebugState());
    }
    setBus(bus) {
        this._bus = bus;
        return this;
    }
    cycle() {
        this._delayQueue.execute(Tia._delayedWrite, this);
        this._collisionUpdateRequired = false;
        if (this._linesSinceChange < 2) {
            this._tickMovement();
            if (this._hstate === 0) {
                this._tickHblank();
            }
            else {
                this._tickHframe();
            }
            if (this._collisionUpdateRequired && !this._frameManager.vblank) {
                this._updateCollision();
            }
        }
        else {
            if (this._hctr === 0) {
                this._cpu.resume();
            }
        }
        if (++this._hctr >= 228) {
            this._nextLine();
        }
        if (this._config.pcmAudio) {
            this._pcmAudio.tick();
        }
    }
    static _delayedWrite(address, value, self) {
        switch (address) {
            case 1:
                self._flushLineCache();
                self._frameManager.setVblank((value & 0x02) > 0);
                break;
            case 42:
                self._flushLineCache();
                self._movementClock = 0;
                self._movementInProgress = true;
                if (!self._extendedHblank) {
                    self._clearHmoveComb();
                    self._extendedHblank = true;
                }
                self._missile0.startMovement();
                self._missile1.startMovement();
                self._player0.startMovement();
                self._player1.startMovement();
                self._ball.startMovement();
                break;
            case 13:
                self._playfield.pf0(value);
                break;
            case 14:
                self._playfield.pf1(value);
                break;
            case 15:
                self._playfield.pf2(value);
                break;
            case 27:
                self._player0.grp(value);
                break;
            case 28:
                self._player1.grp(value);
                break;
            case 240:
                self._player0.shufflePatterns();
                break;
            case 241:
                self._player1.shufflePatterns();
                break;
            case 32:
                self._player0.hmp(value);
                break;
            case 33:
                self._player1.hmp(value);
                break;
            case 34:
                self._missile0.hmm(value);
                break;
            case 35:
                self._missile1.hmm(value);
                break;
            case 36:
                self._ball.hmbl(value);
                break;
            case 43:
                self._missile0.hmm(0);
                self._missile1.hmm(0);
                self._player0.hmp(0);
                self._player1.hmp(0);
                self._ball.hmbl(0);
                break;
            case 11:
                self._player0.refp(value);
                break;
            case 12:
                self._player1.refp(value);
                break;
            case 242:
                self._ball.shuffleStatus();
                break;
            case 31:
                self._ball.enabl(value);
                break;
            case 29:
                self._missile0.enam(value);
                break;
            case 30:
                self._missile1.enam(value);
                break;
        }
    }
    static _onNewFrame(surface, self) {
        const linesTotal = self._frameManager.getCurrentLine();
        if (linesTotal > self._maxLinesTotal) {
            self._maxLinesTotal = linesTotal;
        }
        if (linesTotal < self._maxLinesTotal) {
            const buffer = surface.getBuffer(), base = 160 * linesTotal, boundary = self._maxLinesTotal * 160;
            for (let i = base; i < boundary; i++) {
                buffer[i] = 0xff000000;
            }
        }
        self.newFrame.dispatch(surface);
    }
    _tickMovement() {
        if (!this._movementInProgress) {
            return;
        }
        if ((this._hctr & 0x3) === 0) {
            const apply = this._hstate === 0;
            let m = false;
            const movementCounter = this._movementClock > 15 ? 0 : this._movementClock;
            m = this._missile0.movementTick(movementCounter, apply) || m;
            m = this._missile1.movementTick(movementCounter, apply) || m;
            m = this._player0.movementTick(movementCounter, apply) || m;
            m = this._player1.movementTick(movementCounter, apply) || m;
            m = this._ball.movementTick(movementCounter, apply) || m;
            this._movementInProgress = m;
            this._collisionUpdateRequired = m;
            this._movementClock++;
        }
    }
    _tickHblank() {
        switch (this._hctr) {
            case 0:
                this._extendedHblank = false;
                this._cpu.resume();
                break;
            case 67:
                if (!this._extendedHblank) {
                    this._hstate = 1;
                }
                break;
            case 75:
                if (this._extendedHblank) {
                    this._hstate = 1;
                }
                break;
        }
        if (this._extendedHblank && this._hctr > 67) {
            this._playfield.tick(this._hctr - 68 + this._xDelta);
        }
    }
    _tickHframe() {
        const y = this._frameManager.getCurrentLine(), x = this._hctr - 68 + this._xDelta;
        this._collisionUpdateRequired = true;
        this._playfield.tick(x);
        this._tickSprites();
        if (this._frameManager.isRendering()) {
            this._renderPixel(x, y);
        }
    }
    _tickSprites() {
        this._missile0.tick(true);
        this._missile1.tick(true);
        this._player0.tick();
        this._player1.tick();
        this._ball.tick(true);
    }
    _nextLine() {
        if (this._linesSinceChange >= 2) {
            this._cloneLastLine();
        }
        this._hctr = 0;
        this._playfield.tick(0);
        if (!this._movementInProgress) {
            this._linesSinceChange++;
        }
        this._hstate = 0;
        this._xDelta = 0;
        this._frameManager.nextLine();
        if (this._frameManager.isRendering() && this._frameManager.getCurrentLine() === 0) {
            this._flushLineCache();
        }
    }
    _cloneLastLine() {
        const y = this._frameManager.getCurrentLine();
        if (!this._frameManager.isRendering() || y === 0) {
            return;
        }
        const delta = y * 160, prevDelta = (y - 1) * 160;
        for (let x = 0; x < 160; x++) {
            this._frameManager.surfaceBuffer[delta + x] = this._frameManager.surfaceBuffer[prevDelta + x];
        }
    }
    _getPalette(config) {
        switch (config.tvMode) {
            case 0:
                return palette.NTSC;
            case 1:
                return palette.PAL;
            case 2:
                return palette.SECAM;
            default:
                throw new Error('invalid TV mode');
        }
    }
    _getClockFreq(config) {
        return config.tvMode === 0
            ? 60 * 228 * 262
            : 50 * 228 * 312;
    }
    _renderPixel(x, y) {
        if (this._frameManager.vblank) {
            this._frameManager.surfaceBuffer[y * 160 + x] = 0xff000000;
            return;
        }
        let color = this._colorBk;
        switch (this._priority) {
            case 0:
                color = this._playfield.getPixel(color);
                color = this._ball.getPixel(color);
                color = this._missile1.getPixel(color);
                color = this._player1.getPixel(color);
                color = this._missile0.getPixel(color);
                color = this._player0.getPixel(color);
                break;
            case 1:
                color = this._missile1.getPixel(color);
                color = this._player1.getPixel(color);
                color = this._missile0.getPixel(color);
                color = this._player0.getPixel(color);
                color = this._playfield.getPixel(color);
                color = this._ball.getPixel(color);
                break;
            case 2:
                color = this._ball.getPixel(color);
                color = this._missile1.getPixel(color);
                color = this._player1.getPixel(color);
                color = this._playfield.getPixel(color);
                color = this._missile0.getPixel(color);
                color = this._player0.getPixel(color);
                break;
            default:
                throw new Error('invalid priority');
        }
        this._frameManager.surfaceBuffer[y * 160 + x] = color;
    }
    _updateCollision() {
        this._collisionMask |=
            ~this._player0.collision &
                ~this._player1.collision &
                ~this._missile0.collision &
                ~this._missile1.collision &
                ~this._ball.collision &
                ~this._playfield.collision;
    }
    _clearHmoveComb() {
        if (this._frameManager.isRendering() && this._hstate === 0) {
            const offset = this._frameManager.getCurrentLine() * 160;
            for (let i = 0; i < 8; i++) {
                this._frameManager.surfaceBuffer[offset + i] = 0xff000000;
            }
        }
    }
    _resxCounter() {
        return this._hstate === 0
            ? this._hctr >= 73
                ? 158
                : 159
            : 157;
    }
    _rsync() {
        const x = this._hctr > 68 ? this._hctr - 68 : 0;
        this._xDelta = 157 - x;
        if (this._frameManager.isRendering()) {
            const y = this._frameManager.getCurrentLine(), base = y * 160 + x, boundary = base + (y + 1) * 160;
            for (let i = base; i < boundary; i++) {
                this._frameManager.surfaceBuffer[i] = 0xff000000;
            }
        }
        this._hctr = 225;
    }
    _setPriority(value) {
        const priority = value & 0x04 ? 1 : value & 0x02 ? 2 : 0;
        if (priority !== this._priority) {
            this._flushLineCache();
            this._priority = priority;
        }
    }
    _flushLineCache() {
        const wasCaching = this._linesSinceChange >= 2;
        this._linesSinceChange = 0;
        if (wasCaching) {
            const rewindCycles = this._hctr;
            for (this._hctr = 0; this._hctr < rewindCycles; this._hctr++) {
                if (this._hstate === 0) {
                    this._tickHblank();
                }
                else {
                    this._tickHframe();
                }
            }
        }
    }
}
exports.default = Tia;
(function (Tia) {
    class TrapPayload {
        constructor(reason, tia, message) {
            this.reason = reason;
            this.tia = tia;
            this.message = message;
        }
    }
    Tia.TrapPayload = TrapPayload;
})(Tia || (Tia = {}));
exports.default = Tia;
//# sourceMappingURL=Tia.js.map