"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const microevent_ts_1 = require("microevent.ts");
const BoardInterface_1 = tslib_1.__importDefault(require("../board/BoardInterface"));
const StateMachineCpu_1 = tslib_1.__importDefault(require("../cpu/StateMachineCpu"));
const Memory_1 = tslib_1.__importDefault(require("./Memory"));
class Board {
    constructor(cpuFactory) {
        this.clock = new microevent_ts_1.Event();
        this.trap = new microevent_ts_1.Event();
        this.systemReset = new microevent_ts_1.Event();
        this._trap = false;
        this._clockMode = 1;
        this._timer = {
            tick: (clocks) => this._tick(clocks),
            start: (scheduler, sliceHint) => this._start(scheduler, sliceHint),
            stop: () => this._stop(),
            isRunning: () => !!this._runTask
        };
        this.cpuClock = this.clock;
        this._bus = this._createBus();
        if (typeof cpuFactory === 'undefined') {
            cpuFactory = bus => new StateMachineCpu_1.default(bus);
        }
        this._cpu = cpuFactory(this._bus);
        this._cpu.setInvalidInstructionCallback(() => this._onInvalidInstruction());
    }
    getCpu() {
        return this._cpu;
    }
    getBus() {
        return this._bus;
    }
    getTimer() {
        return this._timer;
    }
    reset(hard) {
        this._cpu.reset();
        this._bus.reset();
        if (hard) {
            this._bus.clear();
        }
        this.systemReset.dispatch();
        return this;
    }
    boot() {
        let clock = 0;
        if (this._cpu.executionState !== 0) {
            throw new Error('Already booted!');
        }
        while (this._cpu.executionState !== 1) {
            this._cpu.cycle();
            clock++;
        }
        this.clock.dispatch(clock);
        return this;
    }
    suspend() { }
    resume() { }
    triggerTrap(reason, message) {
        this._stop();
        this._trap = true;
        if (this.trap.hasHandlers) {
            this.trap.dispatch(new BoardInterface_1.default.TrapPayload(reason, this, message));
        }
        else {
            throw new Error(message);
        }
        return this;
    }
    getBoardStateDebug() {
        return undefined;
    }
    setClockMode(clockMode) {
        this._clockMode = clockMode;
        return this;
    }
    getClockMode() {
        return this._clockMode;
    }
    _createBus() {
        return new Memory_1.default();
    }
    _tick(clocks) {
        let i = 0, clock = 0;
        this._trap = false;
        while (i++ < clocks && !this._trap) {
            this._cpu.cycle();
            clock++;
            if (this._clockMode === 0 &&
                this._cpu.executionState === 1 &&
                this.clock.hasHandlers) {
                this.clock.dispatch(clock);
                clock = 0;
            }
        }
        if (clock > 0 && this.clock.hasHandlers) {
            this.clock.dispatch(clock);
        }
        return clock;
    }
    _start(scheduler, sliceHint = 100000) {
        if (this._runTask) {
            return;
        }
        this._sliceHint = sliceHint;
        this._runTask = scheduler.start(this._executeSlice, this);
    }
    _executeSlice(board) {
        board._tick(board._sliceHint);
    }
    _stop() {
        if (!this._runTask) {
            return;
        }
        this._runTask.stop();
        this._runTask = undefined;
    }
    _onInvalidInstruction() {
        this.triggerTrap(0, 'invalid instruction');
    }
}
exports.default = Board;
//# sourceMappingURL=Board.js.map