"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const fs_1 = tslib_1.__importDefault(require("fs"));
const dasm_1 = tslib_1.__importDefault(require("dasm"));
const CartridgeFactory_1 = tslib_1.__importDefault(require("../machine/stella/cartridge/CartridgeFactory"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("../machine/stella/cartridge/CartridgeInfo"));
const Config_1 = tslib_1.__importDefault(require("../machine/stella/Config"));
const Board_1 = tslib_1.__importDefault(require("../machine/stella/Board"));
const Instruction_1 = tslib_1.__importDefault(require("../machine/cpu/Instruction"));
class VcsRunner {
    constructor() {
        this._defaultCycles = 500000;
        this._traps = [];
        this._cpuClocks = 0;
        this._colorClocks = 0;
    }
    static fromSource(source, config = {}) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const self = new VcsRunner();
            config = Object.assign({ defaultCycles: self._defaultCycles }, config);
            self._defaultCycles = config.defaultCycles;
            const result = dasm_1.default(source, {
                format: 3,
                machine: 'atari2600',
                includes: config.includes,
            });
            if (!result.success) {
                throw new Error(`assembly failed: ${result.output}`);
            }
            self._assembly = result.data;
            self._symbols = new Map();
            result.symbols.forEach((s) => self._symbols.set(s.name, s.value | (s.isLabel ? 0xf000 : 0)));
            const factory = new CartridgeFactory_1.default();
            self._cartridge = yield factory.createCartridge(self._assembly, CartridgeInfo_1.default.CartridgeType.vanilla_4k);
            self._board = new Board_1.default(Config_1.default.create(config.stellaConfig), self._cartridge);
            return self;
        });
    }
    static fromFile(fileName, config) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this.fromSource(fs_1.default.readFileSync(fileName, 'utf-8'), config);
        });
    }
    boot() {
        return this.runUntil(() => this._board.getCpu().executionState !== 0);
    }
    cld() {
        this.getBoard().getCpu().state.flags &= ~8;
        return this;
    }
    modifyCpuState(mapping) {
        Object.assign(this.getBoard().getCpu().state, mapping(this.getBoard().getCpu().state));
        return this;
    }
    getCpuState() {
        return this.getBoard().getCpu().state;
    }
    runUntil(condition, maxCycles = this._defaultCycles, stepping = VcsRunner.Stepping.instruction) {
        for (let i = 0; i < maxCycles; i++) {
            switch (stepping) {
                case VcsRunner.Stepping.colorClock:
                    this._tick();
                    break;
                case VcsRunner.Stepping.cpuClock:
                    do {
                        this._tick();
                    } while (this._board.getSubclock() !== 0);
                    break;
                case VcsRunner.Stepping.instruction:
                    do {
                        this._tick();
                    } while (this._board.getSubclock() !== 0 ||
                        this._board.getCpu().executionState !== 1);
                    break;
            }
            this._traps.forEach((trap) => trap.condition(this) && trap.handler());
            if (condition(this)) {
                return this;
            }
        }
        throw new Error(`break condition not met after ${maxCycles} cycles`);
    }
    runTo(label, maxCycles) {
        return this.runUntil(() => this.hasReachedLabel(label), maxCycles, VcsRunner.Stepping.instruction);
    }
    runToRts(maxCycles) {
        return this.runUntil(() => this.nextInstruction().operation === 42, maxCycles, VcsRunner.Stepping.instruction);
    }
    hasReachedLabel(label) {
        return (this._board.getCpu().state.p & 0x1fff) === (this._resolveLabel(label) & 0x1fff);
    }
    nextInstruction() {
        return Instruction_1.default.opcodes[this.getBoard().getBus().peek(this.getBoard().getCpu().state.p)];
    }
    jumpTo(label) {
        this.getBoard().getCpu().state.p = this._resolveLabel(label) & 0xffff;
        return this;
    }
    getBoard() {
        return this._board;
    }
    getCpuCycles() {
        return this._cpuClocks;
    }
    getColorClocks() {
        return this._colorClocks;
    }
    readMemory(address) {
        return this.getBoard().getBus().read(address);
    }
    writeMemoryAt(label, value) {
        this.getBoard().getBus().write(this._resolveLabel(label), value);
        return this;
    }
    readMemoryAt(label) {
        return this.getBoard().getBus().peek(this._resolveLabel(label));
    }
    trap(condition, handler) {
        this._traps.push({ condition, handler });
        return this;
    }
    trapAt(label, handler) {
        return this.trap(() => this.hasReachedLabel(label), handler);
    }
    trapAlways(handler) {
        return this.trap(() => true, handler);
    }
    _resolveLabel(label) {
        if (!this._symbols.has(label)) {
            throw new Error(`invalid label ${label}`);
        }
        return this._symbols.get(label) & 0xffff;
    }
    _tick() {
        this._board.tick(1);
        this._colorClocks++;
        if (this._board.getSubclock() === 0) {
            this._cpuClocks++;
        }
    }
}
(function (VcsRunner) {
    let Stepping;
    (function (Stepping) {
        Stepping[Stepping["instruction"] = 0] = "instruction";
        Stepping[Stepping["cpuClock"] = 1] = "cpuClock";
        Stepping[Stepping["colorClock"] = 2] = "colorClock";
    })(Stepping = VcsRunner.Stepping || (VcsRunner.Stepping = {}));
})(VcsRunner || (VcsRunner = {}));
exports.default = VcsRunner;
//# sourceMappingURL=VcsRunner.js.map