"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.auxTargets = exports.joystickTargets = void 0;
const tslib_1 = require("tslib");
const microevent_ts_1 = require("microevent.ts");
const ShadowSwitch_1 = tslib_1.__importDefault(require("./gamepad/ShadowSwitch"));
const defaultMapping_1 = require("./gamepad/defaultMapping");
const MIN_POLL_INTERVAL = 50;
exports.joystickTargets = [
    "left",
    "right",
    "up",
    "down",
    "fire",
    "start",
    "select",
    "pause"
];
exports.auxTargets = ["start", "select", "pause"];
function readButton(button) {
    return typeof button === 'object' ? button.pressed : button > 0.5;
}
class GamepadDriver {
    constructor() {
        this._onGamepadConnect = () => this.probeGamepads();
        this._onGamepadDisconnect = () => this.probeGamepads();
        this.gamepadCountChanged = new microevent_ts_1.Event();
        this._shadows = null;
        this._mappings = new Map();
        this._mappingStates = new WeakMap();
        this._mappingTargets = new WeakMap();
        this._bound = false;
        this._joysticks = [];
        this._auxSwitches = {};
        this._gamepadCount = 0;
        this._lastPoll = 0;
        this.setMapping(defaultMapping_1.defaultMapping);
    }
    static probeGamepadCount() {
        let cnt = 0;
        const gamepads = navigator.getGamepads();
        for (let i = 0; i < gamepads.length; i++) {
            if (gamepads[i]) {
                cnt++;
            }
        }
        return cnt;
    }
    init() {
        if (!navigator.getGamepads) {
            throw new Error(`gamepad API not available`);
        }
        this.probeGamepads();
        window.addEventListener('gamepadconnected', this._onGamepadConnect);
        window.addEventListener('gamepaddisconnected', this._onGamepadDisconnect);
    }
    deinit() {
        this.unbind();
        window.removeEventListener('gamepadconnected', this._onGamepadConnect);
        window.removeEventListener('gamepaddisconnected', this._onGamepadDisconnect);
    }
    bind(joysticks = [], auxSwitches = {}) {
        if (this._bound) {
            return;
        }
        this._bound = true;
        this._joysticks = joysticks;
        this._auxSwitches = auxSwitches;
        this._bound = true;
        this._shadows = new WeakMap();
        this._controlledSwitches().forEach(swtch => {
            const shadow = new ShadowSwitch_1.default();
            this._shadows.set(swtch, shadow);
            shadow.setState(swtch.read());
            swtch.beforeRead.addHandler(GamepadDriver._onBeforeSwitchRead, this);
        });
    }
    unbind() {
        if (!this._bound) {
            return;
        }
        this._controlledSwitches().forEach(swtch => swtch.beforeRead.removeHandler(GamepadDriver._onBeforeSwitchRead, this));
        this._shadows = null;
        this._auxSwitches = {};
        this._joysticks = [];
        this._bound = false;
    }
    getGamepadCount() {
        return this._gamepadCount;
    }
    setMapping(mapping, id) {
        if (typeof id !== 'undefined') {
            this._mappings.set(id, mapping);
        }
        const states = new Map();
        const targets = [];
        for (const m of mapping) {
            if (targets.indexOf(m.target)) {
                targets.push(m.target);
                states.set(m.target, false);
            }
        }
        this._mappingStates.set(mapping, states);
        this._mappingTargets.set(mapping, targets);
    }
    clearMapping(id) {
        this._mappings.delete(id);
    }
    poll() {
        this._readGamepads();
    }
    static _onBeforeSwitchRead(swtch, self) {
        self.poll();
    }
    probeGamepads() {
        const cnt = GamepadDriver.probeGamepadCount();
        if (cnt !== this._gamepadCount) {
            this._gamepadCount = cnt;
            this.gamepadCountChanged.dispatch(this._gamepadCount);
        }
    }
    _getSwitchForTarget(target, joystick = null) {
        if (this._auxSwitches[target]) {
            return this._auxSwitches[target];
        }
        if (!joystick) {
            return null;
        }
        switch (target) {
            case "up":
                return joystick.getUp();
            case "down":
                return joystick.getDown();
            case "left":
                return joystick.getLeft();
            case "right":
                return joystick.getRight();
            case "fire":
                return joystick.getFire();
            default:
                return null;
        }
    }
    _controlledSwitches() {
        const switches = [
            ...Object.keys(this._auxSwitches).map(target => this._auxSwitches[target])
        ];
        for (const joystick of this._joysticks) {
            switches.push(joystick.getLeft(), joystick.getRight(), joystick.getUp(), joystick.getDown(), joystick.getFire());
        }
        return switches;
    }
    _readGamepads() {
        const now = Date.now();
        if (this._gamepadCount === 0 || now - this._lastPoll < MIN_POLL_INTERVAL) {
            return;
        }
        this._lastPoll = now;
        const gamepads = navigator.getGamepads();
        let joystickIndex = 0;
        for (let i = 0; i < gamepads.length; i++) {
            const gamepad = gamepads[i];
            if (!gamepad) {
                continue;
            }
            const mapping = this._mappings.get(gamepad.id) || defaultMapping_1.defaultMapping;
            const states = this._mappingStates.get(mapping);
            const targets = this._mappingTargets.get(mapping);
            for (const target of targets) {
                states.set(target, false);
            }
            for (const mappingEntry of mapping) {
                switch (mappingEntry.type) {
                    case "button":
                        const button = gamepad.buttons[mappingEntry.index];
                        if (typeof button !== 'undefined' && readButton(button)) {
                            states.set(mappingEntry.target, true);
                        }
                        break;
                    case "axis":
                        const axis = gamepad.axes[mappingEntry.index];
                        if (typeof axis !== 'undefined' &&
                            (mappingEntry.sign === "positive" ? axis > 0.5 : axis < -0.5)) {
                            states.set(mappingEntry.target, true);
                        }
                        break;
                }
            }
            for (const target of targets) {
                const swtch = this._getSwitchForTarget(target, this._joysticks[joystickIndex]);
                if (!swtch) {
                    continue;
                }
                const shadow = this._shadows.get(swtch);
                shadow.toggle(states.get(target));
                shadow.sync(swtch);
            }
            joystickIndex++;
        }
    }
}
exports.default = GamepadDriver;
//# sourceMappingURL=Gamepad.js.map