"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const shader_1 = require("./video/shader");
const Program_1 = tslib_1.__importDefault(require("./video/Program"));
const PhosphorProcessor_1 = tslib_1.__importDefault(require("./video/PhosphorProcessor"));
const NtscProcessor_1 = tslib_1.__importDefault(require("./video/NtscProcessor"));
const ScanlineProcessor_1 = tslib_1.__importDefault(require("./video/ScanlineProcessor"));
const IntegerScalingProcessor_1 = tslib_1.__importDefault(require("./video/IntegerScalingProcessor"));
const RingBuffer_1 = tslib_1.__importDefault(require("../../tools/RingBuffer"));
const Capabilities_1 = require("./video/Capabilities");
const MAX_CONSECUTIVE_UNDERFLOWS = 5;
class Video {
    constructor(_canvas, config = {}) {
        this._canvas = _canvas;
        this._onContextLost = (e) => {
            e.preventDefault();
            this._gl = null;
            this._mainProgram = null;
            this._vertexCoordinateBuffer = null;
            this._textureCoordinateBuffer = null;
            this._sourceTexture = null;
            this._phosphorProcessor = null;
            this._ntscProcessor = null;
            this._integerScalingProcessor = null;
            this._scanlineProcessor = null;
            this._processors = [];
        };
        this._onContextRestored = () => {
            this._getRenderingContext();
            this.init();
        };
        this._config = null;
        this._gl = null;
        this._video = null;
        this._capabilities = null;
        this._mainProgram = null;
        this._vertexCoordinateBuffer = null;
        this._textureCoordinateBuffer = null;
        this._sourceTexture = null;
        this._anmiationFrameHandle = 0;
        this._phosphorProcessor = null;
        this._ntscProcessor = null;
        this._scanlineProcessor = null;
        this._integerScalingProcessor = null;
        this._processors = [];
        this._pendingFrames = new RingBuffer_1.default(3);
        this._consecutiveUnderflows = 0;
        this._hasFrame = false;
        const defaultConfig = {
            gamma: 1,
            scalingMode: "qis",
            phosphorLevel: 0.5,
            scanlineLevel: 0.2,
            tvEmulation: "composite"
        };
        this._config = Object.assign(Object.assign({}, defaultConfig), config);
        this._getRenderingContext();
        this._canvas.addEventListener('webglcontextlost', this._onContextLost);
        this._canvas.addEventListener('webglcontextrestored', this._onContextRestored);
        this._pendingFrames.evict.addHandler(frame => frame.release());
    }
    init() {
        if (!this._gl)
            return this;
        this._phosphorProcessor = new PhosphorProcessor_1.default(this._gl, this._capabilities);
        this._ntscProcessor = new NtscProcessor_1.default(this._gl, this._capabilities);
        this._scanlineProcessor = new ScanlineProcessor_1.default(this._gl, this._capabilities);
        this._integerScalingProcessor = new IntegerScalingProcessor_1.default(this._gl, this._capabilities);
        this._updateCanvasSize();
        this._mainProgram = Program_1.default.compile(this._gl, shader_1.vsh.plain.source(this._capabilities), shader_1.fsh.blitWithGamma.source(this._capabilities));
        this._mainProgram.use();
        this._mainProgram.uniform1i("u_Sampler0", 0);
        this._createVertexCoordinateBuffer();
        this._createTextureCoordinateBuffer();
        this._configureSourceTexture();
        this._applyConfiguration();
        return this;
    }
    close() {
        const gl = this._gl;
        this._cancelDraw();
        this.unbind();
        if (gl) {
            this._mainProgram.delete();
            gl.deleteBuffer(this._vertexCoordinateBuffer);
            gl.deleteBuffer(this._textureCoordinateBuffer);
            gl.deleteTexture(this._sourceTexture);
            this._ntscProcessor.destroy();
            this._phosphorProcessor.destroy();
            this._scanlineProcessor.destroy();
            this._integerScalingProcessor.destroy();
            return this;
        }
        this._canvas.removeEventListener('webglcontextlost', this._onContextLost);
        this._canvas.removeEventListener('webglcontextrestored', this._onContextRestored);
        this._pendingFrames.forEach(f => f.release());
    }
    resize(width, height) {
        this._updateCanvasSize(width, height);
        this._updateVertexBuffer();
        if (this._video) {
            this._configureProcessors();
            this._scheduleDraw();
        }
        if (this._hasFrame)
            this._draw();
        return this;
    }
    getCanvas() {
        return this._canvas;
    }
    bind(video) {
        if (this._video) {
            return this;
        }
        this._video = video;
        this._video.newFrame.addHandler(Video._frameHandler, this);
        this._configureProcessors();
        this._scheduleDraw();
        return this;
    }
    unbind() {
        if (!this._video) {
            return this;
        }
        this._cancelDraw();
        this._video.newFrame.removeHandler(Video._frameHandler, this);
        this._video = null;
        this._pendingFrames.forEach(f => f.release());
        this._pendingFrames.clear();
        return this;
    }
    getConfig() {
        return this._config;
    }
    updateConfig(config) {
        return this;
    }
    static _frameHandler(imageDataPoolMember, self) {
        if (self._pendingFrames.size() === self._pendingFrames.capacity()) {
            self._pendingFrames.forEach(f => f.release());
            self._pendingFrames.clear();
        }
        self._pendingFrames.push(imageDataPoolMember);
        self._scheduleDraw();
    }
    _getRenderingContext() {
        const contextOptions = {
            alpha: false,
            depth: false,
            antialias: false
        };
        this._gl =
            this._canvas.getContext('webgl', contextOptions) ||
                this._canvas.getContext('experimental-webgl', contextOptions);
        if (!this._gl) {
            throw new Error('unable to acquire webgl context');
        }
        this._capabilities = Capabilities_1.detect(this._gl);
    }
    _scheduleDraw() {
        if (this._anmiationFrameHandle !== 0) {
            return;
        }
        this._anmiationFrameHandle = requestAnimationFrame(() => this._onAnimationFrame());
    }
    _cancelDraw() {
        if (this._anmiationFrameHandle !== 0) {
            cancelAnimationFrame(this._anmiationFrameHandle);
        }
        this._anmiationFrameHandle = 0;
    }
    _onAnimationFrame() {
        const gl = this._gl;
        this._anmiationFrameHandle = 0;
        if (this._pendingFrames.size() === 0) {
            if (this._consecutiveUnderflows++ <= MAX_CONSECUTIVE_UNDERFLOWS) {
                this._scheduleDraw();
            }
            return;
        }
        this._consecutiveUnderflows = 0;
        this._scheduleDraw();
        const frame = this._pendingFrames.pop();
        if (!gl)
            return;
        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, this._sourceTexture);
        gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, frame.get());
        frame.release();
        this._hasFrame = true;
        this._draw();
    }
    _draw() {
        if (!this._gl)
            return;
        const gl = this._gl;
        let texture = this._sourceTexture;
        for (const processor of this._processors) {
            processor.render(texture);
            texture = processor.getTexture();
        }
        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, this._config.scalingMode === "none" ? gl.NEAREST : gl.LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, this._config.scalingMode === "none" ? gl.NEAREST : gl.LINEAR);
        this._mainProgram.use();
        this._mainProgram.bindVertexAttribArray("a_VertexPosition", this._vertexCoordinateBuffer, 2, gl.FLOAT, false, 0, 0);
        this._mainProgram.bindVertexAttribArray("a_TextureCoordinate", this._textureCoordinateBuffer, 2, gl.FLOAT, false, 0, 0);
        gl.bindFramebuffer(gl.FRAMEBUFFER, null);
        gl.viewport(0, 0, gl.drawingBufferWidth, gl.drawingBufferHeight);
        gl.clearColor(0, 0, 0, 1);
        gl.clear(gl.COLOR_BUFFER_BIT);
        gl.drawArrays(gl.TRIANGLE_STRIP, 0, 4);
    }
    _createVertexCoordinateBuffer() {
        if (!this._gl)
            return;
        const gl = this._gl, targetWidth = gl.drawingBufferWidth, targetHeight = gl.drawingBufferHeight, scaleX = targetWidth > 0 ? 2 / targetWidth : 1, scaleY = targetHeight > 0 ? 2 / targetHeight : 1;
        let width, height, west, north;
        if ((4 / 3) * targetHeight <= targetWidth) {
            height = 2;
            width = (4 / 3) * targetHeight * scaleX;
            north = 1;
            west = (Math.floor((-4 / 3) * targetHeight) / 2) * scaleX;
        }
        else {
            height = (targetWidth / (4 / 3)) * scaleY;
            width = 2;
            north = (Math.floor(targetWidth / (4 / 3)) / 2) * scaleY;
            west = -1;
        }
        const vertexData = [west + width, north, west, north, west + width, north - height, west, north - height];
        this._vertexCoordinateBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this._vertexCoordinateBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(vertexData), gl.STATIC_DRAW);
    }
    _updateVertexBuffer() {
        if (!this._gl)
            return;
        this._gl.deleteBuffer(this._vertexCoordinateBuffer);
        this._createVertexCoordinateBuffer();
    }
    _createTextureCoordinateBuffer() {
        if (!this._gl)
            return;
        const gl = this._gl;
        const textureCoordinateData = [1, 1, 0, 1, 1, 0, 0, 0];
        this._textureCoordinateBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this._textureCoordinateBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(textureCoordinateData), gl.STATIC_DRAW);
    }
    _updateCanvasSize(width, height) {
        if (typeof width === 'undefined' || typeof height === 'undefined') {
            width = this.getCanvas().clientWidth;
            height = this.getCanvas().clientHeight;
        }
        const pixelRatio = window.devicePixelRatio || 1;
        this.getCanvas().width = width * pixelRatio;
        this.getCanvas().height = height * pixelRatio;
    }
    _configureSourceTexture() {
        if (!this._gl)
            return;
        const gl = this._gl;
        if (!this._sourceTexture) {
            this._sourceTexture = gl.createTexture();
        }
        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, this._sourceTexture);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
        gl.pixelStorei(gl.UNPACK_FLIP_Y_WEBGL, 1);
    }
    _configureProcessors() {
        if (!this._video || !this._gl)
            return;
        this._phosphorProcessor.configure(this._config.phosphorLevel);
        this._scanlineProcessor.configure(this._config.scanlineLevel);
        this._integerScalingProcessor.configure(this._gl.drawingBufferWidth, this._gl.drawingBufferHeight);
        switch (this._config.tvEmulation) {
            case "composite":
                this._ntscProcessor.configure("composite");
                break;
            case "svideo":
                this._ntscProcessor.configure("svideo");
                break;
        }
        let width = this._video.getWidth();
        let height = this._video.getHeight();
        for (const processor of this._processors) {
            processor.resize(width, height);
            width = processor.getWidth();
            height = processor.getHeight();
        }
    }
    _applyConfiguration() {
        if (!this._gl)
            return;
        this._mainProgram.use();
        this._mainProgram.uniform1f("u_Gamma", this._config.gamma);
        this._processors = [];
        if (this._config.tvEmulation !== "none") {
            this._ntscProcessor.init();
            this._processors.push(this._ntscProcessor);
        }
        if (this._config.phosphorLevel > 0) {
            this._phosphorProcessor.init();
            this._processors.push(this._phosphorProcessor);
        }
        if (this._config.scanlineLevel > 0) {
            this._scanlineProcessor.init();
            this._processors.push(this._scanlineProcessor);
        }
        if (this._config.scalingMode === "qis") {
            this._integerScalingProcessor.init();
            this._processors.push(this._integerScalingProcessor);
        }
        this._configureProcessors();
    }
}
exports.default = Video;
//# sourceMappingURL=Video.js.map