"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const async_mutex_1 = require("async-mutex");
const WaveformChannel_1 = tslib_1.__importDefault(require("./audio/WaveformChannel"));
const PCMChannel_1 = tslib_1.__importDefault(require("./audio/PCMChannel"));
const browser_1 = require("../../tools/browser");
const audioNeedsInteraction = browser_1.isIOS || browser_1.isSafari;
const INTERACTION_EVENTS = ['touchstart', 'click', 'keydown'];
const audioContextCtor = window.AudioContext || window.webkitAudioContext;
function waitMax(waitFor, timeout = 500) {
    return new Promise((r) => {
        waitFor.then(() => r());
        setTimeout(r, timeout);
    });
}
class PreallocatedContext {
    constructor() {
        this._interactionListener = () => {
            const context = this.context;
            this.interactionRequired = false;
            INTERACTION_EVENTS.forEach((event) => document.removeEventListener(event, this._interactionListener));
            this.mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
                yield waitMax(context.resume());
                return new Promise((r) => setTimeout(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
                    yield waitMax(context.suspend());
                    r(undefined);
                }), 100));
            }));
        };
        this.mutex = new async_mutex_1.Mutex();
        this.context = null;
        this.interactionRequired = true;
        if (!audioContextCtor) {
            return;
        }
        this.context = new audioContextCtor();
        try {
            this.context.destination.channelCount = 1;
        }
        catch (e) {
            console.warn('audio driver: failed to set channel count');
        }
        INTERACTION_EVENTS.forEach((event) => document.addEventListener(event, this._interactionListener));
    }
    stopListening() {
        INTERACTION_EVENTS.forEach((event) => document.removeEventListener(event, this._interactionListener));
    }
}
let preallocatedContext = audioNeedsInteraction ? new PreallocatedContext() : null;
class WebAudioDriver {
    constructor(waveformChannels = 0, pcmChannels = 0, fragmentSize) {
        this._touchListener = () => {
            INTERACTION_EVENTS.forEach((event) => document.removeEventListener(event, this._touchListener, true));
            if (!this._context) {
                return;
            }
            this._context.resume();
            setTimeout(() => {
                this._mutex.runExclusive(() => (this._suspended ? this._context.suspend() : this._context.resume()));
            }, 10);
        };
        this._context = null;
        this._merger = null;
        this._waveformChannels = null;
        this._pcmChannels = null;
        this._channels = null;
        this._cache = new Map();
        this._mutex = new async_mutex_1.Mutex();
        this._suspended = true;
        this._isBound = false;
        this._waveformChannels = new Array(waveformChannels);
        this._pcmChannels = new Array(pcmChannels);
        for (let i = 0; i < waveformChannels; i++) {
            this._waveformChannels[i] = new WaveformChannel_1.default(this._cache);
        }
        for (let i = 0; i < pcmChannels; i++) {
            this._pcmChannels[i] = new PCMChannel_1.default(fragmentSize);
        }
        this._channels = [...this._waveformChannels, ...this._pcmChannels];
    }
    init() {
        if (preallocatedContext) {
            const p = preallocatedContext;
            preallocatedContext = new PreallocatedContext();
            this._context = p.context;
            p.stopListening();
            this._mutex = p.mutex;
            if (p.interactionRequired) {
                INTERACTION_EVENTS.forEach((event) => document.addEventListener(event, this._touchListener, true));
            }
        }
        else {
            if (!audioContextCtor) {
                throw new Error(`web audio is not supported by runtime`);
            }
            this._context = new audioContextCtor();
            try {
                this._context.destination.channelCount = 1;
            }
            catch (e) {
                console.warn('audio driver: failed to set channel count');
            }
        }
        this._merger = this._context.createChannelMerger(this._channels.length);
        this._merger.connect(this._context.destination);
        this._channels.forEach((channel) => channel.init(this._context, this._merger));
    }
    bind(waveformSources = [], pcmSources = []) {
        if (this._isBound) {
            return;
        }
        if (waveformSources.length !== this._waveformChannels.length) {
            throw new Error(`invalid number of waveform sources: expected ${this._waveformChannels.length}, got ${waveformSources.length}`);
        }
        if (pcmSources.length !== this._pcmChannels.length) {
            throw new Error(`invalid number of waveform sources: expected ${this._pcmChannels.length}, got ${pcmSources.length}`);
        }
        this._waveformChannels.forEach((channel, i) => channel.bind(waveformSources[i]));
        this._pcmChannels.forEach((channel, i) => channel.bind(pcmSources[i]));
        this._isBound = true;
        this.resume();
    }
    unbind() {
        if (!this._isBound) {
            return;
        }
        this._channels.forEach((channel) => channel.unbind());
        this._isBound = false;
        this.pause();
    }
    setMasterVolume(channel, volume) {
        this._channels[channel].setMasterVolume(volume);
    }
    pause() {
        return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this._suspended) {
                return;
            }
            this._suspended = true;
            yield waitMax(this._context.suspend());
        }));
    }
    resume() {
        return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this._suspended) {
                return;
            }
            this._suspended = false;
            yield waitMax(this._context.resume());
        }));
    }
    close() {
        this._mutex.runExclusive(() => this._context.close());
    }
}
exports.default = WebAudioDriver;
//# sourceMappingURL=WebAudio.js.map