"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const RingBuffer_1 = tslib_1.__importDefault(require("../../../tools/RingBuffer"));
const LinearResampler_1 = tslib_1.__importDefault(require("./LinearResampler"));
class PCMChannel {
    constructor(_hostFragmentSize = 1024) {
        this._hostFragmentSize = _hostFragmentSize;
        this._outputSampleRate = 0;
        this._bufferSize = 0;
        this._volume = 1;
        this._gain = null;
        this._processor = null;
        this._bufferUnderrun = false;
        this._fragmentRing = null;
        this._fragmentSize = 0;
        this._inputSampleRate = 0;
        this._fragmentIndex = 0;
        this._currentFragment = null;
        this._lastFragment = null;
        this._resampler = new LinearResampler_1.default();
    }
    init(context, target) {
        this._outputSampleRate = context.sampleRate;
        this._gain = context.createGain();
        this._gain.gain.value = this._volume;
        this._gain.connect(target);
        this._processor = context.createScriptProcessor(this._hostFragmentSize, 1, 1);
        this._bufferSize = this._processor.bufferSize;
        this._processor.connect(this._gain);
        this._processor.onaudioprocess = e => this._processAudio(e);
        const buffer = context.createBuffer(1, 1, context.sampleRate);
        buffer.getChannelData(0).set([0]);
    }
    bind(audio) {
        this.unbind();
        this._audio = audio;
        this._fragmentSize = audio.getFrameSize();
        this._inputSampleRate = audio.getSampleRate();
        this._fragmentIndex = 0;
        this._lastFragment = null;
        this._bufferUnderrun = true;
        this._fragmentRing = new RingBuffer_1.default(Math.ceil(((4 * this._bufferSize) / this._outputSampleRate / this._fragmentSize) * this._inputSampleRate));
        this._fragmentRing.evict.addHandler(b => b.release());
        this._audio.newFrame.addHandler(PCMChannel._onNewFragment, this);
        this._resampler.reset(this._inputSampleRate, this._outputSampleRate);
    }
    unbind() {
        if (!this._audio) {
            return;
        }
        this._audio.newFrame.removeHandler(PCMChannel._onNewFragment, this);
        if (this._lastFragment) {
            this._lastFragment.release();
            this._lastFragment = null;
        }
        if (this._currentFragment) {
            this._currentFragment.release();
            this._currentFragment = null;
        }
        if (this._fragmentRing) {
            this._fragmentRing.forEach(b => b.release());
            this._fragmentRing.clear();
            this._fragmentRing = null;
        }
    }
    setMasterVolume(volume) {
        this._volume = volume;
    }
    static _onNewFragment(fragment, self) {
        self._fragmentRing.push(fragment);
        if (!self._currentFragment) {
            self._currentFragment = self._fragmentRing.pop();
            self._fragmentIndex = 0;
        }
    }
    _processAudio(e) {
        if (!this._audio) {
            return;
        }
        const outputBuffer = e.outputBuffer.getChannelData(0);
        let bufferIndex = 0;
        const fillBuffer = (until) => {
            const previousFragmentBuffer = this._lastFragment && this._lastFragment.get();
            while (bufferIndex < until) {
                if (this._resampler.needsData()) {
                    this._resampler.push((this._audio && this._audio.isPaused()) || !previousFragmentBuffer
                        ? 0
                        : previousFragmentBuffer[this._fragmentIndex++] * this._volume);
                    if (this._fragmentIndex >= this._fragmentSize) {
                        this._fragmentIndex = 0;
                    }
                }
                outputBuffer[bufferIndex++] = this._resampler.get();
            }
        };
        if (this._currentFragment && this._bufferUnderrun) {
            fillBuffer(this._bufferSize >>> 1);
            this._bufferUnderrun = false;
        }
        let fragmentBuffer = this._currentFragment && this._currentFragment.get();
        while (bufferIndex < this._bufferSize && this._currentFragment) {
            if (this._resampler.needsData()) {
                this._resampler.push(fragmentBuffer[this._fragmentIndex++] * this._volume);
                if (this._fragmentIndex >= this._fragmentSize) {
                    this._fragmentIndex = 0;
                    if (this._lastFragment) {
                        this._lastFragment.release();
                    }
                    this._lastFragment = this._currentFragment;
                    this._currentFragment = this._fragmentRing.pop();
                    fragmentBuffer = this._currentFragment && this._currentFragment.get();
                }
            }
            outputBuffer[bufferIndex++] = this._resampler.get();
        }
        if (bufferIndex < this._bufferSize) {
            this._bufferUnderrun = true;
        }
        fillBuffer(this._bufferSize);
    }
}
exports.default = PCMChannel;
//# sourceMappingURL=PCMChannel.js.map