"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const Program_1 = tslib_1.__importDefault(require("./Program"));
const shader_1 = require("./shader");
class NtscProcessor {
    constructor(_gl, _capabilities) {
        this._gl = _gl;
        this._capabilities = _capabilities;
        this._height = 0;
        this._programPass1 = null;
        this._programPass2 = null;
        this._targetPass1 = null;
        this._targetPass2 = null;
        this._framebuffer = null;
        this._vertexCoordinateBuffer = null;
        this._textureCoordinateBuffer = null;
        this._initialized = false;
    }
    init() {
        if (this._initialized)
            return;
        const gl = this._gl;
        gl.getExtension('WEBGL_color_buffer_float');
        gl.getExtension('OES_texture_float');
        this._programPass1 = Program_1.default.compile(gl, shader_1.vsh.plain.source(this._capabilities), shader_1.fsh.ntscPass1.source(this._capabilities));
        this._programPass2 = Program_1.default.compile(gl, shader_1.vsh.plain.source(this._capabilities), shader_1.fsh.ntscPass2.source(this._capabilities));
        this._programPass1.use();
        this._programPass1.uniform1i("u_Sampler0", 0);
        this._programPass2.use();
        this._programPass2.uniform1i("u_Sampler0", 0);
        this._framebuffer = gl.createFramebuffer();
        this._vertexCoordinateBuffer = gl.createBuffer();
        this._textureCoordinateBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this._vertexCoordinateBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array([1, 1, -1, 1, 1, -1, -1, -1]), gl.STATIC_DRAW);
        gl.bindBuffer(gl.ARRAY_BUFFER, this._textureCoordinateBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array([1, 1, 0, 1, 1, 0, 0, 0]), gl.STATIC_DRAW);
        this._initialized = true;
    }
    destroy() {
        if (!this._initialized)
            return;
        const gl = this._gl;
        this._programPass1.delete();
        this._programPass2.delete();
        gl.deleteFramebuffer(this._framebuffer);
        gl.deleteTexture(this._targetPass1);
        gl.deleteTexture(this._targetPass2);
        gl.deleteBuffer(this._vertexCoordinateBuffer);
        gl.deleteBuffer(this._textureCoordinateBuffer);
        this._initialized = false;
    }
    render(texture) {
        this._pass(texture, this._targetPass1, this._programPass1, 960);
        this._pass(this._targetPass1, this._targetPass2, this._programPass2, 480);
    }
    getWidth() {
        return 480;
    }
    getHeight() {
        return this._height;
    }
    getTexture() {
        return this._targetPass2;
    }
    resize(width, height) {
        if (!this._initialized)
            return;
        if (width !== 160) {
            throw new Error('NTSC postprocessor supports only for width = 160');
        }
        this._height = height;
        const gl = this._gl;
        if (this._targetPass1)
            gl.deleteTexture(this._targetPass1);
        if (this._targetPass2)
            gl.deleteTexture(this._targetPass2);
        this._targetPass1 = gl.createTexture();
        this._targetPass2 = gl.createTexture();
        gl.activeTexture(gl.TEXTURE0);
        for (const texture of [this._targetPass1, this._targetPass2]) {
            gl.bindTexture(gl.TEXTURE_2D, texture);
            gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, texture === this._targetPass2 ? 480 : 960, height, 0, gl.RGBA, texture === this._targetPass1 ? this._textureType() : gl.UNSIGNED_BYTE, null);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
        }
    }
    configure(mode) {
        if (!this._initialized)
            return;
        this._programPass1.use();
        if (mode === "composite") {
            this._programPass1.uniform1f("u_Artifacting", 1);
            this._programPass1.uniform1f("u_Fringing", 1);
        }
        else {
            this._programPass1.uniform1f("u_Artifacting", 0);
            this._programPass1.uniform1f("u_Fringing", 0);
        }
    }
    _pass(textureIn, textureOut, program, width) {
        const gl = this._gl;
        program.use();
        program.bindVertexAttribArray("a_VertexPosition", this._vertexCoordinateBuffer, 2, gl.FLOAT, false, 0, 0);
        program.bindVertexAttribArray("a_TextureCoordinate", this._textureCoordinateBuffer, 2, gl.FLOAT, false, 0, 0);
        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, textureIn);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
        gl.bindFramebuffer(gl.FRAMEBUFFER, this._framebuffer);
        gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, textureOut, 0);
        gl.viewport(0, 0, width, this._height);
        gl.clearColor(0, 0, 0, 1);
        gl.clear(gl.COLOR_BUFFER_BIT);
        gl.drawArrays(gl.TRIANGLE_STRIP, 0, 4);
    }
    _textureType() {
        const gl = this._gl;
        if (this._capabilities.floatTextures) {
            return gl.FLOAT;
        }
        if (this._capabilities.halfFloatTextures) {
            return gl.getExtension('OES_texture_half_float').HALF_FLOAT_OES;
        }
        return gl.UNSIGNED_BYTE;
    }
}
exports.default = NtscProcessor;
//# sourceMappingURL=NtscProcessor.js.map