"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const Program_1 = tslib_1.__importDefault(require("./Program"));
const shader_1 = require("./shader");
class PhosphorProcessor {
    constructor(_gl, _capabilities) {
        this._gl = _gl;
        this._capabilities = _capabilities;
        this._width = 0;
        this._height = 0;
        this._texture0 = null;
        this._texture1 = null;
        this._framebuffer = null;
        this._program = null;
        this._vertexCoordinateBuffer = null;
        this._textureCoordinateBuffer = null;
        this._initialized = false;
    }
    init() {
        if (this._initialized)
            return;
        const gl = this._gl;
        this._framebuffer = gl.createFramebuffer();
        this._program = Program_1.default.compile(gl, shader_1.vsh.plain.source(this._capabilities), shader_1.fsh.phosphor.source(this._capabilities));
        this._program.use();
        this._program.uniform1i("u_Sampler_NewImage", 0);
        this._program.uniform1i("u_Sampler_PreviousImage", 1);
        this._vertexCoordinateBuffer = gl.createBuffer();
        this._textureCoordinateBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this._vertexCoordinateBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array([1, 1, -1, 1, 1, -1, -1, -1]), gl.STATIC_DRAW);
        gl.bindBuffer(gl.ARRAY_BUFFER, this._textureCoordinateBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array([1, 1, 0, 1, 1, 0, 0, 0]), gl.STATIC_DRAW);
        this._initialized = true;
    }
    destroy() {
        if (!this._initialized)
            return;
        const gl = this._gl;
        this._program.delete();
        gl.deleteFramebuffer(this._framebuffer);
        gl.deleteBuffer(this._vertexCoordinateBuffer);
        gl.deleteBuffer(this._textureCoordinateBuffer);
        if (this._texture0)
            gl.deleteTexture(this._texture0);
        if (this._texture1)
            gl.deleteTexture(this._texture1);
        this._initialized = false;
    }
    render(texture) {
        const gl = this._gl;
        this._program.use();
        this._program.bindVertexAttribArray("a_VertexPosition", this._vertexCoordinateBuffer, 2, gl.FLOAT, false, 0, 0);
        this._program.bindVertexAttribArray("a_TextureCoordinate", this._textureCoordinateBuffer, 2, gl.FLOAT, false, 0, 0);
        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
        gl.activeTexture(gl.TEXTURE1);
        gl.bindTexture(gl.TEXTURE_2D, this._texture0);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
        gl.bindFramebuffer(gl.FRAMEBUFFER, this._framebuffer);
        gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, this._texture1, 0);
        gl.viewport(0, 0, this._width, this._height);
        gl.clearColor(0, 0, 0, 1);
        gl.clear(gl.COLOR_BUFFER_BIT);
        gl.drawArrays(gl.TRIANGLE_STRIP, 0, 4);
        const t = this._texture1;
        this._texture1 = this._texture0;
        this._texture0 = t;
    }
    getWidth() {
        return this._width;
    }
    getHeight() {
        return this._height;
    }
    getTexture() {
        return this._texture0;
    }
    resize(width, height) {
        if (!this._initialized)
            return;
        this._width = width;
        this._height = height;
        const gl = this._gl;
        if (this._texture0)
            gl.deleteTexture(this._texture0);
        if (this._texture1)
            gl.deleteTexture(this._texture1);
        this._texture0 = gl.createTexture();
        this._texture1 = gl.createTexture();
        gl.bindFramebuffer(gl.FRAMEBUFFER, this._framebuffer);
        gl.viewport(0, 0, width, height);
        gl.clearColor(0, 0, 0, 1);
        gl.activeTexture(gl.TEXTURE0);
        for (const texture of [this._texture0, this._texture1]) {
            gl.bindTexture(gl.TEXTURE_2D, texture);
            gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, width, height, 0, gl.RGBA, gl.UNSIGNED_BYTE, null);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
            gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, texture, 0);
            gl.clear(gl.COLOR_BUFFER_BIT);
        }
    }
    configure(level) {
        if (!this._initialized)
            return;
        this._program.use();
        this._program.uniform1f("u_PhosphorLevel", level);
    }
}
exports.default = PhosphorProcessor;
//# sourceMappingURL=PhosphorProcessor.js.map