"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function compileShader(gl, type, source) {
    const shader = gl.createShader(type);
    gl.shaderSource(shader, source);
    gl.compileShader(shader);
    if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
        throw new Error(`failed to compile shader:\n\n${gl.getShaderInfoLog(shader)}\n\n${source}`);
    }
    return shader;
}
class Program {
    constructor(_gl, _program, _vsh, _fsh) {
        this._gl = _gl;
        this._program = _program;
        this._vsh = _vsh;
        this._fsh = _fsh;
        this._attributeLocations = new Map();
        this._uniformLocations = new Map();
    }
    static compile(gl, vshSource, fshSource, layout = { a_VertexPosition: 0 }) {
        const vsh = compileShader(gl, gl.VERTEX_SHADER, vshSource);
        const fsh = compileShader(gl, gl.FRAGMENT_SHADER, fshSource);
        const program = gl.createProgram();
        for (const attrib of Object.keys(layout)) {
            gl.bindAttribLocation(program, layout[attrib], attrib);
        }
        gl.attachShader(program, vsh);
        gl.attachShader(program, fsh);
        gl.linkProgram(program);
        if (!gl.getProgramParameter(program, gl.LINK_STATUS)) {
            throw new Error(`failed to link program:\n\n${gl.getProgramInfoLog(program)}`);
        }
        return new Program(gl, program, vsh, fsh);
    }
    delete() {
        const gl = this._gl;
        gl.deleteProgram(this._program);
        gl.deleteShader(this._vsh);
        gl.deleteShader(this._fsh);
    }
    use() {
        this._gl.useProgram(this._program);
    }
    getAttribLocation(name) {
        if (!this._attributeLocations.has(name)) {
            const location = this._gl.getAttribLocation(this._program, name);
            if (location < 0) {
                throw new Error(`invalid attribute ${name}`);
            }
            this._attributeLocations.set(name, location);
        }
        return this._attributeLocations.get(name);
    }
    getUniformLocation(name) {
        if (!this._uniformLocations.has(name)) {
            const location = this._gl.getUniformLocation(this._program, name);
            if (location === null) {
                throw new Error(`invalid uniform ${name}`);
            }
            this._uniformLocations.set(name, location);
        }
        return this._uniformLocations.get(name);
    }
    bindVertexAttribArray(attribute, buffer, size, type, normalized, stride, offset) {
        const gl = this._gl;
        gl.bindBuffer(gl.ARRAY_BUFFER, buffer);
        gl.vertexAttribPointer(this.getAttribLocation(attribute), size, type, normalized, stride, offset);
        gl.enableVertexAttribArray(this.getAttribLocation(attribute));
    }
    uniform1i(uniform, value) {
        this._gl.uniform1i(this.getUniformLocation(uniform), value);
    }
    uniform1f(uniform, value) {
        this._gl.uniform1f(this.getUniformLocation(uniform), value);
    }
}
exports.default = Program;
//# sourceMappingURL=Program.js.map