"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = exports.Stellerator = void 0;
const tslib_1 = require("tslib");
const async_mutex_1 = require("async-mutex");
const microevent_ts_1 = require("microevent.ts");
const EmulationServiceInterface_1 = tslib_1.__importDefault(require("../../stella/service/EmulationServiceInterface"));
const EmulationService_1 = tslib_1.__importDefault(require("../../stella/service/worker/EmulationService"));
const DriverManager_1 = tslib_1.__importDefault(require("../../stella/service/DriverManager"));
const Video_1 = tslib_1.__importDefault(require("../../driver/Video"));
const WebAudio_1 = tslib_1.__importDefault(require("../../stella/driver/WebAudio"));
const KeyboardIO_1 = tslib_1.__importDefault(require("../../stella/driver/KeyboardIO"));
const TouchIO_1 = tslib_1.__importDefault(require("../../stella/driver/TouchIO"));
const MouseAsPaddle_1 = tslib_1.__importDefault(require("../../driver/MouseAsPaddle"));
const Gamepad_1 = tslib_1.__importDefault(require("../../driver/Gamepad"));
const FullscreenVideo_1 = tslib_1.__importDefault(require("../../driver/FullscreenVideo"));
const CartridgeInfo_1 = tslib_1.__importDefault(require("../../../machine/stella/cartridge/CartridgeInfo"));
const Config_1 = tslib_1.__importDefault(require("../../../machine/stella/Config"));
const base64_1 = require("../../../tools/base64");
const ControlPanelProxy_1 = tslib_1.__importDefault(require("./ControlPanelProxy"));
const Factory_1 = tslib_1.__importDefault(require("../../../machine/cpu/Factory"));
const AsyncIO_1 = tslib_1.__importDefault(require("../../driver/AsyncIO"));
function cpuType(config = Stellerator.CpuAccuracy.cycle) {
    switch (config) {
        case Stellerator.CpuAccuracy.cycle:
            return Factory_1.default.Type.stateMachine;
        case Stellerator.CpuAccuracy.instruction:
            return Factory_1.default.Type.batchedAccess;
        default:
            throw new Error(`invalid CPU Accuracy: ${config}`);
    }
}
class Stellerator {
    constructor(canvasElt = null, workerUrl, config = {}) {
        this._pauseHandler = () => {
            switch (this._emulationService.getState()) {
                case EmulationServiceInterface_1.default.State.paused:
                    this.resume();
                    break;
                case EmulationServiceInterface_1.default.State.running:
                    this.pause();
                    break;
            }
        };
        this._canvasElt = null;
        this._config = null;
        this._emulationService = null;
        this._serviceInitialized = null;
        this._videoDriver = null;
        this._fullscreenVideo = null;
        this._audioDriver = null;
        this._keyboardIO = null;
        this._asyncIO = null;
        this._touchIO = null;
        this._paddle = null;
        this._gamepad = null;
        this._controlPanel = new ControlPanelProxy_1.default();
        this._state = Stellerator.State.stopped;
        this._driverManager = new DriverManager_1.default();
        this._mutex = new async_mutex_1.Mutex();
        this._canvasElt = canvasElt;
        this._config = Object.assign({ gamma: 1, scalingMode: Stellerator.ScalingMode.qis, tvEmulation: Stellerator.TvEmulation.composite, phosphorLevel: 0.5, scanlineLevel: 0.2, audio: true, volume: 0.5, enableKeyboard: true, enableTouch: true, touchLeftHanded: false, touchJoystickSensitivity: 15, keyboardTarget: document, fullscreenViaKeyboard: true, paddleViaMouse: true, pauseViaKeyboard: true, pauseViaTouch: true, fullscreenViaTouch: true, enableGamepad: true, resetViaKeyboard: true }, config);
        this._emulationService = new EmulationService_1.default(workerUrl);
        this.frequencyUpdate = this._emulationService.frequencyUpdate;
        const stateChange = new microevent_ts_1.Event();
        this._emulationService.stateChanged.addHandler((newState) => stateChange.dispatch(this._mapState(newState)));
        this.stateChange = stateChange;
        this._createDrivers();
        this._driverManager.addDriver(this._controlPanel, (context) => this._controlPanel.bind(context.getControlPanel()));
        this._driverManager.bind(this._emulationService);
        this._serviceInitialized = this._emulationService.init().then(undefined, (e) => {
            console.log(e);
            throw e;
        });
    }
    setGamma(gamma) {
        this._config.gamma = gamma;
        this._createVideoDriver();
        return this;
    }
    getGamma() {
        return this._config.gamma;
    }
    setScalingMode(scalingMode) {
        this._config.scalingMode = scalingMode;
        this._createVideoDriver();
        return this;
    }
    getScalingMode() {
        return this._config.scalingMode;
    }
    setTvEmulation(tvEmulation) {
        this._config.tvEmulation = tvEmulation;
        this._createVideoDriver();
        return this;
    }
    getTvEmulation() {
        return this._config.tvEmulation;
    }
    setPhosphorLevel(phosphorLevel) {
        this._config.phosphorLevel = phosphorLevel;
        this._createVideoDriver();
        return this;
    }
    getPhosphorLevel() {
        return this._config.phosphorLevel;
    }
    setScanlineLevel(scanlineLevel) {
        this._config.scanlineLevel = scanlineLevel;
        this._createVideoDriver();
        return this;
    }
    getScanlineLevel() {
        return this._config.scanlineLevel;
    }
    toggleFullscreen(fullscreen) {
        if (!this._fullscreenVideo) {
            return this;
        }
        if (typeof fullscreen === 'undefined') {
            this._fullscreenVideo.toggle();
        }
        else {
            fullscreen ? this._fullscreenVideo.engage() : this._fullscreenVideo.disengage();
        }
        return this;
    }
    isFullscreen() {
        return this._fullscreenVideo ? this._fullscreenVideo.isEngaged() : false;
    }
    setVolume(volume) {
        if (this._audioDriver) {
            this._audioDriver.setMasterVolume(Math.max(Math.min(volume, 1), 0));
        }
        return this;
    }
    audioEnabled() {
        return !!this._audioDriver;
    }
    getVolume() {
        return this._audioDriver ? this._audioDriver.getMasterVolume() : 0;
    }
    resize() {
        if (this._videoDriver) {
            this._videoDriver.resize();
        }
        return this;
    }
    getState() {
        return this._state;
    }
    getControlPanel() {
        return this._controlPanel;
    }
    setCanvas(canvas) {
        this._canvasElt = canvas;
        this._createVideoDriver();
        this._createTouchDriver();
        return this;
    }
    releaseCanvas() {
        this._removeVideoDriver();
        this._removeTouchDriver();
        this._canvasElt = null;
        return this;
    }
    start(cartridge, tvMode, config = {}) {
        return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (typeof cartridge === 'string') {
                cartridge = base64_1.decode(cartridge);
            }
            const stellaConfig = Config_1.default.create({
                tvMode: this._convertTvMode(tvMode),
                pcmAudio: true,
            });
            if (typeof config.randomSeed !== 'undefined' && config.randomSeed > 0) {
                stellaConfig.randomSeed = config.randomSeed;
            }
            if (typeof config.emulatePaddles !== 'undefined') {
                stellaConfig.emulatePaddles = config.emulatePaddles;
            }
            if (typeof config.frameStart !== 'undefined') {
                stellaConfig.frameStart = config.frameStart;
            }
            stellaConfig.cpuType = cpuType(config.cpuAccuracy);
            yield this._serviceInitialized;
            return (this._state = this._mapState(yield this._emulationService.start(cartridge, Object.assign(Object.assign({}, stellaConfig), { asyncIO: config.asyncIO }), config.cartridgeType)));
        }));
    }
    run(cartridge, tvMode, config = {}) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if ((yield this.start(cartridge, tvMode, config)) === Stellerator.State.paused) {
                return this.resume();
            }
        });
    }
    pause() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._serviceInitialized;
            return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () { return (this._state = this._mapState(yield this._emulationService.pause())); }));
        });
    }
    resume() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._serviceInitialized;
            return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () { return (this._state = this._mapState(yield this._emulationService.resume())); }));
        });
    }
    stop() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._serviceInitialized;
            return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () { return (this._state = this._mapState(yield this._emulationService.stop())); }));
        });
    }
    reset() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._serviceInitialized;
            return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () { return (this._state = this._mapState(yield this._emulationService.reset())); }));
        });
    }
    lastError() {
        return this._emulationService.getLastError();
    }
    asyncIOSend(message) {
        this._asyncIO.send(message);
        return this;
    }
    _convertTvMode(tvMode) {
        switch (tvMode) {
            case Stellerator.TvMode.ntsc:
                return 0;
            case Stellerator.TvMode.pal:
                return 1;
            case Stellerator.TvMode.secam:
                return 2;
            default:
                throw new Error(`invalid TV mode '${tvMode}'`);
        }
    }
    _createDrivers() {
        this._createVideoDriver();
        if (this._config.audio) {
            try {
                this._audioDriver = new WebAudio_1.default();
                this._audioDriver.init();
                this._audioDriver.setMasterVolume(this._config.volume);
                this._driverManager.addDriver(this._audioDriver, (context) => this._audioDriver.bind(true, [context.getPCMChannel()]));
                this._emulationService.stateChanged.addHandler((newState) => {
                    switch (newState) {
                        case EmulationServiceInterface_1.default.State.running:
                            this._audioDriver.resume();
                            break;
                        default:
                            this._audioDriver.pause();
                            break;
                    }
                });
            }
            catch (e) {
                console.error(`failed to initialize audio: ${e && e.message}`);
            }
        }
        this._createTouchDriver();
        if (this._config.enableKeyboard) {
            this._keyboardIO = new KeyboardIO_1.default(this._config.keyboardTarget);
            this._driverManager.addDriver(this._keyboardIO, (context) => this._keyboardIO.bind(context.getJoystick(0), context.getJoystick(1), context.getControlPanel()));
            if (this._config.fullscreenViaKeyboard) {
                this._keyboardIO.toggleFullscreen.addHandler(() => this._fullscreenVideo && this._fullscreenVideo.toggle());
            }
            if (this._config.pauseViaKeyboard) {
                this._keyboardIO.togglePause.addHandler(this._pauseHandler);
            }
        }
        if (this._config.resetViaKeyboard) {
            this._keyboardIO.hardReset.addHandler(() => this.reset());
        }
        if (this._config.enableGamepad) {
            this._gamepad = new Gamepad_1.default();
            this._gamepad.init();
            this._driverManager.addDriver(this._gamepad, (context) => this._gamepad.bind([context.getJoystick(0), context.getJoystick(1)], {
                ["start"]: context.getControlPanel().getResetButton(),
                ["select"]: context.getControlPanel().getSelectSwitch(),
            }));
        }
        if (this._config.paddleViaMouse) {
            this._paddle = new MouseAsPaddle_1.default();
            this._driverManager.addDriver(this._paddle, (context) => this._paddle.bind(context.getPaddle(0)));
        }
        this._asyncIO = new AsyncIO_1.default();
        this.asyncIOMessage = this._asyncIO.message;
        this._driverManager.addDriver(this._asyncIO, (context, driver) => driver.bind(context.getAsyncIO()));
    }
    _removeVideoDriver() {
        if (!this._videoDriver) {
            return;
        }
        this._videoDriver.unbind();
        this._driverManager.removeDriver(this._videoDriver);
        this._fullscreenVideo.disengage();
        this._fullscreenVideo = null;
        this._videoDriver = null;
    }
    _createVideoDriver() {
        if (this._videoDriver) {
            this._removeVideoDriver();
        }
        if (!this._canvasElt) {
            return;
        }
        this._videoDriver = new Video_1.default(this._canvasElt, {
            gamma: this._config.gamma,
            scalingMode: this._config.scalingMode,
            tvEmulation: this._config.tvEmulation,
            phosphorLevel: this._config.phosphorLevel,
            scanlineLevel: this._config.scanlineLevel,
        }).init();
        this._driverManager.addDriver(this._videoDriver, (context) => this._videoDriver.bind(context.getVideo()));
        this._fullscreenVideo = new FullscreenVideo_1.default(this._videoDriver);
    }
    _removeTouchDriver() {
        if (!this._touchIO) {
            return;
        }
        this._touchIO.unbind();
        this._driverManager.removeDriver(this._touchIO);
        this._touchIO = null;
    }
    _createTouchDriver() {
        if (this._touchIO) {
            this._removeTouchDriver();
        }
        if (this._config.enableTouch) {
            this._touchIO = new TouchIO_1.default(this._canvasElt, this._config.touchJoystickSensitivity, this._config.touchLeftHanded);
            this._driverManager.addDriver(this._touchIO, (context) => this._touchIO.bind(context.getJoystick(0), context.getControlPanel()));
            if (this._config.pauseViaTouch) {
                this._touchIO.togglePause.addHandler(this._pauseHandler);
            }
            if (this._config.fullscreenViaTouch) {
                this._touchIO.toggleFullscreen.addHandler(() => this._fullscreenVideo && this._fullscreenVideo.toggle());
            }
        }
    }
    _mapState(state) {
        switch (state) {
            case EmulationServiceInterface_1.default.State.stopped:
                return Stellerator.State.stopped;
            case EmulationServiceInterface_1.default.State.running:
                return Stellerator.State.running;
            case EmulationServiceInterface_1.default.State.paused:
                return Stellerator.State.paused;
            case EmulationServiceInterface_1.default.State.error:
                return Stellerator.State.error;
            default:
                throw new Error('cannot happen');
        }
    }
}
exports.Stellerator = Stellerator;
exports.default = Stellerator;
(function (Stellerator) {
    let TvMode;
    (function (TvMode) {
        TvMode["ntsc"] = "ntsc";
        TvMode["pal"] = "pal";
        TvMode["secam"] = "secam";
    })(TvMode = Stellerator.TvMode || (Stellerator.TvMode = {}));
    let TvEmulation;
    (function (TvEmulation) {
        TvEmulation["composite"] = "composite";
        TvEmulation["svideo"] = "svideo";
        TvEmulation["none"] = "none";
    })(TvEmulation = Stellerator.TvEmulation || (Stellerator.TvEmulation = {}));
    let ScalingMode;
    (function (ScalingMode) {
        ScalingMode["qis"] = "qis";
        ScalingMode["bilinear"] = "bilinear";
        ScalingMode["none"] = "none";
    })(ScalingMode = Stellerator.ScalingMode || (Stellerator.ScalingMode = {}));
    Stellerator.CartridgeType = CartridgeInfo_1.default.CartridgeType;
    Stellerator.describeCartridgeType = CartridgeInfo_1.default.describeCartridgeType;
    Stellerator.allCartridgeTypes = CartridgeInfo_1.default.getAllTypes;
    let State;
    (function (State) {
        State["running"] = "running";
        State["paused"] = "paused";
        State["stopped"] = "stopped";
        State["error"] = "error";
    })(State = Stellerator.State || (Stellerator.State = {}));
    let CpuAccuracy;
    (function (CpuAccuracy) {
        CpuAccuracy["cycle"] = "cycle";
        CpuAccuracy["instruction"] = "instruction";
    })(CpuAccuracy = Stellerator.CpuAccuracy || (Stellerator.CpuAccuracy = {}));
})(Stellerator = exports.Stellerator || (exports.Stellerator = {}));
exports.default = Stellerator;
//# sourceMappingURL=Stellerator.js.map