"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const microevent_ts_1 = require("microevent.ts");
function mkSwitch(swtch) {
    return {
        type: 0,
        swtch
    };
}
function mkTrigger(event) {
    return {
        type: 1,
        trigger: event
    };
}
class KeyboardIO {
    constructor(_target, mappings = KeyboardIO.defaultMappings) {
        this._target = _target;
        this.toggleFullscreen = new microevent_ts_1.Event();
        this.hardReset = new microevent_ts_1.Event();
        this.togglePause = new microevent_ts_1.Event();
        this._keydownListener = null;
        this._keyupListener = null;
        this._joystick0 = null;
        this._joystick1 = null;
        this._controlPanel = null;
        this._dispatchTable = {};
        this._compiledMappings = new Map();
        this._compileMappings(mappings);
    }
    bind(joystick0, joystick1, controlPanel) {
        if (this._joystick0) {
            return;
        }
        this._joystick0 = joystick0;
        this._joystick1 = joystick1;
        this._controlPanel = controlPanel;
        this._updateActionTable();
        this._keydownListener = e => {
            if (!this._compiledMappings.has(e.keyCode)) {
                return;
            }
            const modifiers = (e.shiftKey ? 4 : 0) |
                (e.ctrlKey ? 1 : 0) |
                (e.altKey ? 2 : 0);
            if (!this._compiledMappings.get(e.keyCode).has(modifiers)) {
                return;
            }
            const action = this._compiledMappings.get(e.keyCode).get(modifiers);
            if (typeof action !== 'undefined') {
                e.preventDefault();
                const dispatch = this._dispatchTable[action];
                switch (dispatch.type) {
                    case 0:
                        dispatch.swtch.toggle(true);
                        break;
                    case 1:
                        dispatch.trigger.dispatch(undefined);
                        break;
                    default:
                }
            }
        };
        this._keyupListener = e => {
            if (!this._compiledMappings.has(e.keyCode)) {
                return;
            }
            for (const action of this._compiledMappings.get(e.keyCode).values()) {
                e.preventDefault();
                const dispatch = this._dispatchTable[action];
                switch (dispatch.type) {
                    case 0:
                        dispatch.swtch.toggle(false);
                        break;
                    default:
                }
            }
        };
        this._target.addEventListener('keydown', this._keydownListener);
        this._target.addEventListener('keyup', this._keyupListener);
    }
    unbind() {
        if (!this._joystick0) {
            return;
        }
        this._target.removeEventListener('keydown', this._keydownListener);
        this._target.removeEventListener('keyup', this._keyupListener);
        this._joystick0 = this._joystick1 = this._controlPanel = null;
        this._keydownListener = this._keyupListener = null;
    }
    _updateActionTable() {
        this._dispatchTable[12] = mkTrigger(this.toggleFullscreen);
        this._dispatchTable[13] = mkTrigger(this.hardReset);
        this._dispatchTable[14] = mkTrigger(this.togglePause);
        this._dispatchTable[0] = mkSwitch(this._controlPanel.getSelectSwitch());
        this._dispatchTable[1] = mkSwitch(this._controlPanel.getResetButton());
        this._dispatchTable[2] = mkSwitch(this._joystick0.getLeft());
        this._dispatchTable[3] = mkSwitch(this._joystick0.getRight());
        this._dispatchTable[4] = mkSwitch(this._joystick0.getUp());
        this._dispatchTable[5] = mkSwitch(this._joystick0.getDown());
        this._dispatchTable[10] = mkSwitch(this._joystick0.getFire());
        this._dispatchTable[6] = mkSwitch(this._joystick1.getLeft());
        this._dispatchTable[7] = mkSwitch(this._joystick1.getRight());
        this._dispatchTable[8] = mkSwitch(this._joystick1.getUp());
        this._dispatchTable[9] = mkSwitch(this._joystick1.getDown());
        this._dispatchTable[11] = mkSwitch(this._joystick1.getFire());
    }
    _compileMappings(mappings) {
        const compileMapping = (action, keycode, modifiers) => {
            if ((modifiers & ~(4 | 1 | 2)) !== 0) {
                throw new Error(`invalid modifier set ${modifiers}`);
            }
            if (!this._compiledMappings.has(keycode)) {
                this._compiledMappings.set(keycode, new Map());
            }
            this._compiledMappings.get(keycode).set(modifiers, action);
        };
        mappings.forEach(mapping => {
            const action = mapping.action, specs = Array.isArray(mapping.spec) ? mapping.spec : [mapping.spec];
            specs.forEach(spec => compileMapping(action, typeof spec === 'object' ? spec.keycode : spec, typeof spec === 'object' ? spec.modifiers : 0));
        });
    }
}
exports.default = KeyboardIO;
(function (KeyboardIO) {
    KeyboardIO.defaultMappings = [
        {
            action: 0,
            spec: {
                keycode: 32,
                modifiers: 4
            }
        },
        {
            action: 1,
            spec: {
                keycode: 13,
                modifiers: 4
            }
        },
        {
            action: 2,
            spec: [
                65,
                37
            ]
        },
        {
            action: 3,
            spec: [
                68,
                39
            ]
        },
        {
            action: 4,
            spec: [
                87,
                38
            ]
        },
        {
            action: 5,
            spec: [
                83,
                40
            ]
        },
        {
            action: 10,
            spec: [
                32,
                86
            ]
        },
        {
            action: 6,
            spec: 74
        },
        {
            action: 7,
            spec: 76
        },
        {
            action: 8,
            spec: 73
        },
        {
            action: 9,
            spec: 75
        },
        {
            action: 11,
            spec: 66
        },
        {
            action: 12,
            spec: 13
        },
        {
            action: 13,
            spec: {
                keycode: 82,
                modifiers: 4
            }
        },
        {
            action: 14,
            spec: 80
        }
    ];
})(KeyboardIO || (KeyboardIO = {}));
exports.default = KeyboardIO;
//# sourceMappingURL=KeyboardIO.js.map