"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const microevent_ts_1 = require("microevent.ts");
const DoubleTapDetector_1 = tslib_1.__importDefault(require("./touch/DoubleTapDetector"));
class TouchIO {
    constructor(_canvas, _joystickSensitivity = 15, _leftHanded = false) {
        this._canvas = _canvas;
        this._joystickSensitivity = _joystickSensitivity;
        this._leftHanded = _leftHanded;
        this._onTouchStart = (e) => {
            for (let i = 0; i < e.changedTouches.length; i++) {
                const normalizedTouch = new NormalizedTouch(e.changedTouches.item(i), this._canvas), id = normalizedTouch.touch.identifier;
                if (this._leftHanded ? normalizedTouch.x > 0.5 : normalizedTouch.x <= 0.5) {
                    if (normalizedTouch.y <= 0.5) {
                        normalizedTouch.type = "alt";
                    }
                    else {
                        normalizedTouch.type = this._isAlt ? "pause" : "fire";
                    }
                }
                else {
                    if (normalizedTouch.y <= 0.5) {
                        normalizedTouch.type = this._isAlt ? "select" : "joystick";
                    }
                    else {
                        normalizedTouch.type = this._isAlt ? "reset" : "joystick";
                    }
                }
                if (this._pendingTouches.has(id) || this._pendingTouches.has(normalizedTouch.type)) {
                    continue;
                }
                this._pendingTouches.set(id, normalizedTouch);
                this._pendingTouches.set(normalizedTouch.type, normalizedTouch);
                switch (normalizedTouch.type) {
                    case "alt":
                        this._isAlt = true;
                        this._fullscreenDoubleTapDetector.startTouch();
                        break;
                    case "fire":
                        this._joystick.getFire().toggle(true);
                        break;
                    case "pause":
                        this.togglePause.dispatch(undefined);
                        this._fullscreenDoubleTapDetector.cancelTouch();
                        break;
                    case "select":
                        this._select.toggle(true);
                        this._fullscreenDoubleTapDetector.cancelTouch();
                        break;
                    case "reset":
                        this._reset.toggle(true);
                        this._fullscreenDoubleTapDetector.cancelTouch();
                        break;
                    case "joystick":
                        break;
                    default:
                        throw new Error('invalid touch type');
                }
            }
            e.preventDefault();
        };
        this._onTouchEnd = (e) => {
            for (let i = 0; i < e.changedTouches.length; i++) {
                const normalizedTouch = this._pendingTouches.get(e.changedTouches.item(i).identifier);
                if (!normalizedTouch) {
                    continue;
                }
                switch (normalizedTouch.type) {
                    case "alt":
                        this._isAlt = false;
                        this._fullscreenDoubleTapDetector.endTouch();
                        break;
                    case "fire":
                        this._joystick.getFire().toggle(false);
                        break;
                    case "select":
                        this._select.toggle(false);
                        break;
                    case "reset":
                        this._reset.toggle(false);
                        break;
                    case "joystick":
                        this._joystick.getDown().toggle(false);
                        this._joystick.getUp().toggle(false);
                        this._joystick.getLeft().toggle(false);
                        this._joystick.getRight().toggle(false);
                        break;
                    case "pause":
                        break;
                    default:
                        throw new Error('invalid touch type');
                }
                this._pendingTouches.delete(normalizedTouch.type);
                this._pendingTouches.delete(normalizedTouch.touch.identifier);
            }
            e.preventDefault();
        };
        this._onTouchMove = (e) => {
            for (let i = 0; i < e.changedTouches.length; i++) {
                const touch = e.changedTouches.item(i), normalizedTouch = this._pendingTouches.get(touch.identifier);
                if (!normalizedTouch) {
                    continue;
                }
                if (normalizedTouch.type !== "joystick") {
                    continue;
                }
                const deltaX = touch.clientX - normalizedTouch.x0, deltaY = touch.clientY - normalizedTouch.y0, abs = Math.sqrt(deltaX * deltaX + deltaY * deltaY), sin = Math.abs(deltaY / abs), cos = Math.abs(deltaX / abs), trigger = abs > this._joystickSensitivity;
                this._joystick.getLeft().toggle(trigger && deltaX < 0 && cos > 0.5);
                this._joystick.getRight().toggle(trigger && deltaX > 0 && cos > 0.5);
                this._joystick.getUp().toggle(trigger && deltaY < 0 && sin > 0.5);
                this._joystick.getDown().toggle(trigger && deltaY > 0 && sin > 0.5);
                if (abs > 3 * this._joystickSensitivity) {
                    normalizedTouch.x0 += (deltaX / abs) * (abs - 3 * this._joystickSensitivity);
                    normalizedTouch.y0 += (deltaY / abs) * (abs - 3 * this._joystickSensitivity);
                }
            }
            e.preventDefault();
        };
        this.togglePause = new microevent_ts_1.Event();
        this._fullscreenDoubleTapDetector = new DoubleTapDetector_1.default();
        this._bound = false;
        this._joystick = null;
        this._select = null;
        this._reset = null;
        this._isAlt = false;
        this._pendingTouches = new Map();
        this.toggleFullscreen = this._fullscreenDoubleTapDetector.trigger;
    }
    static isSupported() {
        return 'ontouchstart' in window;
    }
    bind(joystick, controlPanel) {
        if (this._bound) {
            return;
        }
        this._bound = true;
        this._joystick = joystick;
        this._select = controlPanel.getSelectSwitch();
        this._reset = controlPanel.getResetButton();
        this._bindListeners();
    }
    unbind() {
        if (!this._bound) {
            return;
        }
        this._unbindListeners();
        (this._bound = false), (this._joystick = this._reset = this._select = null);
    }
    _bindListeners() {
        this._canvas.addEventListener('touchstart', this._onTouchStart);
        this._canvas.addEventListener('touchend', this._onTouchEnd);
        this._canvas.addEventListener('touchmove', this._onTouchMove);
    }
    _unbindListeners() {
        this._canvas.removeEventListener('touchstart', this._onTouchStart);
        this._canvas.removeEventListener('touchend', this._onTouchEnd);
        this._canvas.removeEventListener('touchmove', this._onTouchMove);
    }
}
class NormalizedTouch {
    constructor(touch, canvas) {
        this.touch = touch;
        this.type = "unknown";
        const boundingRect = canvas.getBoundingClientRect();
        this.x = (touch.clientX - boundingRect.left) / boundingRect.width;
        this.y = (touch.clientY - boundingRect.top) / boundingRect.height;
        this.x0 = touch.clientX;
        this.y0 = touch.clientY;
    }
}
exports.default = TouchIO;
//# sourceMappingURL=TouchIO.js.map