"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const EmulationServiceInterface_1 = tslib_1.__importDefault(require("./EmulationServiceInterface"));
class DriverManager {
    constructor() {
        this._drivers = new Map();
        this._driversBound = false;
    }
    bind(emulationService) {
        if (this._driversBound) {
            return this;
        }
        this._emulationService = emulationService;
        if (this._shouldBindDrivers()) {
            this._bindDrivers();
        }
        this._emulationService.stateChanged.addHandler(DriverManager._onEmuStateChange, this);
        return this;
    }
    unbind() {
        if (!this._emulationService) {
            return this;
        }
        this._unbindDrivers();
        this._emulationService.stateChanged.removeHandler(DriverManager._onEmuStateChange, this);
        this._emulationService = null;
        return this;
    }
    addDriver(driver, binder) {
        this._drivers.set(driver, new DriverManager.DriverContext(driver, binder));
        if (this._driversBound) {
            binder(this._emulationService.getEmulationContext(), driver);
        }
        return this;
    }
    removeDriver(driver) {
        if (!this._drivers.get(driver)) {
            return this;
        }
        driver.unbind();
        this._drivers.delete(driver);
        return this;
    }
    static _onEmuStateChange(newState, self) {
        if (self._shouldBindDrivers(newState)) {
            self._bindDrivers();
        }
        else {
            self._unbindDrivers();
        }
    }
    _shouldBindDrivers(state = this._emulationService ? this._emulationService.getState() : undefined) {
        return (this._emulationService &&
            (state === EmulationServiceInterface_1.default.State.running || state === EmulationServiceInterface_1.default.State.paused));
    }
    _bindDrivers() {
        if (this._driversBound) {
            return;
        }
        this._drivers.forEach(driverContext => driverContext.binder(this._emulationService.getEmulationContext(), driverContext.driver));
        this._driversBound = true;
    }
    _unbindDrivers() {
        if (!this._driversBound) {
            return;
        }
        this._drivers.forEach(driverContext => driverContext.driver.unbind());
        this._driversBound = false;
    }
}
exports.default = DriverManager;
(function (DriverManager) {
    class DriverContext {
        constructor(driver, binder) {
            this.driver = driver;
            this.binder = binder;
        }
    }
    DriverManager.DriverContext = DriverContext;
})(DriverManager || (DriverManager = {}));
exports.default = DriverManager;
//# sourceMappingURL=DriverManager.js.map