"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const microevent_ts_1 = require("microevent.ts");
const EmulationServiceInterface_1 = tslib_1.__importDefault(require("../EmulationServiceInterface"));
const EmulationContext_1 = tslib_1.__importDefault(require("./EmulationContext"));
const Board_1 = tslib_1.__importDefault(require("../../../../machine/stella/Board"));
const CartridgeFactory_1 = tslib_1.__importDefault(require("../../../../machine/stella/cartridge/CartridgeFactory"));
const AsyncIO_1 = tslib_1.__importDefault(require("../../../../machine/stella/AsyncIO"));
const Factory_1 = tslib_1.__importDefault(require("../../../../tools/scheduler/Factory"));
const ClockProbe_1 = tslib_1.__importDefault(require("../../../../tools/ClockProbe"));
const PeriodicScheduler_1 = tslib_1.__importDefault(require("../../../../tools/scheduler/PeriodicScheduler"));
const async_mutex_1 = require("async-mutex");
const CLOCK_UPDATE_INTERVAL = 2000;
class EmulationService {
    constructor() {
        this.stateChanged = new microevent_ts_1.Event();
        this.emulationError = new microevent_ts_1.Event();
        this._enforceRateLimit = true;
        this._state = EmulationServiceInterface_1.default.State.stopped;
        this._lastError = null;
        this._scheduler = null;
        this._clockProbe = new ClockProbe_1.default(new PeriodicScheduler_1.default(CLOCK_UPDATE_INTERVAL));
        this._mutex = new async_mutex_1.Mutex();
        this._schedulerFactory = new Factory_1.default();
        this._limitingStrategy = 1;
        this.frequencyUpdate = this._clockProbe.frequencyUpdate;
        this._updateScheduler();
    }
    init() {
        return Promise.resolve();
    }
    start(buffer, config, cartridgeType) {
        const factory = new CartridgeFactory_1.default();
        return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
            try {
                this._stop();
                this._limitingStrategy = config.pcmAudio
                    ? 2
                    : 1;
                this._updateScheduler();
                if (this._state === EmulationServiceInterface_1.default.State.error) {
                    return this._state;
                }
                const cartridge = yield factory.createCartridge(buffer, cartridgeType), board = new Board_1.default(config, cartridge);
                this._board = board;
                this._board.trap.addHandler(EmulationService._trapHandler, this);
                if (config.asyncIO) {
                    this._asyncIO = new AsyncIO_1.default(board);
                }
                this._context = new EmulationContext_1.default(board, this._asyncIO);
                this._clockProbe.attach(this._board.clock);
                this._setState(EmulationServiceInterface_1.default.State.paused);
            }
            catch (e) {
                this._setError(e);
            }
            return this._state;
        }));
    }
    stop() {
        return this._mutex.runExclusive(() => this._stop());
    }
    pause() {
        return this._mutex.runExclusive(() => {
            try {
                if (this._state === EmulationServiceInterface_1.default.State.running) {
                    this._board.getTimer().stop();
                    this._board.suspend();
                    this._setState(EmulationServiceInterface_1.default.State.paused);
                    this._clockProbe.stop();
                }
            }
            catch (e) {
                this._setError(e);
            }
            return this._state;
        });
    }
    resume() {
        return this._mutex.runExclusive(() => {
            if (this._state === EmulationServiceInterface_1.default.State.paused) {
                try {
                    this._tryToResume();
                }
                catch (e) {
                    this._setError(e);
                }
            }
            return this._state;
        });
    }
    reset() {
        return this._mutex.runExclusive(() => {
            try {
                switch (this._state) {
                    case EmulationServiceInterface_1.default.State.running:
                    case EmulationServiceInterface_1.default.State.paused:
                        this._board.reset();
                        break;
                    case EmulationServiceInterface_1.default.State.error:
                        this._board.reset();
                        this._tryToResume();
                        break;
                }
            }
            catch (e) {
                this._setError(e);
            }
            return this._state;
        });
    }
    getState() {
        return this._state;
    }
    getEmulationContext() {
        switch (this._state) {
            case EmulationServiceInterface_1.default.State.running:
            case EmulationServiceInterface_1.default.State.paused:
                return this._context;
            default:
                return null;
        }
    }
    getLastError() {
        return this._lastError;
    }
    getFrequency() {
        return this._clockProbe.getFrequency();
    }
    getRateLimit() {
        return this._enforceRateLimit;
    }
    setRateLimit(enforce) {
        if (this._enforceRateLimit === enforce) {
            return Promise.resolve(undefined);
        }
        return this._mutex.runExclusive(() => {
            if (this._state === EmulationServiceInterface_1.default.State.running) {
                this._board.getTimer().stop();
            }
            this._enforceRateLimit = enforce;
            this._updateScheduler();
            if (this._state === EmulationServiceInterface_1.default.State.running) {
                this._board.getTimer().start(this._scheduler);
            }
        });
    }
    static _trapHandler(trap, self) {
        self._setError(new Error(`TRAP: ${trap.message}`));
        self.emulationError.dispatch(self._lastError);
    }
    _stop() {
        try {
            if (this._state === EmulationServiceInterface_1.default.State.running) {
                this._board.getTimer().stop();
                this._board.suspend();
                this._board.trap.removeHandler(EmulationService._trapHandler, this);
                this._clockProbe.stop().detach();
            }
            this._board = null;
            this._context = null;
            this._asyncIO = undefined;
            this._setState(EmulationServiceInterface_1.default.State.stopped);
        }
        catch (e) {
            this._setError(e);
        }
        return this._state;
    }
    _tryToResume() {
        if (this._state === EmulationServiceInterface_1.default.State.running) {
            return;
        }
        this._board.getTimer().start(this._scheduler);
        this._board.resume();
        this._setState(EmulationServiceInterface_1.default.State.running);
        this._clockProbe.start();
    }
    _setError(e) {
        this._lastError = e;
        this._setState(EmulationServiceInterface_1.default.State.error);
    }
    _setState(state) {
        if (state !== this._state) {
            this._state = state;
            this.stateChanged.dispatch(state);
        }
        return this._state;
    }
    _updateScheduler() {
        this._scheduler = this._enforceRateLimit
            ? this._schedulerFactory.createLimitingScheduler(this._limitingStrategy)
            : this._schedulerFactory.createImmediateScheduler();
    }
}
exports.default = EmulationService;
//# sourceMappingURL=EmulationService.js.map