"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const EmulationService_1 = tslib_1.__importDefault(require("../vanilla/EmulationService"));
const DriverManager_1 = tslib_1.__importDefault(require("../DriverManager"));
const VideoDriver_1 = tslib_1.__importDefault(require("./VideoDriver"));
const ControlDriver_1 = tslib_1.__importDefault(require("./ControlDriver"));
const WaveformAudioDriver_1 = tslib_1.__importDefault(require("./WaveformAudioDriver"));
const PCMAudioDriver_1 = tslib_1.__importDefault(require("./PCMAudioDriver"));
const AsyncIO_1 = tslib_1.__importDefault(require("../../../driver/AsyncIO"));
const messages_1 = require("./messages");
class EmulationBackend {
    constructor(_rpc) {
        this._rpc = _rpc;
        this._videoDriver = null;
        this._service = new EmulationService_1.default();
    }
    startup() {
        const driverManager = new DriverManager_1.default(), videoDriver = new VideoDriver_1.default(this._rpc), controlDriver = new ControlDriver_1.default(this._rpc), waveformAduioDrivers = [0, 1].map(i => new WaveformAudioDriver_1.default(i, this._rpc)), pcmAudioDriver = new PCMAudioDriver_1.default(0, this._rpc), asyncIODriver = new AsyncIO_1.default();
        this._rpc
            .registerRpcHandler(messages_1.RPC_TYPE.setup, this._onSetup.bind(this))
            .registerRpcHandler(messages_1.RPC_TYPE.emulationFetchLastError, this._onFetchLastError.bind(this))
            .registerRpcHandler(messages_1.RPC_TYPE.emulationPause, this._onEmulationPause.bind(this))
            .registerRpcHandler(messages_1.RPC_TYPE.emulationReset, this._onEmulationReset.bind(this))
            .registerRpcHandler(messages_1.RPC_TYPE.emulationResume, this._onEmulationResume.bind(this))
            .registerRpcHandler(messages_1.RPC_TYPE.emulationSetRateLimit, this._onEmulationSetRateLimit.bind(this))
            .registerRpcHandler(messages_1.RPC_TYPE.emulationStart, this._onEmulationStart.bind(this))
            .registerRpcHandler(messages_1.RPC_TYPE.emulationStop, this._onEmulationStop.bind(this))
            .registerSignalHandler(messages_1.SIGNAL_TYPE.messageToAsyncIO, data => asyncIODriver.send(data));
        asyncIODriver.message.addHandler(message => this._rpc.signal(messages_1.SIGNAL_TYPE.messageFromAsyncIO, message));
        this._service.frequencyUpdate.addHandler(EmulationBackend._onFrequencyUpdate, this);
        this._service.emulationError.addHandler(EmulationBackend._onEmulationError, this);
        this._videoDriver = videoDriver;
        videoDriver.init();
        controlDriver.init();
        driverManager
            .addDriver(videoDriver, (context, driver) => driver.bind(context.getRawVideo()))
            .addDriver(controlDriver, (context, driver) => driver.bind(context))
            .addDriver(pcmAudioDriver, (context, driver) => driver.bind(context.getPCMChannel()))
            .addDriver(asyncIODriver, (context, driver) => driver.bind(context.getAsyncIO()))
            .bind(this._service);
        for (let i = 0; i < 2; i++) {
            driverManager.addDriver(waveformAduioDrivers[i], (context, driver) => driver.bind(context.getWaveformChannels()[i]));
        }
    }
    static _onFrequencyUpdate(frequency, self) {
        self._rpc.signal(messages_1.SIGNAL_TYPE.emulationFrequencyUpdate, frequency);
    }
    static _onEmulationError(error, self) {
        self._rpc.signal(messages_1.SIGNAL_TYPE.emulationError, error ? error.message : null);
    }
    _onSetup(msg) {
        this._videoDriver.init();
    }
    _onFetchLastError() {
        const lastError = this._service.getLastError();
        return lastError ? lastError.message : null;
    }
    _onEmulationPause() {
        return this._service.pause();
    }
    _onEmulationReset() {
        return this._service.reset();
    }
    _onEmulationResume() {
        return this._service.resume();
    }
    _onEmulationStart(message) {
        return this._service.start(message.buffer, message.config, message.cartridgeType);
    }
    _onEmulationStop() {
        return this._service.stop();
    }
    _onEmulationSetRateLimit(message) {
        return this._service.setRateLimit(message);
    }
}
exports.default = EmulationBackend;
//# sourceMappingURL=EmulationBackend.js.map