"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const microevent_ts_1 = require("microevent.ts");
const worker_rpc_1 = require("worker-rpc");
const EmulationServiceInterface_1 = tslib_1.__importDefault(require("../EmulationServiceInterface"));
const EmulationContext_1 = tslib_1.__importDefault(require("./EmulationContext"));
const VideoProxy_1 = tslib_1.__importDefault(require("./VideoProxy"));
const ControlProxy_1 = tslib_1.__importDefault(require("./ControlProxy"));
const WaveformAudioProxy_1 = tslib_1.__importDefault(require("./WaveformAudioProxy"));
const PCMAudioProxy_1 = tslib_1.__importDefault(require("./PCMAudioProxy"));
const async_mutex_1 = require("async-mutex");
const messages_1 = require("./messages");
const AsyncIOProxy_1 = tslib_1.__importDefault(require("./AsyncIOProxy"));
const CONTROL_PROXY_UPDATE_INTERVAL = 25;
class EmulationService {
    constructor(_stellaWorkerUri) {
        this._stellaWorkerUri = _stellaWorkerUri;
        this.stateChanged = new microevent_ts_1.Event();
        this.frequencyUpdate = new microevent_ts_1.Event();
        this._rateLimitEnforced = true;
        this._mutex = new async_mutex_1.Mutex();
        this._worker = null;
        this._rpc = null;
        this._state = EmulationServiceInterface_1.default.State.stopped;
        this._lastError = null;
        this._emulationContext = null;
        this._frequency = 0;
        this._waveformChannels = new Array(2);
        this._pcmChannel = null;
        this._controlProxy = null;
        this._controlProxyUpdateHandle = null;
        this._proxyState = 0;
        this._savedConfig = null;
    }
    init() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            this._worker = new Worker(this._stellaWorkerUri);
            this._rpc = new worker_rpc_1.RpcProvider((message, transfer) => this._worker.postMessage(message, transfer));
            this._pcmChannel = new PCMAudioProxy_1.default(0, this._rpc).init();
            for (let i = 0; i < 2; i++) {
                this._waveformChannels[i] = new WaveformAudioProxy_1.default(i, this._rpc).init();
            }
            const videoProxy = new VideoProxy_1.default(this._rpc), controlProxy = new ControlProxy_1.default(this._rpc), asyncIOProxy = new AsyncIOProxy_1.default(this._rpc);
            asyncIOProxy.init();
            videoProxy.init();
            this._emulationContext = new EmulationContext_1.default(videoProxy, controlProxy, this._waveformChannels, this._pcmChannel, asyncIOProxy);
            this._worker.onmessage = messageEvent => this._rpc.dispatch(messageEvent.data);
            this._rpc
                .registerSignalHandler(messages_1.SIGNAL_TYPE.emulationFrequencyUpdate, this._onFrequencyUpdate.bind(this))
                .registerSignalHandler(messages_1.SIGNAL_TYPE.emulationError, this._onEmulationError.bind(this));
            this._controlProxy = controlProxy;
        });
    }
    start(buffer, config, cartridgeType) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this.stop();
            return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
                const state = yield this._rpc.rpc(messages_1.RPC_TYPE.emulationStart, {
                    buffer,
                    config,
                    cartridgeType
                });
                if (state === EmulationServiceInterface_1.default.State.paused) {
                    this._savedConfig = config;
                    this._emulationContext.setConfig(config);
                    yield this._startProxies(config);
                }
                else {
                    this._savedConfig = null;
                }
                return this._applyState(state);
            }));
        });
    }
    pause() {
        return this._mutex.runExclusive(() => this._rpc.rpc(messages_1.RPC_TYPE.emulationPause).then(state => {
            this._pauseProxies();
            return this._applyState(state);
        }));
    }
    stop() {
        return this._mutex.runExclusive(() => this._rpc.rpc(messages_1.RPC_TYPE.emulationStop).then(state => {
            this._stopProxies();
            return this._applyState(state);
        }));
    }
    reset() {
        return this._mutex.runExclusive(() => tslib_1.__awaiter(this, void 0, void 0, function* () {
            const state = yield this._rpc.rpc(messages_1.RPC_TYPE.emulationReset);
            if (this._state === EmulationServiceInterface_1.default.State.error &&
                (state === EmulationServiceInterface_1.default.State.running ||
                    state === EmulationServiceInterface_1.default.State.paused) &&
                this._savedConfig) {
                yield this._startProxies(this._savedConfig);
            }
            return this._applyState(state);
        }));
    }
    resume() {
        return this._mutex.runExclusive(() => this._rpc.rpc(messages_1.RPC_TYPE.emulationResume).then(state => {
            this._resumeProxies();
            return this._applyState(state);
        }));
    }
    setRateLimit(enforce) {
        this._rateLimitEnforced = enforce;
        return this._rpc.rpc(messages_1.RPC_TYPE.emulationSetRateLimit, enforce);
    }
    getFrequency() {
        return this._frequency;
    }
    getRateLimit() {
        return this._rateLimitEnforced;
    }
    getState() {
        return this._state;
    }
    getLastError() {
        return this._lastError;
    }
    getEmulationContext() {
        switch (this._state) {
            case EmulationServiceInterface_1.default.State.running:
            case EmulationServiceInterface_1.default.State.paused:
                return this._emulationContext;
            default:
                return null;
        }
    }
    _fetchLastError() {
        return this._rpc
            .rpc(messages_1.RPC_TYPE.emulationFetchLastError)
            .then(message => (message ? new Error(message) : null));
    }
    _applyState(state) {
        if (state === EmulationServiceInterface_1.default.State.error) {
            return this._fetchLastError().then(error => {
                this._state = state;
                this._lastError = error;
                this._stopProxies();
                this.stateChanged.dispatch(state);
                return state;
            });
        }
        else {
            this._state = state;
            this.stateChanged.dispatch(state);
            return state;
        }
    }
    _onFrequencyUpdate(message) {
        this._frequency = message;
        this.frequencyUpdate.dispatch(this._frequency);
    }
    _onEmulationError(message) {
        this._lastError = new Error(message || '');
        this._stopProxies();
        this._state = EmulationServiceInterface_1.default.State.error;
        this.stateChanged.dispatch(this._state);
    }
    _startProxies(config) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            yield this._emulationContext.getVideoProxy().start();
            for (let i = 0; i < this._waveformChannels.length; i++) {
                yield this._waveformChannels[i].start(config);
            }
            yield this._pcmChannel.start();
            this._startControlUpdates();
            this._proxyState = 1;
        });
    }
    _stopProxies() {
        if (this._proxyState === 0) {
            return;
        }
        this._emulationContext.getVideoProxy().stop();
        this._pcmChannel.stop();
        this._stopControlUpdates();
        this._proxyState = 0;
    }
    _pauseProxies() {
        if (this._proxyState !== 1) {
            return;
        }
        this._stopControlUpdates();
        this._proxyState = 2;
    }
    _resumeProxies() {
        if (this._proxyState !== 2) {
            return;
        }
        this._startControlUpdates();
        this._proxyState = 1;
    }
    _startControlUpdates() {
        if (this._controlProxyUpdateHandle === null) {
            this._controlProxyUpdateHandle = setInterval(() => this._controlProxy.sendUpdate(), CONTROL_PROXY_UPDATE_INTERVAL);
        }
    }
    _stopControlUpdates() {
        if (this._controlProxyUpdateHandle !== null) {
            clearInterval(this._controlProxyUpdateHandle);
            this._controlProxyUpdateHandle = null;
        }
    }
}
exports.default = EmulationService;
//# sourceMappingURL=EmulationService.js.map