"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = void 0;
const tslib_1 = require("tslib");
const async_mutex_1 = require("async-mutex");
const Pool_1 = tslib_1.__importDefault(require("../../../../tools/pool/Pool"));
const ArrayBufferSurface_1 = tslib_1.__importDefault(require("../../../../video/surface/ArrayBufferSurface"));
const messages_1 = require("./messages");
class VideoDriver {
    constructor(_rpc) {
        this._rpc = _rpc;
        this._active = false;
        this._video = null;
        this._mutex = new async_mutex_1.Mutex();
        this._surfacePool = null;
        this._managedSurfacesById = null;
        this._ids = null;
        this._width = 0;
        this._height = 0;
        this._nextId = 0;
    }
    init() {
        this._rpc
            .registerSignalHandler(messages_1.SIGNAL_TYPE.videoReturnSurface, this._onReturnSurfaceFromHost.bind(this))
            .registerRpcHandler(messages_1.RPC_TYPE.getVideoParameters, this._onGetVideoParameters.bind(this));
    }
    bind(video) {
        this._mutex.runExclusive(() => this._bind(video));
    }
    unbind() {
        this._mutex.runExclusive(() => this._unbind());
    }
    static _onNewFrame(surface, self) {
        if (!self._active) {
            console.warn('new frame passed to inactive driver');
            return;
        }
        if (!self._managedSurfaces.has(surface)) {
            console.warn(`surface not registered`);
            return;
        }
        const buffer = surface.getUnderlyingBuffer(), id = self._ids.get(self._managedSurfaces.get(surface));
        self._rpc.signal(messages_1.SIGNAL_TYPE.videoNewFrame, {
            id,
            width: self._width,
            height: self._height,
            buffer
        }, [buffer]);
    }
    _bind(video) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (this._active) {
                return;
            }
            this._width = video.getWidth();
            this._height = video.getHeight();
            this._video = video;
            this._surfacePool = new Pool_1.default(() => ArrayBufferSurface_1.default.createFromArrayBuffer(this._width, this._height, new ArrayBuffer(4 * this._width * this._height)));
            this._managedSurfacesById = new Map();
            this._managedSurfaces = new WeakMap();
            this._ids = new WeakMap();
            this._video.setSurfaceFactory(() => {
                const managedSurface = this._surfacePool.get(), surface = managedSurface.get();
                const isNewSurface = !this._ids.has(managedSurface);
                if (isNewSurface) {
                    const id = this._nextId++;
                    this._ids.set(managedSurface, id);
                    this._managedSurfacesById.set(id, managedSurface);
                    this._managedSurfaces.set(surface, managedSurface);
                    surface.fill(0xff000000);
                }
                return managedSurface.get();
            });
            this._video.newFrame.addHandler(VideoDriver._onNewFrame, this);
            this._active = true;
        });
    }
    _unbind() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this._active) {
                return;
            }
            this._active = false;
            this._video.setSurfaceFactory(null);
            this._video.newFrame.removeHandler(VideoDriver._onNewFrame, this);
            this._video = null;
            this._surfacePool = null;
            this._managedSurfacesById = null;
            this._ids = null;
        });
    }
    _onReturnSurfaceFromHost(message) {
        if (!this._active) {
            console.warn('surface returned from host to inactive driver');
            return;
        }
        const surface = this._managedSurfacesById.get(message.id);
        if (!surface) {
            console.warn(`invalid member ID ${message.id}`);
            return;
        }
        surface.get().replaceUnderlyingBuffer(this._width, this._height, message.buffer);
        surface.release();
    }
    _onGetVideoParameters() {
        return {
            width: this._width,
            height: this._height
        };
    }
}
exports.default = VideoDriver;
//# sourceMappingURL=VideoDriver.js.map